package consumer

import (
	"context"
	"fmt"
	"os"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/msg"

	"gitea.dwysokinski.me/twhelp/core/cmd/twhelp/internal"
	"gitea.dwysokinski.me/twhelp/core/internal/adapter"
	"gitea.dwysokinski.me/twhelp/core/internal/service"
	"github.com/ThreeDotsLabs/watermill"
	"github.com/ThreeDotsLabs/watermill-amqp/v2/pkg/amqp"
	"github.com/ThreeDotsLabs/watermill/message"
	"github.com/ThreeDotsLabs/watermill/message/router/middleware"
	"github.com/garsue/watermillzap"
	"github.com/uptrace/bun"
	"github.com/urfave/cli/v2"
	"go.uber.org/zap"
)

const (
	routerCloseTimeout      = 30 * time.Second
	maxRetries              = 3
	retryInitialInterval    = 3 * time.Second
	retryMaxInterval        = 10 * time.Second
	retryIntervalMultiplier = 1.5
	healthyFilePath         = "/tmp/healthy"
	healthyFilePerm         = 0644
)

func New() *cli.Command {
	return &cli.Command{
		Name:  "consumer",
		Usage: "Runs RabbitMQ consumers",
		Subcommands: []*cli.Command{
			newServerCommand(),
			newTribeCommand(),
			newPlayerCommand(),
			newVillageCommand(),
			newEnnoblementCommand(),
		},
	}
}

func newServerCommand() *cli.Command {
	return &cli.Command{
		Name: "server",
		Action: func(c *cli.Context) error {
			return runConsumer(
				c,
				"ServerConsumer",
				func(
					c *cli.Context,
					router *message.Router,
					logger watermill.LoggerAdapter,
					publisher *amqp.Publisher,
					subscriber *amqp.Subscriber,
					marshaler msg.Marshaler,
					bunDB *bun.DB,
				) error {
					svc := service.NewServer(adapter.NewServerBunRepository(bunDB), internal.NewTWClient(c.App.Version))
					msg.NewServerConsumer(marshaler, publisher, subscriber, logger, svc).Register(router)
					return nil
				},
			)
		},
	}
}

func newTribeCommand() *cli.Command {
	return &cli.Command{
		Name: "tribe",
		Action: func(c *cli.Context) error {
			return runConsumer(
				c,
				"TribeConsumer",
				func(
					c *cli.Context,
					router *message.Router,
					logger watermill.LoggerAdapter,
					publisher *amqp.Publisher,
					subscriber *amqp.Subscriber,
					marshaler msg.Marshaler,
					bunDB *bun.DB,
				) error {
					svc := service.NewTribe(adapter.NewTribeBunRepository(bunDB), internal.NewTWClient(c.App.Version))
					snapshotSvc := service.NewTribeSnapshot(adapter.NewTribeSnapshotBunRepository(bunDB), svc)
					msg.NewTribeConsumer(
						marshaler,
						publisher,
						subscriber,
						logger,
						svc,
						snapshotSvc,
					).Register(router)
					return nil
				},
			)
		},
	}
}

func newPlayerCommand() *cli.Command {
	return &cli.Command{
		Name: "player",
		Action: func(c *cli.Context) error {
			return runConsumer(
				c,
				"PlayerConsumer",
				func(
					c *cli.Context,
					router *message.Router,
					logger watermill.LoggerAdapter,
					publisher *amqp.Publisher,
					subscriber *amqp.Subscriber,
					marshaler msg.Marshaler,
					bunDB *bun.DB,
				) error {
					tribeChangeSvc := service.NewTribeChange(adapter.NewTribeChangeBunRepository(bunDB))
					svc := service.NewPlayer(adapter.NewPlayerBunRepository(bunDB), tribeChangeSvc, internal.NewTWClient(c.App.Version))
					snapshotSvc := service.NewPlayerSnapshot(adapter.NewPlayerSnapshotBunRepository(bunDB), svc)
					msg.NewPlayerConsumer(
						marshaler,
						publisher,
						subscriber,
						logger,
						svc,
						snapshotSvc,
					).Register(router)
					return nil
				},
			)
		},
	}
}

func newVillageCommand() *cli.Command {
	return &cli.Command{
		Name: "village",
		Action: func(c *cli.Context) error {
			return runConsumer(
				c,
				"VillageConsumer",
				func(
					c *cli.Context,
					router *message.Router,
					logger watermill.LoggerAdapter,
					publisher *amqp.Publisher,
					subscriber *amqp.Subscriber,
					marshaler msg.Marshaler,
					bunDB *bun.DB,
				) error {
					svc := service.NewVillage(adapter.NewVillageBunRepository(bunDB), internal.NewTWClient(c.App.Version))
					msg.NewVillageConsumer(marshaler, publisher, subscriber, logger, svc).Register(router)
					return nil
				},
			)
		},
	}
}

func newEnnoblementCommand() *cli.Command {
	return &cli.Command{
		Name: "ennoblement",
		Action: func(c *cli.Context) error {
			return runConsumer(
				c,
				"EnnoblementConsumer",
				func(
					c *cli.Context,
					router *message.Router,
					logger watermill.LoggerAdapter,
					publisher *amqp.Publisher,
					subscriber *amqp.Subscriber,
					marshaler msg.Marshaler,
					bunDB *bun.DB,
				) error {
					svc := service.NewEnnoblement(adapter.NewEnnoblementBunRepository(bunDB), internal.NewTWClient(c.App.Version))
					msg.NewEnnoblementConsumer(marshaler, publisher, subscriber, logger, svc).Register(router)
					return nil
				},
			)
		},
	}
}

type registerHandlersFunc func(
	c *cli.Context,
	router *message.Router,
	logger watermill.LoggerAdapter,
	publisher *amqp.Publisher,
	subscriber *amqp.Subscriber,
	marshaler msg.Marshaler,
	bunDB *bun.DB,
) error

func runConsumer(c *cli.Context, name string, registerHandlers registerHandlersFunc) error {
	logger := zap.L()

	bunDB, err := internal.NewBunDB()
	if err != nil {
		return fmt.Errorf("internal.NewBunDB: %w", err)
	}
	defer func() {
		_ = bunDB.Close()
	}()

	watermillLogger := watermillzap.NewLogger(logger)

	publisher, err := internal.NewAMQPPublisher(watermillLogger)
	if err != nil {
		return fmt.Errorf("internal.NewAMQPPublisher: %w", err)
	}
	defer func() {
		_ = publisher.Close()
	}()

	subscriber, err := internal.NewAMQPSubscriber(watermillLogger, name)
	if err != nil {
		return fmt.Errorf("internal.NewAMQPSubscriber: %w", err)
	}
	defer func() {
		_ = subscriber.Close()
	}()

	router, err := newRouter(watermillLogger)
	if err != nil {
		return fmt.Errorf("newRouter: %w", err)
	}

	if err = registerHandlers(
		c,
		router,
		watermillLogger,
		publisher,
		subscriber,
		internal.NewWatermillConsumerMarshaler(),
		bunDB,
	); err != nil {
		return fmt.Errorf("registerHandlers: %w", err)
	}

	go runRouter(c.Context, router, logger)
	defer func() {
		_ = router.Close()
	}()

	<-router.Running()
	if err := os.WriteFile(healthyFilePath, []byte("healthy"), healthyFilePerm); err != nil {
		return fmt.Errorf("couldn't create the %s file: %w", healthyFilePath, err)
	}

	logger.Info("consumer is up and running", zap.String("consumer_name", name))

	internal.WaitForSignal(c.Context)

	return nil
}

func newRouter(logger watermill.LoggerAdapter) (*message.Router, error) {
	router, err := message.NewRouter(message.RouterConfig{
		CloseTimeout: routerCloseTimeout,
	}, logger)
	if err != nil {
		return nil, fmt.Errorf("message.NewRouter: %w", err)
	}
	router.AddMiddleware(middleware.Recoverer)
	router.AddMiddleware(newRetryMiddleware(logger).Middleware)
	return router, nil
}

func newRetryMiddleware(logger watermill.LoggerAdapter) middleware.Retry {
	return middleware.Retry{
		MaxRetries:      maxRetries,
		InitialInterval: retryInitialInterval,
		MaxInterval:     retryMaxInterval,
		Multiplier:      retryIntervalMultiplier,
		Logger:          logger,
	}
}

func runRouter(ctx context.Context, router *message.Router, logger *zap.Logger) {
	if err := router.Run(ctx); err != nil {
		logger.Fatal("router.Run: " + err.Error())
	}
}
