package internal

import (
	"fmt"

	"gitea.dwysokinski.me/twhelp/core/internal/adapter"
	"gitea.dwysokinski.me/twhelp/core/internal/msg"

	"github.com/google/uuid"
	"github.com/kelseyhightower/envconfig"

	"github.com/ThreeDotsLabs/watermill"
	"github.com/ThreeDotsLabs/watermill-amqp/v2/pkg/amqp"
)

func NewAMQPPublisher(logger watermill.LoggerAdapter) (*amqp.Publisher, error) {
	cfg, err := newAMQPConfig(amqp.GenerateQueueNameTopicName)
	if err != nil {
		return nil, fmt.Errorf("newAMQPConfig: %w", err)
	}

	publisher, err := amqp.NewPublisher(cfg, logger)
	if err != nil {
		return nil, fmt.Errorf("amqp.NewPublisher: %w", err)
	}

	return publisher, nil
}

func NewAMQPSubscriber(logger watermill.LoggerAdapter, queueNameSuffix string) (*amqp.Subscriber, error) {
	cfg, err := newAMQPConfig(generateQueueNameTopicNameWithSuffix(queueNameSuffix))
	if err != nil {
		return nil, fmt.Errorf("newAMQPConfig: %w", err)
	}

	subscriber, err := amqp.NewSubscriber(cfg, logger)
	if err != nil {
		return nil, fmt.Errorf("amqp.NewSubscriber: %w", err)
	}

	return subscriber, nil
}

type amqpConfig struct {
	URI           string `envconfig:"URI" required:"true"`
	PrefetchCount int    `envconfig:"PREFETCH_COUNT" default:"2"`
}

func newAMQPConfig(generateQueueName amqp.QueueNameGenerator) (amqp.Config, error) {
	var appCfg amqpConfig
	if err := envconfig.Process("AMQP", &appCfg); err != nil {
		return amqp.Config{}, fmt.Errorf("envconfig.Process: %w", err)
	}

	cfg := amqp.NewDurablePubSubConfig(appCfg.URI, generateQueueName)
	cfg.Consume.NoRequeueOnNack = true
	cfg.Consume.Qos.PrefetchCount = appCfg.PrefetchCount
	cfg.Exchange.GenerateName = func(topic string) string {
		return "twhelp_" + topic
	}
	return cfg, nil
}

func NewWatermillPublisherMarshaler() adapter.WatermillMarshaler {
	return adapter.JSONWatermillMarshaler{NewUUID: uuid.NewString}
}

func NewWatermillConsumerMarshaler() msg.Marshaler {
	return msg.JSONMarshaler{NewUUID: uuid.NewString}
}

func generateQueueNameTopicNameWithSuffix(suffix string) amqp.QueueNameGenerator {
	return func(topic string) string {
		return "twhelp_" + topic + "_" + suffix
	}
}
