package adapter

import (
	"context"
	"errors"
	"fmt"
	"sync"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"github.com/uptrace/bun"
)

func appendODSetClauses(q *bun.InsertQuery) *bun.InsertQuery {
	return q.Set("rank_att = EXCLUDED.rank_att").
		Set("score_att = EXCLUDED.score_att").
		Set("rank_def = EXCLUDED.rank_def").
		Set("score_def = EXCLUDED.score_def").
		Set("rank_sup = EXCLUDED.rank_sup").
		Set("score_sup = EXCLUDED.score_sup").
		Set("rank_total = EXCLUDED.rank_total").
		Set("score_total = EXCLUDED.score_total")
}

type paginationApplier struct {
	pagination domain.Pagination
}

func (p paginationApplier) apply(q *bun.SelectQuery) *bun.SelectQuery {
	if p.pagination.Limit > 0 {
		q = q.Limit(int(p.pagination.Limit))
	}

	if p.pagination.Offset > 0 {
		q = q.Offset(int(p.pagination.Offset))
	}

	return q
}

func sortDirectionToString(sd domain.SortDirection) (string, error) {
	switch sd {
	case domain.SortDirectionASC:
		return "ASC", nil
	case domain.SortDirectionDESC:
		return "DESC", nil
	}
	return "", fmt.Errorf("%w: %d", domain.ErrUnsupportedSortDirection, sd)
}

const (
	scanAndCountErrChSize   = 2
	scanAndCountCountChSize = 1
)

func scanAndCount(
	ctx context.Context,
	countQ *bun.SelectQuery,
	scanQ *bun.SelectQuery,
) (int64, error) {
	var wg sync.WaitGroup
	errCh := make(chan error, scanAndCountErrChSize) // 2 goroutines can send err to this channel
	defer close(errCh)
	countCh := make(chan int64, scanAndCountCountChSize) // only 1 goroutine can send cnt to this channel
	defer close(countCh)

	wg.Add(1)
	go func() {
		defer wg.Done()

		count, err := countQ.Count(ctx)
		if err != nil {
			errCh <- err
			return
		}
		countCh <- int64(count)
	}()

	wg.Add(1)
	go func() {
		defer wg.Done()

		if err := scanQ.Scan(ctx); err != nil {
			errCh <- err
		}
	}()

	wg.Wait()

	select {
	case err := <-errCh:
		return 0, err
	default:
	}

	select {
	case cnt := <-countCh:
		return cnt, nil
	default:
		return 0, errors.New("couldn't get the result of the count query from the channel")
	}
}
