package bundbtest

import (
	"context"
	"database/sql"
	"strings"
	"time"
	"unicode"

	"github.com/cenkalti/backoff/v4"
	"github.com/google/uuid"
	"github.com/pressly/goose/v3"
	"github.com/stretchr/testify/require"
	"github.com/uptrace/bun"
	"github.com/uptrace/bun/dialect/pgdialect"
	"github.com/uptrace/bun/driver/pgdriver"
)

const (
	migrationTimeout    = 10 * time.Second
	retryMaxInterval    = 5 * time.Second
	retryMaxElapsedTime = 30 * time.Second
)

// NewDB creates a new *bun.DB. Don't forget to set up goose before using this function.
func NewDB(t TestingT, dsn string) *bun.DB {
	t.Helper()

	schema := generateSchema()

	sqldb := sql.OpenDB(
		pgdriver.NewConnector(
			pgdriver.WithDSN(dsn),
			pgdriver.WithConnParams(map[string]any{
				"search_path": schema,
			}),
		),
	)

	bunDB := bun.NewDB(sqldb, pgdialect.New())
	t.Cleanup(func() {
		_ = bunDB.Close()
	})

	bo := backoff.NewExponentialBackOff()
	bo.MaxInterval = retryMaxInterval
	bo.MaxElapsedTime = retryMaxElapsedTime
	require.NoError(t, backoff.Retry(bunDB.Ping, bo), "couldn't ping DB")

	_, err := bunDB.Exec("CREATE SCHEMA ?", bun.Safe(schema))
	require.NoError(t, err, "couldn't create schema")

	runMigrations(t, sqldb)

	return bunDB
}

func runMigrations(t TestingT, db *sql.DB) {
	t.Helper()

	ctx, cancel := context.WithTimeout(context.Background(), migrationTimeout)
	defer cancel()

	require.NoError(t, goose.UpContext(ctx, db, "."), "couldn't migrate")
}

func generateSchema() string {
	return strings.TrimFunc(strings.ReplaceAll(uuid.NewString(), "-", "_"), unicode.IsNumber)
}
