package bunmodel

import (
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"github.com/uptrace/bun"
)

type Player struct {
	bun.BaseModel `bun:"table:players,alias:player"`

	ID             int64     `bun:"id,nullzero,pk"`
	ServerKey      string    `bun:"server_key,nullzero,pk"`
	Server         Server    `bun:"server,rel:belongs-to,join:server_key=key"`
	Name           string    `bun:"name,nullzero"`
	NumVillages    int64     `bun:"num_villages"`
	Points         int64     `bun:"points"`
	Rank           int64     `bun:"rank"`
	TribeID        int64     `bun:"tribe_id,nullzero"`
	Tribe          Tribe     `bun:"tribe,rel:belongs-to,join:tribe_id=id,join:server_key=server_key"`
	ProfileURL     string    `bun:"profile_url,nullzero"`
	BestRank       int64     `bun:"best_rank"`
	BestRankAt     time.Time `bun:"best_rank_at,nullzero"`
	MostPoints     int64     `bun:"most_points"`
	MostPointsAt   time.Time `bun:"most_points_at,nullzero"`
	MostVillages   int64     `bun:"most_villages"`
	MostVillagesAt time.Time `bun:"most_villages_at,nullzero"`
	LastActivityAt time.Time `bun:"last_activity_at,nullzero"`
	CreatedAt      time.Time `bun:"created_at,nullzero"`
	DeletedAt      time.Time `bun:"deleted_at,nullzero"`

	OpponentsDefeated
}

func NewPlayer(p domain.CreatePlayerParams) Player {
	return Player{
		OpponentsDefeated: OpponentsDefeated(p.OpponentsDefeated),
		ID:                p.ID,
		ServerKey:         p.ServerKey,
		Name:              p.Name,
		NumVillages:       p.NumVillages,
		Points:            p.Points,
		Rank:              p.Rank,
		TribeID:           p.TribeID,
		ProfileURL:        p.ProfileURL,
		BestRank:          p.BestRank,
		BestRankAt:        p.BestRankAt,
		MostPoints:        p.MostPoints,
		MostPointsAt:      p.MostPointsAt,
		MostVillages:      p.MostVillages,
		MostVillagesAt:    p.MostVillagesAt,
		LastActivityAt:    p.LastActivityAt,
		CreatedAt:         time.Now(),
	}
}

func (p Player) ToDomain() domain.Player {
	return domain.Player{
		BasePlayer: domain.BasePlayer{
			OpponentsDefeated: domain.OpponentsDefeated(p.OpponentsDefeated),
			ID:                p.ID,
			Name:              p.Name,
			NumVillages:       p.NumVillages,
			Points:            p.Points,
			Rank:              p.Rank,
			TribeID:           p.TribeID,
			ProfileURL:        p.ProfileURL,
		},
		BestRank:       p.BestRank,
		BestRankAt:     p.BestRankAt,
		MostPoints:     p.MostPoints,
		MostPointsAt:   p.MostPointsAt,
		MostVillages:   p.MostVillages,
		MostVillagesAt: p.MostVillagesAt,
		LastActivityAt: p.LastActivityAt,
		ServerKey:      p.ServerKey,
		CreatedAt:      p.CreatedAt,
		DeletedAt:      p.DeletedAt,
	}
}

func (p Player) ToDomainWithRelations() domain.PlayerWithRelations {
	res := domain.PlayerWithRelations{
		Player: p.ToDomain(),
		Server: p.Server.ToMeta(),
	}

	if p.Tribe.ID > 0 {
		res.Tribe = domain.NullTribeMeta{
			Valid: true,
			Tribe: p.Tribe.ToMeta(),
		}
	}

	return res
}

func (p Player) ToMeta() domain.PlayerMeta {
	return domain.PlayerMeta{
		ID:         p.ID,
		Name:       p.Name,
		ProfileURL: p.ProfileURL,
	}
}

func (p Player) ToMetaWithRelations() domain.PlayerMetaWithRelations {
	res := domain.PlayerMetaWithRelations{
		PlayerMeta: p.ToMeta(),
	}

	if p.Tribe.ID > 0 {
		res.Tribe = domain.NullTribeMeta{
			Valid: true,
			Tribe: p.Tribe.ToMeta(),
		}
	}

	return res
}
