package adapter

import (
	"context"
	"database/sql"
	"errors"
	"fmt"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/adapter/internal/bunmodel"
	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"github.com/uptrace/bun"
)

var (
	playerMetaColumns = []string{"id", "name", "profile_url"}
	playerOrders      = []string{"player.server_key ASC"}
)

type PlayerBunRepository struct {
	db *bun.DB
}

func NewPlayerBunRepository(db *bun.DB) *PlayerBunRepository {
	return &PlayerBunRepository{db: db}
}

func (p *PlayerBunRepository) CreateOrUpdate(ctx context.Context, params ...domain.CreatePlayerParams) error {
	if len(params) == 0 {
		return nil
	}

	players := make([]bunmodel.Player, 0, len(params))
	for _, param := range params {
		players = append(players, bunmodel.NewPlayer(param))
	}
	if _, err := p.db.NewInsert().
		Model(&players).
		On("CONFLICT ON CONSTRAINT players_pkey DO UPDATE").
		Set("name = EXCLUDED.name").
		Set("num_villages = EXCLUDED.num_villages").
		Set("points = EXCLUDED.points").
		Set("rank = EXCLUDED.rank").
		Set("tribe_id = EXCLUDED.tribe_id").
		Set("profile_url = EXCLUDED.profile_url").
		Set("best_rank = EXCLUDED.best_rank").
		Set("best_rank_at = EXCLUDED.best_rank_at").
		Set("most_villages = EXCLUDED.most_villages").
		Set("most_villages_at = EXCLUDED.most_villages_at").
		Set("most_points = EXCLUDED.most_points").
		Set("most_points_at = EXCLUDED.most_points_at").
		Set("last_activity_at = EXCLUDED.last_activity_at").
		Set("deleted_at = null").
		Apply(appendODSetClauses).
		Returning("NULL").
		Exec(ctx); err != nil {
		return fmt.Errorf("something went wrong while inserting players into the db: %w", err)
	}

	return nil
}

func (p *PlayerBunRepository) Delete(ctx context.Context, serverKey string, ids ...int64) error {
	if len(ids) == 0 {
		return nil
	}

	if _, err := p.db.NewUpdate().
		Model(&bunmodel.Player{}).
		Where("deleted_at IS NULL").
		Where("id IN (?)", bun.In(ids)).
		Where("server_key = ?", serverKey).
		Set("deleted_at = ?", time.Now()).
		Set("tribe_id = NULL").
		Returning("NULL").
		Exec(ctx); err != nil {
		return fmt.Errorf("couldn't delete players: %w", err)
	}

	return nil
}

func (p *PlayerBunRepository) List(ctx context.Context, params domain.ListPlayersParams) ([]domain.Player, error) {
	var players []bunmodel.Player

	if err := p.db.NewSelect().
		Model(&players).
		Order(playerOrders...).
		Apply(listPlayersParamsApplier{params: params}.apply).
		Scan(ctx); err != nil && !errors.Is(err, sql.ErrNoRows) {
		return nil, fmt.Errorf("couldn't select players from the db: %w", err)
	}

	result := make([]domain.Player, 0, len(players))
	for _, player := range players {
		result = append(result, player.ToDomain())
	}

	return result, nil
}

func (p *PlayerBunRepository) ListCountWithRelations(
	ctx context.Context,
	params domain.ListPlayersParams,
) ([]domain.PlayerWithRelations, int64, error) {
	var players []bunmodel.Player

	paramsApplier := listPlayersParamsApplier{params}

	subQ := p.db.NewSelect().
		Column("id", "server_key").
		Model(&bunmodel.Player{}).
		Order(playerOrders...).
		Apply(paramsApplier.apply)

	cntQ := p.db.NewSelect().
		Model(&bunmodel.Player{}).
		Apply(paramsApplier.applyFilters)

	q := p.db.NewSelect().
		Model(&players).
		Order(playerOrders...).
		Apply(paramsApplier.applySort).
		Where("(player.id, player.server_key) IN (?)", subQ).
		Relation("Tribe", func(q *bun.SelectQuery) *bun.SelectQuery {
			return q.Column(tribeMetaColumns...)
		}).
		Relation("Server", func(q *bun.SelectQuery) *bun.SelectQuery {
			return q.Column(serverMetaColumns...)
		})

	count, err := scanAndCount(ctx, cntQ, q)
	if err != nil && !errors.Is(err, sql.ErrNoRows) {
		return nil, 0, fmt.Errorf("couldn't select players from the db: %w", err)
	}

	result := make([]domain.PlayerWithRelations, 0, len(players))
	for _, player := range players {
		result = append(result, player.ToDomainWithRelations())
	}

	return result, count, nil
}

type listPlayersParamsApplier struct {
	params domain.ListPlayersParams
}

func (l listPlayersParamsApplier) apply(q *bun.SelectQuery) *bun.SelectQuery {
	return q.Apply(l.applySort).
		Apply(l.applyFilters).
		Apply(l.applyPagination)
}

func (l listPlayersParamsApplier) applyFilters(q *bun.SelectQuery) *bun.SelectQuery {
	if len(l.params.IDs) > 0 {
		q = q.Where("player.id IN (?)", bun.In(l.params.IDs))
	}

	if l.params.IDGT.Valid {
		q = q.Where("player.id > ?", l.params.IDGT.Int64)
	}

	if len(l.params.Names) > 0 {
		q = q.Where("player.name IN (?)", bun.In(l.params.Names))
	}

	if len(l.params.ServerKeys) > 0 {
		q = q.Where("player.server_key IN (?)", bun.In(l.params.ServerKeys))
	}

	if l.params.TribeID.Valid {
		q = q.Where("player.tribe_id = ?", l.params.TribeID.Int64)
	}

	if l.params.Deleted.Valid {
		if l.params.Deleted.Bool {
			q = q.Where("player.deleted_at IS NOT NULL")
		} else {
			q = q.Where("player.deleted_at IS NULL")
		}
	}

	if len(l.params.ServerKeysNIN) > 0 {
		q = q.Where("player.server_key NOT IN (?)", bun.In(l.params.ServerKeysNIN))
	}

	if len(l.params.VersionCodes) > 0 {
		q = q.Join("INNER JOIN servers AS server ON server.key = player.server_key AND server.version_code IN (?)", bun.In(l.params.VersionCodes))
	}

	return q
}

func (l listPlayersParamsApplier) applySort(q *bun.SelectQuery) *bun.SelectQuery {
	if len(l.params.Sort) == 0 {
		return q
	}

	orders := make([]string, 0, len(l.params.Sort))
	for i, s := range l.params.Sort {
		column, err := playerSortByToColumn(s.By)
		if err != nil {
			return q.Err(fmt.Errorf("playerSortByToColumn (index=%d): %w", i, err))
		}

		direction, err := sortDirectionToString(s.Direction)
		if err != nil {
			return q.Err(fmt.Errorf("sortDirectionToString (index=%d): %w", i, err))
		}

		orders = append(orders, column+" "+direction)
	}

	return q.Order(orders...)
}

func (l listPlayersParamsApplier) applyPagination(q *bun.SelectQuery) *bun.SelectQuery {
	return (paginationApplier{pagination: l.params.Pagination}).apply(q)
}

func playerSortByToColumn(sortBy domain.PlayerSortBy) (string, error) {
	switch sortBy {
	case domain.PlayerSortByID:
		return "player.id", nil
	case domain.PlayerSortByScoreAtt:
		return "player.score_att", nil
	case domain.PlayerSortByScoreDef:
		return "player.score_def", nil
	case domain.PlayerSortByScoreSup:
		return "player.score_sup", nil
	case domain.PlayerSortByScoreTotal:
		return "player.score_total", nil
	case domain.PlayerSortByPoints:
		return "player.points", nil
	case domain.PlayerSortByDeletedAt:
		return "player.deleted_at", nil
	}
	return "", fmt.Errorf("%w: %d", domain.ErrUnsupportedSortBy, sortBy)
}
