package adapter_test

import (
	"context"
	"testing"
	"time"

	"github.com/stretchr/testify/require"

	"gitea.dwysokinski.me/twhelp/core/internal/adapter"
	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"github.com/jackc/pgerrcode"
	"github.com/stretchr/testify/assert"
	"github.com/uptrace/bun/driver/pgdriver"
)

func TestPlayerBunRepository_CreateOrUpdate(t *testing.T) {
	t.Parallel()

	if testing.Short() {
		t.Skip("skipping long-running test")
	}

	db := newBunDB(t)
	fixture := loadFixtures(t, db)
	repo := adapter.NewPlayerBunRepository(db)

	t.Run("OK", func(t *testing.T) {
		t.Parallel()

		tribeTT := fixture.Tribe(t, "pl169-tt")
		tribeCSA := fixture.Tribe(t, "pl169-csa")
		serverKey := tribeTT.ServerKey
		createParams := []domain.CreatePlayerParams{
			{
				BasePlayer: domain.BasePlayer{
					OpponentsDefeated: domain.OpponentsDefeated{
						RankAtt:    1,
						ScoreAtt:   2,
						RankDef:    3,
						ScoreDef:   4,
						RankSup:    5,
						ScoreSup:   6,
						RankTotal:  7,
						ScoreTotal: 8,
					},
					ID:          999,
					Name:        "name 999",
					NumVillages: 3,
					Points:      4,
					Rank:        6,
					TribeID:     tribeTT.ID,
					ProfileURL:  "profile-999",
				},
				BestRank:       111,
				BestRankAt:     time.Now().Add(-5 * time.Minute),
				MostPoints:     222,
				MostPointsAt:   time.Now().Add(-1 * time.Minute),
				MostVillages:   333,
				MostVillagesAt: time.Now().Add(-12 * time.Minute),
				LastActivityAt: time.Now().Add(-15 * time.Minute),
				ServerKey:      serverKey,
			},
			{
				BasePlayer: domain.BasePlayer{
					OpponentsDefeated: domain.OpponentsDefeated{
						RankAtt:    8,
						ScoreAtt:   7,
						RankDef:    6,
						ScoreDef:   5,
						RankSup:    4,
						ScoreSup:   3,
						RankTotal:  2,
						ScoreTotal: 1,
					},
					ID:          998,
					Name:        "name 998",
					NumVillages: 5,
					Points:      4,
					Rank:        2,
					ProfileURL:  "profile-998",
				},
				BestRank:       333,
				BestRankAt:     time.Now().Add(-51 * time.Minute),
				MostPoints:     111,
				MostPointsAt:   time.Now().Add(-11 * time.Minute),
				MostVillages:   222,
				MostVillagesAt: time.Now().Add(-15 * time.Minute),
				LastActivityAt: time.Now().Add(-25 * time.Minute),
				ServerKey:      serverKey,
			},
		}
		updateParams := make([]domain.CreatePlayerParams, 0, len(createParams))
		ids := make([]int64, 0, len(createParams))
		for i, p := range createParams {
			var tribeID int64
			if i%2 == 0 {
				tribeID = tribeCSA.ID
			}

			updateParams = append(updateParams, domain.CreatePlayerParams{
				BasePlayer: domain.BasePlayer{
					OpponentsDefeated: domain.OpponentsDefeated{
						RankAtt:    p.RankAtt + 1,
						ScoreAtt:   p.ScoreAtt + 1,
						RankDef:    p.RankDef + 1,
						ScoreDef:   p.ScoreDef + 1,
						RankSup:    p.RankSup + 1,
						ScoreSup:   p.ScoreSup + 1,
						RankTotal:  p.RankTotal + 1,
						ScoreTotal: p.ScoreTotal + 1,
					},
					ID:          p.ID,
					Name:        p.Name + "Update",
					NumVillages: p.NumVillages + 1,
					Points:      p.Points + 1,
					Rank:        p.Rank + 1,
					TribeID:     tribeID,
					ProfileURL:  p.ProfileURL + "Update",
				},
				BestRank:       p.BestRank - 1,
				BestRankAt:     p.BestRankAt.Add(time.Minute),
				MostPoints:     p.MostPoints + 1,
				MostPointsAt:   p.MostPointsAt.Add(time.Minute),
				MostVillages:   p.MostVillages + 1,
				MostVillagesAt: p.MostVillagesAt.Add(time.Minute),
				LastActivityAt: p.LastActivityAt.Add(time.Minute),
				ServerKey:      p.ServerKey,
			})
			ids = append(ids, p.ID)
		}
		listParams := domain.ListPlayersParams{
			IDs:        ids,
			ServerKeys: []string{serverKey},
		}
		ctx := context.Background()

		assert.NoError(t, repo.CreateOrUpdate(ctx, createParams...))
		createdPlayers, err := repo.List(ctx, listParams)
		assert.NoError(t, err)
		assertCreatedPlayers(t, createParams, createdPlayers)

		assert.NoError(t, repo.CreateOrUpdate(ctx, updateParams...))
		updatedPlayers, err := repo.List(ctx, listParams)
		assert.NoError(t, err)
		assertCreatedPlayers(t, updateParams, updatedPlayers)
	})

	t.Run("OK: len(params) == 0", func(t *testing.T) {
		t.Parallel()

		assert.NoError(t, repo.CreateOrUpdate(context.Background()))
	})

	t.Run("ERR: server doesn't exist", func(t *testing.T) {
		t.Parallel()

		err := repo.CreateOrUpdate(context.Background(), domain.CreatePlayerParams{
			BasePlayer: domain.BasePlayer{
				OpponentsDefeated: domain.OpponentsDefeated{
					RankAtt:    1,
					ScoreAtt:   2,
					RankDef:    3,
					ScoreDef:   4,
					RankSup:    5,
					ScoreSup:   6,
					RankTotal:  7,
					ScoreTotal: 8,
				},
				ID:          999,
				Name:        "name 999",
				NumVillages: 3,
				Points:      4,
				Rank:        6,
			},
			ServerKey: "pl999",
		})
		var pgErr pgdriver.Error
		assert.ErrorAs(t, err, &pgErr)
		assert.Equal(t, pgerrcode.ForeignKeyViolation, pgErr.Field('C'))
		assert.Equal(t, "players_server_key_fkey", pgErr.Field('n'))
	})
}

func TestPlayerBunRepository_Delete(t *testing.T) {
	t.Parallel()

	if testing.Short() {
		t.Skip("skipping long-running test")
	}

	db := newBunDB(t)
	fixture := loadFixtures(t, db)
	repo := adapter.NewPlayerBunRepository(db)

	t.Run("OK", func(t *testing.T) {
		t.Parallel()

		playerChceszRemont := fixture.Player(t, "pl169-chcesz-remont")
		require.Zero(t, playerChceszRemont.DeletedAt)
		playerMaddov := fixture.Player(t, "pl169-maddov")
		require.Zero(t, playerMaddov.DeletedAt)

		assert.NoError(t, repo.Delete(context.Background(), "pl169", playerChceszRemont.ID, playerMaddov.ID))
		players, err := repo.List(context.Background(), domain.ListPlayersParams{
			IDs:        []int64{playerChceszRemont.ID, playerMaddov.ID},
			ServerKeys: []string{playerChceszRemont.ServerKey},
		})
		assert.NoError(t, err)
		assert.Len(t, players, 2)
		for _, player := range players {
			assert.WithinDuration(t, time.Now(), player.DeletedAt, 10*time.Second)
			assert.Zero(t, player.TribeID)
		}
	})

	t.Run("OK: len(ids) == 0", func(t *testing.T) {
		t.Parallel()

		assert.NoError(t, repo.Delete(context.Background(), "pl169"))
	})
}

func TestPlayerBunRepository_List_ListCountWithRelations(t *testing.T) {
	t.Parallel()

	if testing.Short() {
		t.Skip("skipping long-running test")
	}

	db := newBunDB(t)
	fixture := loadFixtures(t, db)
	repo := adapter.NewPlayerBunRepository(db)
	players := fixture.Players(t)
	tribeCSA := fixture.Tribe(t, "pl169-csa")

	type expectedPlayer struct {
		id        int64
		serverKey string
	}

	allPlayers := make([]expectedPlayer, 0, len(players))
	//nolint:prealloc
	var playersIT70 []expectedPlayer
	var playersPL []expectedPlayer
	var playersPL169CSA []expectedPlayer
	var playersExceptPL169 []expectedPlayer
	for _, p := range players {
		allPlayers = append(allPlayers, expectedPlayer{
			id:        p.ID,
			serverKey: p.ServerKey,
		})

		if p.ServerKey == "it70" {
			playersIT70 = append(playersIT70, expectedPlayer{
				id:        p.ID,
				serverKey: p.ServerKey,
			})
		}

		if p.ServerKey == "pl169" && p.TribeID == tribeCSA.ID {
			playersPL169CSA = append(playersPL169CSA, expectedPlayer{
				id:        p.ID,
				serverKey: p.ServerKey,
			})
		}

		if fixture.Server(t, p.ServerKey).VersionCode == "pl" {
			playersPL = append(playersPL, expectedPlayer{
				id:        p.ID,
				serverKey: p.ServerKey,
			})
		}

		if p.ServerKey != "pl169" {
			playersExceptPL169 = append(playersExceptPL169, expectedPlayer{
				id:        p.ID,
				serverKey: p.ServerKey,
			})
		}
	}

	tests := []struct {
		name            string
		params          domain.ListPlayersParams
		expectedPlayers []expectedPlayer
		expectedCount   int64
		expectedErr     error
	}{
		{
			name: "Sort=[{By=ID,Direction=ASC}]",
			params: domain.ListPlayersParams{
				Sort: []domain.PlayerSort{{By: domain.PlayerSortByID, Direction: domain.SortDirectionASC}},
			},
			expectedPlayers: allPlayers,
			expectedCount:   int64(len(allPlayers)),
		},
		{
			name: "ServerKeys=[it70],Sort=[{By=ID,Direction=ASC}]",
			params: domain.ListPlayersParams{
				ServerKeys: []string{"it70"},
				Sort:       []domain.PlayerSort{{By: domain.PlayerSortByID, Direction: domain.SortDirectionASC}},
			},
			expectedPlayers: playersIT70,
			expectedCount:   int64(len(playersIT70)),
		},
		{
			name:   "IDs=[699734930]",
			params: domain.ListPlayersParams{IDs: []int64{699734930}},
			expectedPlayers: []expectedPlayer{
				{
					id:        699734930,
					serverKey: "pl169",
				},
			},
			expectedCount: 1,
		},
		{
			name: "IDs=[6180190,8419570,699736927],IDGT=6180190,ServerKeys=[pl169],Sort=[{By=ID,Direction=ASC}]",
			params: domain.ListPlayersParams{
				IDs:        []int64{6180190, 8419570, 699736927},
				IDGT:       domain.NullInt64{Valid: true, Int64: 6180190},
				ServerKeys: []string{"pl169"},
				Sort:       []domain.PlayerSort{{By: domain.PlayerSortByID, Direction: domain.SortDirectionASC}},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        8419570,
					serverKey: "pl169",
				},
				{
					id:        699736927,
					serverKey: "pl169",
				},
			},
			expectedCount: 2,
		},
		{
			name: "IDs=[1111,1112,11428054],Deleted=false,ServerKeys=[en113]",
			params: domain.ListPlayersParams{
				IDs:        []int64{1111, 1112, 11428054},
				ServerKeys: []string{"en113"},
				Deleted: domain.NullBool{
					Bool:  false,
					Valid: true,
				},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        11428054,
					serverKey: "en113",
				},
			},
			expectedCount: 1,
		},
		{
			name: "IDs=[6180190,8419570,699736927],Sort=[{By=ID,Direction=ASC}],Limit=1,ServerKeys=[pl169]",
			params: domain.ListPlayersParams{
				IDs:        []int64{6180190, 8419570, 699736927},
				ServerKeys: []string{"pl169"},
				Sort:       []domain.PlayerSort{{By: domain.PlayerSortByID, Direction: domain.SortDirectionASC}},
				Pagination: domain.Pagination{
					Limit: 1,
				},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        6180190,
					serverKey: "pl169",
				},
			},
			expectedCount: 3,
		},
		{
			name: "IDs=[6180190,8419570,699736927],Sort=[{By=ID,Direction=DESC}],Offset=1,ServerKeys=[pl169]",
			params: domain.ListPlayersParams{
				IDs:        []int64{6180190, 8419570, 699736927},
				ServerKeys: []string{"pl169"},
				Sort:       []domain.PlayerSort{{By: domain.PlayerSortByID, Direction: domain.SortDirectionDESC}},
				Pagination: domain.Pagination{
					Offset: 1,
				},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        6180190,
					serverKey: "pl169",
				},
				{
					id:        8419570,
					serverKey: "pl169",
				},
			},
			expectedCount: 3,
		},
		{
			name: "IDs=[6180190,8419570,699736927],Sort=[{By=ScoreAtt,Direction=ASC}],Offset=2,Limit=1,ServerKeys=[pl169]",
			params: domain.ListPlayersParams{
				IDs:        []int64{6180190, 8419570, 699736927},
				ServerKeys: []string{"pl169"},
				Sort:       []domain.PlayerSort{{By: domain.PlayerSortByScoreAtt, Direction: domain.SortDirectionASC}},
				Pagination: domain.Pagination{
					Limit:  1,
					Offset: 2,
				},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        699736927,
					serverKey: "pl169",
				},
			},
			expectedCount: 3,
		},
		{
			name: "IDs=[6180190,8419570,699736927],Sort=[{By=ScoreAtt,Direction=DESC}],Offset=2,Limit=1,ServerKeys=[pl169]",
			params: domain.ListPlayersParams{
				IDs:        []int64{6180190, 8419570, 699736927},
				ServerKeys: []string{"pl169"},
				Sort:       []domain.PlayerSort{{By: domain.PlayerSortByScoreAtt, Direction: domain.SortDirectionDESC}},
				Pagination: domain.Pagination{
					Limit:  1,
					Offset: 2,
				},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        8419570,
					serverKey: "pl169",
				},
			},
			expectedCount: 3,
		},
		{
			name: "IDs=[6180190,8419570,699736927],Sort=[{By=ScoreDef,Direction=ASC}],Limit=1,ServerKeys=[pl169]",
			params: domain.ListPlayersParams{
				IDs:        []int64{6180190, 8419570, 699736927},
				ServerKeys: []string{"pl169"},
				Sort:       []domain.PlayerSort{{By: domain.PlayerSortByScoreDef, Direction: domain.SortDirectionASC}},
				Pagination: domain.Pagination{
					Limit: 1,
				},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        8419570,
					serverKey: "pl169",
				},
			},
			expectedCount: 3,
		},
		{
			name: "IDs=[6180190,8419570,699736927],Sort=[{By=ScoreDef,Direction=DESC}],Limit=1,ServerKeys=[pl169]",
			params: domain.ListPlayersParams{
				IDs:        []int64{6180190, 8419570, 699736927},
				ServerKeys: []string{"pl169"},
				Sort:       []domain.PlayerSort{{By: domain.PlayerSortByScoreDef, Direction: domain.SortDirectionDESC}},
				Pagination: domain.Pagination{
					Limit: 1,
				},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        6180190,
					serverKey: "pl169",
				},
			},
			expectedCount: 3,
		},
		{
			name: "IDs=[6180190,8419570,699736927],Sort=[{By=ScoreSup,Direction=ASC}],Limit=1,ServerKeys=[pl169]",
			params: domain.ListPlayersParams{
				IDs:        []int64{6180190, 8419570, 699736927},
				ServerKeys: []string{"pl169"},
				Sort:       []domain.PlayerSort{{By: domain.PlayerSortByScoreSup, Direction: domain.SortDirectionASC}},
				Pagination: domain.Pagination{
					Limit: 1,
				},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        699736927,
					serverKey: "pl169",
				},
			},
			expectedCount: 3,
		},
		{
			name: "IDs=[6180190,8419570,699736927],Sort=[{By=ScoreSup,Direction=DESC}],Limit=1,ServerKeys=[pl169]",
			params: domain.ListPlayersParams{
				IDs:        []int64{6180190, 8419570, 699736927},
				ServerKeys: []string{"pl169"},
				Sort:       []domain.PlayerSort{{By: domain.PlayerSortByScoreSup, Direction: domain.SortDirectionDESC}},
				Pagination: domain.Pagination{
					Limit: 1,
				},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        8419570,
					serverKey: "pl169",
				},
			},
			expectedCount: 3,
		},
		{
			name: "IDs=[6180190,8419570,699736927],Sort=[{By=ScoreTotal,Direction=ASC}],Limit=1,ServerKeys=[pl169]",
			params: domain.ListPlayersParams{
				IDs:        []int64{6180190, 8419570, 699736927},
				ServerKeys: []string{"pl169"},
				Sort:       []domain.PlayerSort{{By: domain.PlayerSortByScoreTotal, Direction: domain.SortDirectionASC}},
				Pagination: domain.Pagination{
					Limit: 1,
				},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        8419570,
					serverKey: "pl169",
				},
			},
			expectedCount: 3,
		},
		{
			name: "IDs=[6180190,8419570,699736927],Sort=[{By=ScoreTotal,Direction=DESC}],Limit=1,ServerKeys=[pl169]",
			params: domain.ListPlayersParams{
				IDs:        []int64{6180190, 8419570, 699736927},
				ServerKeys: []string{"pl169"},
				Sort:       []domain.PlayerSort{{By: domain.PlayerSortByScoreTotal, Direction: domain.SortDirectionDESC}},
				Pagination: domain.Pagination{
					Limit: 1,
				},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        6180190,
					serverKey: "pl169",
				},
			},
			expectedCount: 3,
		},
		{
			name: "IDs=[6180190,8419570,699736927],Sort=[{By=Points,Direction=ASC}],Limit=1,ServerKeys=[pl169]",
			params: domain.ListPlayersParams{
				IDs:        []int64{6180190, 8419570, 699736927},
				ServerKeys: []string{"pl169"},
				Sort:       []domain.PlayerSort{{By: domain.PlayerSortByPoints, Direction: domain.SortDirectionASC}},
				Pagination: domain.Pagination{
					Limit: 1,
				},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        699736927,
					serverKey: "pl169",
				},
			},
			expectedCount: 3,
		},
		{
			name: "IDs=[1111,1112],Deleted=true,Sort=[{By=DeletedAt,Direction=ASC}],Limit=1,ServerKeys=[en113]",
			params: domain.ListPlayersParams{
				IDs:        []int64{1111, 1112},
				ServerKeys: []string{"en113"},
				Deleted: domain.NullBool{
					Bool:  true,
					Valid: true,
				},
				Sort: []domain.PlayerSort{{By: domain.PlayerSortByDeletedAt, Direction: domain.SortDirectionASC}},
				Pagination: domain.Pagination{
					Limit: 1,
				},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        1112,
					serverKey: "en113",
				},
			},
			expectedCount: 2,
		},
		{
			name: "IDs=[1111,1112],Deleted=true,Sort=[{By=DeletedAt,Direction=DESC}],Limit=1,ServerKeys=[en113]",
			params: domain.ListPlayersParams{
				IDs:        []int64{1111, 1112},
				ServerKeys: []string{"en113"},
				Deleted: domain.NullBool{
					Bool:  true,
					Valid: true,
				},
				Sort: []domain.PlayerSort{{By: domain.PlayerSortByDeletedAt, Direction: domain.SortDirectionDESC}},
				Pagination: domain.Pagination{
					Limit: 1,
				},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        1111,
					serverKey: "en113",
				},
			},
			expectedCount: 2,
		},
		{
			name: "ServerKey=[pl169],TribeID=28,Sort=[{By=ID,Direction=ASC}]",
			params: domain.ListPlayersParams{
				ServerKeys: []string{"pl169"},
				TribeID: domain.NullInt64{
					Valid: true,
					Int64: tribeCSA.ID,
				},
				Sort: []domain.PlayerSort{{By: domain.PlayerSortByID, Direction: domain.SortDirectionASC}},
			},
			expectedPlayers: playersPL169CSA,
			expectedCount:   int64(len(playersPL169CSA)),
		},
		{
			name: "VersionCodes=[pl],Sort=[{By=ID,Direction=ASC}]",
			params: domain.ListPlayersParams{
				VersionCodes: []string{"pl"},
				Sort:         []domain.PlayerSort{{By: domain.PlayerSortByID, Direction: domain.SortDirectionASC}},
			},
			expectedPlayers: playersPL,
			expectedCount:   int64(len(playersPL)),
		},
		{
			name: "ServerKeysNIN=[pl169],Sort=[{By=ID,Direction=ASC}]",
			params: domain.ListPlayersParams{
				ServerKeysNIN: []string{"pl169"},
				Sort:          []domain.PlayerSort{{By: domain.PlayerSortByID, Direction: domain.SortDirectionASC}},
			},
			expectedPlayers: playersExceptPL169,
			expectedCount:   int64(len(playersExceptPL169)),
		},
		{
			name: "Names=[Zero-Two,makobex],Sort=[{By=ID,Direction=ASC}],ServerKeys=[en113]",
			params: domain.ListPlayersParams{
				Names:      []string{"Zero-Two", "makobex"},
				ServerKeys: []string{"en113"},
				Sort:       []domain.PlayerSort{{By: domain.PlayerSortByID, Direction: domain.SortDirectionASC}},
			},
			expectedPlayers: []expectedPlayer{
				{
					id:        8625053,
					serverKey: "en113",
				},
				{
					id:        8625053,
					serverKey: "en113",
				},
			},
			expectedCount: 2,
		},
		{
			name: "ERR: unsupported sort by",
			params: domain.ListPlayersParams{
				Sort: []domain.PlayerSort{
					{By: 100, Direction: domain.SortDirectionDESC},
				},
			},
			expectedPlayers: nil,
			expectedCount:   0,
			expectedErr:     domain.ErrUnsupportedSortBy,
		},
		{
			name: "ERR: unsupported sort direction",
			params: domain.ListPlayersParams{
				Sort: []domain.PlayerSort{
					{By: domain.PlayerSortByScoreAtt, Direction: 100},
				},
			},
			expectedPlayers: nil,
			expectedCount:   0,
			expectedErr:     domain.ErrUnsupportedSortDirection,
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			resListCountWithRelations, count, err := repo.ListCountWithRelations(context.Background(), tt.params)
			assert.ErrorIs(t, err, tt.expectedErr)
			assert.Equal(t, tt.expectedCount, count)
			assert.Len(t, resListCountWithRelations, len(tt.expectedPlayers))
			for _, expPlayer := range tt.expectedPlayers {
				found := false

				for _, player := range resListCountWithRelations {
					if player.ID != expPlayer.id || player.ServerKey != expPlayer.serverKey {
						continue
					}

					if player.TribeID != 0 && (!player.Tribe.Valid || player.Tribe.Tribe.ID != player.TribeID) {
						continue
					}

					if player.TribeID == 0 && player.Tribe.Valid {
						continue
					}

					if player.ServerKey != player.Server.Key {
						continue
					}

					found = true
					break
				}

				assert.True(t, found, "player (id=%d,serverkey=%s) not found",
					expPlayer.id,
					expPlayer.serverKey,
				)
			}

			resList, err := repo.List(context.Background(), tt.params)
			assert.ErrorIs(t, err, tt.expectedErr)
			assert.Len(t, resList, len(resListCountWithRelations))
			for i, player := range resList {
				assert.Equal(t, resListCountWithRelations[i].Player, player)
			}
		})
	}
}

func assertCreatedPlayers(tb testing.TB, params []domain.CreatePlayerParams, players []domain.Player) {
	tb.Helper()

	assert.Len(tb, players, len(params))
	for _, p := range params {
		var player domain.Player
		for _, pl := range players {
			if pl.ID == p.ID && pl.ServerKey == p.ServerKey {
				player = pl
				break
			}
		}
		assert.EqualValues(tb, p.OpponentsDefeated, player.OpponentsDefeated)
		assert.Equal(tb, p.ID, player.ID)
		assert.Equal(tb, p.Name, player.Name)
		assert.Equal(tb, p.NumVillages, player.NumVillages)
		assert.Equal(tb, p.Points, player.Points)
		assert.Equal(tb, p.Rank, player.Rank)
		assert.Equal(tb, p.TribeID, player.TribeID)
		assert.Equal(tb, p.ProfileURL, player.ProfileURL)
		assert.Equal(tb, p.BestRank, player.BestRank)
		assert.WithinDuration(tb, p.BestRankAt, player.BestRankAt, time.Second)
		assert.Equal(tb, p.MostPoints, player.MostPoints)
		assert.WithinDuration(tb, p.MostPointsAt, player.MostPointsAt, time.Second)
		assert.Equal(tb, p.MostVillages, player.MostVillages)
		assert.WithinDuration(tb, p.MostVillagesAt, player.MostVillagesAt, time.Second)
		assert.WithinDuration(tb, p.LastActivityAt, player.LastActivityAt, time.Second)
		assert.Equal(tb, p.ServerKey, player.ServerKey)
		assert.Zero(tb, player.DeletedAt)
		assert.WithinDuration(tb, player.CreatedAt, time.Now(), 1*time.Second)
	}
}
