package adapter

import (
	"context"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"gitea.dwysokinski.me/twhelp/core/internal/tw"
)

type TWHTTPClient struct {
	client *tw.Client
}

func NewTWHTTPClient(client *tw.Client) *TWHTTPClient {
	return &TWHTTPClient{client: client}
}

func (t *TWHTTPClient) GetOpenServers(ctx context.Context, baseURL string) ([]domain.OpenServer, error) {
	servers, err := t.client.GetOpenServers(ctx, baseURL)
	if err != nil {
		return nil, err
	}

	res := make([]domain.OpenServer, 0, len(servers))
	for _, s := range servers {
		res = append(res, domain.OpenServer(s))
	}

	return res, nil
}

func (t *TWHTTPClient) GetServerConfig(ctx context.Context, baseURL string) (domain.ServerConfig, error) {
	cfg, err := t.client.GetServerConfig(ctx, baseURL)
	if err != nil {
		return domain.ServerConfig{}, err
	}
	return t.convertServerConfigToDomain(cfg), nil
}

func (t *TWHTTPClient) GetBuildingInfo(ctx context.Context, baseURL string) (domain.BuildingInfo, error) {
	cfg, err := t.client.GetBuildingInfo(ctx, baseURL)
	if err != nil {
		return domain.BuildingInfo{}, err
	}
	return t.convertBuildingInfoToDomain(cfg), nil
}

func (t *TWHTTPClient) GetUnitInfo(ctx context.Context, baseURL string) (domain.UnitInfo, error) {
	cfg, err := t.client.GetUnitInfo(ctx, baseURL)
	if err != nil {
		return domain.UnitInfo{}, err
	}
	return t.convertUnitInfoToDomain(cfg), nil
}

func (t *TWHTTPClient) GetTribes(ctx context.Context, baseURL string) ([]domain.BaseTribe, error) {
	tribes, err := t.client.GetTribes(ctx, baseURL)
	if err != nil {
		return nil, err
	}

	res := make([]domain.BaseTribe, 0, len(tribes))
	for _, tr := range tribes {
		res = append(res, t.convertTribeToDomain(tr))
	}

	return res, nil
}

func (t *TWHTTPClient) GetPlayers(ctx context.Context, baseURL string) ([]domain.BasePlayer, error) {
	players, err := t.client.GetPlayers(ctx, baseURL)
	if err != nil {
		return nil, err
	}

	res := make([]domain.BasePlayer, 0, len(players))
	for _, p := range players {
		res = append(res, t.convertPlayerToDomain(p))
	}

	return res, nil
}

func (t *TWHTTPClient) GetVillages(ctx context.Context, baseURL string) ([]domain.BaseVillage, error) {
	villages, err := t.client.GetVillages(ctx, baseURL)
	if err != nil {
		return nil, err
	}

	res := make([]domain.BaseVillage, 0, len(villages))
	for _, v := range villages {
		res = append(res, t.convertVillageToDomain(v))
	}

	return res, nil
}

func (t *TWHTTPClient) GetEnnoblements(ctx context.Context, baseURL string, since time.Time) ([]domain.BaseEnnoblement, error) {
	ennoblements, err := t.client.GetEnnoblements(ctx, baseURL, since)
	if err != nil {
		return nil, err
	}

	res := make([]domain.BaseEnnoblement, 0, len(ennoblements))
	for _, e := range ennoblements {
		res = append(res, t.convertEnnoblementToDomain(e))
	}

	return res, nil
}

func (t *TWHTTPClient) convertServerConfigToDomain(cfg tw.ServerConfig) domain.ServerConfig {
	return domain.ServerConfig{
		Speed:     cfg.Speed,
		UnitSpeed: cfg.UnitSpeed,
		Moral:     cfg.Moral,
		Build:     domain.ServerConfigBuild(cfg.Build),
		Misc:      domain.ServerConfigMisc(cfg.Misc),
		Commands:  domain.ServerConfigCommands(cfg.Commands),
		Newbie:    domain.ServerConfigNewbie(cfg.Newbie),
		Game:      t.convertServerConfigGameToDomain(cfg.Game),
		Buildings: domain.ServerConfigBuildings(cfg.Buildings),
		Snob:      domain.ServerConfigSnob(cfg.Snob),
		Ally:      domain.ServerConfigAlly(cfg.Ally),
		Coord:     domain.ServerConfigCoord(cfg.Coord),
		Sitter:    domain.ServerConfigSitter(cfg.Sitter),
		Sleep:     domain.ServerConfigSleep(cfg.Sleep),
		Night:     domain.ServerConfigNight(cfg.Night),
		Win:       domain.ServerConfigWin(cfg.Win),
	}
}

func (t *TWHTTPClient) convertServerConfigGameToDomain(cfg tw.ServerConfigGame) domain.ServerConfigGame {
	return domain.ServerConfigGame{
		BuildtimeFormula:   cfg.BuildtimeFormula,
		Knight:             cfg.Knight,
		KnightNewItems:     cfg.KnightNewItems.Int8(),
		Archer:             cfg.Archer,
		Tech:               cfg.Tech,
		FarmLimit:          cfg.FarmLimit,
		Church:             cfg.Church,
		Watchtower:         cfg.Watchtower,
		Stronghold:         cfg.Stronghold,
		FakeLimit:          cfg.FakeLimit,
		BarbarianRise:      cfg.BarbarianRise,
		BarbarianShrink:    cfg.BarbarianShrink,
		BarbarianMaxPoints: cfg.BarbarianMaxPoints,
		Scavenging:         cfg.Scavenging,
		Hauls:              cfg.Hauls,
		HaulsBase:          cfg.HaulsBase,
		HaulsMax:           cfg.HaulsMax,
		BaseProduction:     cfg.BaseProduction,
		Event:              cfg.Event,
		SuppressEvents:     cfg.SuppressEvents,
	}
}

func (t *TWHTTPClient) convertUnitInfoToDomain(info tw.UnitInfo) domain.UnitInfo {
	return domain.UnitInfo{
		Spear:    domain.Unit(info.Spear),
		Sword:    domain.Unit(info.Sword),
		Axe:      domain.Unit(info.Axe),
		Archer:   domain.Unit(info.Archer),
		Spy:      domain.Unit(info.Spy),
		Light:    domain.Unit(info.Light),
		Marcher:  domain.Unit(info.Marcher),
		Heavy:    domain.Unit(info.Heavy),
		Ram:      domain.Unit(info.Ram),
		Catapult: domain.Unit(info.Catapult),
		Knight:   domain.Unit(info.Knight),
		Snob:     domain.Unit(info.Snob),
		Militia:  domain.Unit(info.Militia),
	}
}

func (t *TWHTTPClient) convertBuildingInfoToDomain(info tw.BuildingInfo) domain.BuildingInfo {
	return domain.BuildingInfo{
		Main:       domain.Building(info.Main),
		Barracks:   domain.Building(info.Barracks),
		Stable:     domain.Building(info.Stable),
		Garage:     domain.Building(info.Garage),
		Watchtower: domain.Building(info.Watchtower),
		Snob:       domain.Building(info.Snob),
		Smith:      domain.Building(info.Smith),
		Place:      domain.Building(info.Place),
		Statue:     domain.Building(info.Statue),
		Market:     domain.Building(info.Market),
		Wood:       domain.Building(info.Wood),
		Stone:      domain.Building(info.Stone),
		Iron:       domain.Building(info.Iron),
		Farm:       domain.Building(info.Farm),
		Storage:    domain.Building(info.Storage),
		Hide:       domain.Building(info.Hide),
		Wall:       domain.Building(info.Wall),
	}
}

func (t *TWHTTPClient) convertOpponentsDefeatedToDomain(od tw.OpponentsDefeated) domain.OpponentsDefeated {
	return domain.OpponentsDefeated(od)
}

func (t *TWHTTPClient) convertTribeToDomain(tr tw.Tribe) domain.BaseTribe {
	return domain.BaseTribe{
		OpponentsDefeated: t.convertOpponentsDefeatedToDomain(tr.OpponentsDefeated),
		ID:                tr.ID,
		Name:              tr.Name,
		Tag:               tr.Tag,
		NumMembers:        tr.NumMembers,
		NumVillages:       tr.NumVillages,
		Points:            tr.Points,
		AllPoints:         tr.AllPoints,
		Rank:              tr.Rank,
		ProfileURL:        tr.ProfileURL,
	}
}

func (t *TWHTTPClient) convertPlayerToDomain(p tw.Player) domain.BasePlayer {
	return domain.BasePlayer{
		OpponentsDefeated: t.convertOpponentsDefeatedToDomain(p.OpponentsDefeated),
		ID:                p.ID,
		Name:              p.Name,
		NumVillages:       p.NumVillages,
		Points:            p.Points,
		Rank:              p.Rank,
		TribeID:           p.TribeID,
		ProfileURL:        p.ProfileURL,
	}
}

func (t *TWHTTPClient) convertVillageToDomain(v tw.Village) domain.BaseVillage {
	return domain.BaseVillage(v)
}

func (t *TWHTTPClient) convertEnnoblementToDomain(e tw.Ennoblement) domain.BaseEnnoblement {
	return domain.BaseEnnoblement(e)
}
