package adapter

import (
	"context"
	"database/sql"
	"errors"
	"fmt"

	"gitea.dwysokinski.me/twhelp/core/internal/adapter/internal/bunmodel"
	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"github.com/uptrace/bun"
)

type VersionBunRepository struct {
	db *bun.DB
}

func NewVersionBunRepository(db *bun.DB) *VersionBunRepository {
	return &VersionBunRepository{db: db}
}

func (v *VersionBunRepository) List(ctx context.Context) ([]domain.Version, error) {
	var versions []bunmodel.Version

	if err := v.buildListQuery(&versions).Scan(ctx); err != nil && !errors.Is(err, sql.ErrNoRows) {
		return nil, fmt.Errorf("couldn't select versions from the db: %w", err)
	}

	result := make([]domain.Version, 0, len(versions))
	for _, version := range versions {
		result = append(result, version.ToDomain())
	}

	return result, nil
}

func (v *VersionBunRepository) ListCount(ctx context.Context) ([]domain.Version, int64, error) {
	var versions []bunmodel.Version

	count, err := v.buildListQuery(&versions).ScanAndCount(ctx)
	if err != nil && !errors.Is(err, sql.ErrNoRows) {
		return nil, 0, fmt.Errorf("couldn't select versions from the db: %w", err)
	}

	result := make([]domain.Version, 0, len(versions))
	for _, version := range versions {
		result = append(result, version.ToDomain())
	}

	return result, int64(count), nil
}

func (v *VersionBunRepository) buildListQuery(versions *[]bunmodel.Version) *bun.SelectQuery {
	return v.db.NewSelect().
		Model(versions).
		Order("code ASC")
}

func (v *VersionBunRepository) GetByCode(ctx context.Context, code string) (domain.Version, error) {
	var version bunmodel.Version

	if err := v.db.NewSelect().
		Model(&version).
		Where("code = ?", code).
		Scan(ctx); err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			return domain.Version{}, domain.VersionNotFoundError{VerCode: code}
		}
		return domain.Version{}, fmt.Errorf(
			"something went wrong while selecting version (code=%s) from the db: %w",
			code,
			err,
		)
	}

	return version.ToDomain(), nil
}
