package domain

import (
	"fmt"
	"time"
)

type BasePlayer struct {
	OpponentsDefeated

	ID          int64
	Name        string
	NumVillages int64
	Points      int64
	Rank        int64
	TribeID     int64
	ProfileURL  string
}

type Player struct {
	BasePlayer

	BestRank       int64
	BestRankAt     time.Time
	MostPoints     int64
	MostPointsAt   time.Time
	MostVillages   int64
	MostVillagesAt time.Time
	LastActivityAt time.Time
	ServerKey      string
	CreatedAt      time.Time
	DeletedAt      time.Time
}

type PlayerWithRelations struct {
	Player

	Tribe  NullTribeMeta
	Server ServerMeta
}

type PlayerMeta struct {
	ID         int64
	Name       string
	ProfileURL string
}

type NullPlayerMeta struct {
	Valid  bool
	Player PlayerMeta
}

type PlayerMetaWithRelations struct {
	PlayerMeta

	Tribe NullTribeMeta
}

type NullPlayerMetaWithRelations struct {
	Valid  bool
	Player PlayerMetaWithRelations
}

type CreatePlayerParams struct {
	BasePlayer

	BestRank       int64
	BestRankAt     time.Time
	MostPoints     int64
	MostPointsAt   time.Time
	MostVillages   int64
	MostVillagesAt time.Time
	LastActivityAt time.Time
	ServerKey      string
}

type PlayerSortBy uint8

const (
	PlayerSortByID PlayerSortBy = iota
	PlayerSortByScoreAtt
	PlayerSortByScoreDef
	PlayerSortByScoreSup
	PlayerSortByScoreTotal
	PlayerSortByPoints
	PlayerSortByDeletedAt
)

func newPlayerSortBy(s string) (PlayerSortBy, error) {
	switch s {
	case "id":
		return PlayerSortByID, nil
	case "scoreAtt":
		return PlayerSortByScoreAtt, nil
	case "scoreDef":
		return PlayerSortByScoreDef, nil
	case "scoreSup":
		return PlayerSortByScoreSup, nil
	case "scoreTotal":
		return PlayerSortByScoreTotal, nil
	case "points":
		return PlayerSortByPoints, nil
	case "deletedAt":
		return PlayerSortByDeletedAt, nil
	}
	return 0, fmt.Errorf("%w: \"%s\"", ErrUnsupportedSortBy, s)
}

type PlayerSort struct {
	By        PlayerSortBy
	Direction SortDirection
}

func NewPlayerSort(by, dir string) (PlayerSort, error) {
	convBy, err := newPlayerSortBy(by)
	if err != nil {
		return PlayerSort{}, err
	}

	convDir, err := newSortDirection(dir)
	if err != nil {
		return PlayerSort{}, err
	}

	return PlayerSort{
		By:        convBy,
		Direction: convDir,
	}, nil
}

type ListPlayersParams struct {
	IDs           []int64
	IDGT          NullInt64
	ServerKeys    []string
	ServerKeysNIN []string
	VersionCodes  []string
	Names         []string
	Deleted       NullBool
	TribeID       NullInt64
	Pagination    Pagination
	Sort          []PlayerSort
}

type PlayerNotFoundError struct {
	ID int64
}

func (e PlayerNotFoundError) Error() string {
	return fmt.Sprintf("player (id=%d) not found", e.ID)
}

func (e PlayerNotFoundError) UserError() string {
	return e.Error()
}

func (e PlayerNotFoundError) Code() ErrorCode {
	return ErrorCodeEntityNotFound
}
