package domain

import (
	"fmt"
	"time"
)

type BaseTribe struct {
	OpponentsDefeated

	ID          int64
	Name        string
	Tag         string
	NumMembers  int64
	NumVillages int64
	Points      int64
	AllPoints   int64
	Rank        int64
	ProfileURL  string
}

type Tribe struct {
	BaseTribe

	Dominance      float64
	BestRank       int64
	BestRankAt     time.Time
	MostPoints     int64
	MostPointsAt   time.Time
	MostVillages   int64
	MostVillagesAt time.Time
	ServerKey      string
	CreatedAt      time.Time
	DeletedAt      time.Time
}

type TribeMeta struct {
	ID         int64
	Name       string
	Tag        string
	ProfileURL string
}

type NullTribeMeta struct {
	Valid bool
	Tribe TribeMeta
}

type CreateTribeParams struct {
	BaseTribe

	BestRank       int64
	BestRankAt     time.Time
	MostPoints     int64
	MostPointsAt   time.Time
	MostVillages   int64
	MostVillagesAt time.Time
	ServerKey      string
}

type TribeSortBy uint8

const (
	TribeSortByID TribeSortBy = iota
	TribeSortByScoreAtt
	TribeSortByScoreDef
	TribeSortByScoreTotal
	TribeSortByPoints
	TribeSortByDominance
	TribeSortByDeletedAt
)

func newTribeSortBy(s string) (TribeSortBy, error) {
	switch s {
	case "id":
		return TribeSortByID, nil
	case "scoreAtt":
		return TribeSortByScoreAtt, nil
	case "scoreDef":
		return TribeSortByScoreDef, nil
	case "scoreTotal":
		return TribeSortByScoreTotal, nil
	case "points":
		return TribeSortByPoints, nil
	case "dominance":
		return TribeSortByDominance, nil
	case "deletedAt":
		return TribeSortByDeletedAt, nil
	}
	return 0, fmt.Errorf("%w: \"%s\"", ErrUnsupportedSortBy, s)
}

type TribeSort struct {
	By        TribeSortBy
	Direction SortDirection
}

func NewTribeSort(by, dir string) (TribeSort, error) {
	convBy, err := newTribeSortBy(by)
	if err != nil {
		return TribeSort{}, err
	}

	convDir, err := newSortDirection(dir)
	if err != nil {
		return TribeSort{}, err
	}

	return TribeSort{
		By:        convBy,
		Direction: convDir,
	}, nil
}

type ListTribesParams struct {
	IDs        []int64
	IDGT       NullInt64
	Tags       []string
	ServerKeys []string
	Deleted    NullBool
	Pagination Pagination
	Sort       []TribeSort
}

type TribeNotFoundError struct {
	ID int64
}

func (e TribeNotFoundError) Error() string {
	return fmt.Sprintf("tribe (id=%d) not found", e.ID)
}

func (e TribeNotFoundError) UserError() string {
	return e.Error()
}

func (e TribeNotFoundError) Code() ErrorCode {
	return ErrorCodeEntityNotFound
}
