package msg_test

import (
	"context"
	"flag"
	"fmt"
	"log"
	"net"
	urlpkg "net/url"
	"os"
	"testing"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/adapter/bundbtest"
	"github.com/ThreeDotsLabs/watermill"
	"github.com/ThreeDotsLabs/watermill-amqp/v2/pkg/amqp"
	"github.com/ThreeDotsLabs/watermill/message"
	"github.com/ThreeDotsLabs/watermill/pubsub/gochannel"
	"github.com/cenkalti/backoff/v4"
	"github.com/ory/dockertest/v3"
	"github.com/ory/dockertest/v3/docker"
	"github.com/pressly/goose/v3"
	"github.com/stretchr/testify/require"
	"github.com/uptrace/bun"
)

var (
	dbDSN       = os.Getenv("TESTS_DB_DSN")
	rabbitMQURI = os.Getenv("TESTS_AMQP_URI")
)

func TestMain(m *testing.M) {
	setUpGoose()

	// https://github.com/golang/go/blob/7cfa7d69259590319524c3715df4a39b39924bc3/src/testing/testing.go#L224
	flag.Parse()

	if (dbDSN != "" && rabbitMQURI != "") || testing.Short() {
		os.Exit(m.Run())
	}

	pool, err := dockertest.NewPool("")
	if err != nil {
		log.Fatalf("couldn't construct pool: %s", err)
	}

	pool.MaxWait = 30 * time.Second
	if err = pool.Client.Ping(); err != nil {
		log.Fatalf("couldn't connect to Docker: %s", err)
	}

	var resources []*dockertest.Resource

	if dbDSN == "" {
		resource, dsn := newPostgres(pool)
		dbDSN = dsn.String()
		resources = append(resources, resource)
	}

	if rabbitMQURI == "" {
		resource, uri := newRabbitMQ(pool)
		rabbitMQURI = uri.String()
		resources = append(resources, resource)
	}

	code := m.Run()

	for _, r := range resources {
		_ = pool.Purge(r)
	}

	os.Exit(code)
}

func setUpGoose() {
	goose.SetBaseFS(os.DirFS("../../migrations"))
	goose.SetLogger(goose.NopLogger())
	if err := goose.SetDialect("postgres"); err != nil {
		log.Fatal(err)
	}
}

func newPostgres(pool *dockertest.Pool) (*dockertest.Resource, *urlpkg.URL) {
	resource, dsn, err := bundbtest.NewPostgres(pool, 240)
	if err != nil {
		log.Fatal(err)
	}
	return resource, dsn
}

func newRabbitMQ(pool *dockertest.Pool) (*dockertest.Resource, *urlpkg.URL) {
	uri := &urlpkg.URL{
		Scheme: "amqp",
		User:   urlpkg.UserPassword("rmq", "rmq"),
		Path:   "/",
	}

	pw, _ := uri.User.Password()

	resource, err := pool.RunWithOptions(&dockertest.RunOptions{
		Repository: "rabbitmq",
		Tag:        "3.12.7-alpine",
		Env: []string{
			fmt.Sprintf("RABBITMQ_DEFAULT_USER=%s", uri.User.Username()),
			fmt.Sprintf("RABBITMQ_DEFAULT_PASS=%s", pw),
		},
	}, func(config *docker.HostConfig) {
		config.AutoRemove = true
		config.RestartPolicy = docker.RestartPolicy{
			Name: "no",
		}
	})
	if err != nil {
		log.Fatalf("couldn't start rabbitmq: %s", err)
	}

	if err = resource.Expire(180); err != nil {
		log.Fatal(err)
	}

	uri.Host, err = getHostPort(resource, "5672/tcp")
	if err != nil {
		log.Fatal(err)
	}

	return resource, uri
}

func getHostPort(resource *dockertest.Resource, id string) (string, error) {
	dockerURL := os.Getenv("DOCKER_HOST")
	if dockerURL == "" {
		return resource.GetHostPort(id), nil
	}

	u, err := urlpkg.Parse(dockerURL)
	if err != nil {
		return "", err
	}

	return net.JoinHostPort(u.Hostname(), resource.GetPort(id)), nil
}

func newBunDB(tb testing.TB) *bun.DB {
	tb.Helper()
	return bundbtest.NewDB(tb, dbDSN)
}

func newRabbitMQConnection(tb testing.TB) *amqp.ConnectionWrapper {
	tb.Helper()

	bo := backoff.NewExponentialBackOff()
	bo.MaxInterval = 5 * time.Second
	bo.MaxElapsedTime = 30 * time.Second
	require.NoError(tb, backoff.Retry(func() error {
		conn, err := amqp.NewConnection(amqp.ConnectionConfig{AmqpURI: rabbitMQURI}, watermill.NopLogger{})
		if err != nil {
			return err
		}
		_ = conn.Close()
		return nil
	}, bo))

	conn, err := amqp.NewConnection(amqp.ConnectionConfig{AmqpURI: rabbitMQURI}, watermill.NopLogger{})
	require.NoError(tb, err)
	tb.Cleanup(func() {
		_ = conn.Close()
	})

	return conn
}

func newRabbitMQPubSub(
	tb testing.TB,
	conn *amqp.ConnectionWrapper,
	queueNameGenerator amqp.QueueNameGenerator,
	exchangeAndRoutingKeyNameGenerator func(topic string) string,
) (*amqp.Publisher, *amqp.Subscriber) {
	tb.Helper()

	cfg := amqp.NewNonDurablePubSubConfig("", queueNameGenerator)
	cfg.Consume.NoRequeueOnNack = true
	cfg.Consume.Qos.PrefetchCount = 1
	cfg.Exchange.GenerateName = exchangeAndRoutingKeyNameGenerator
	cfg.QueueBind.GenerateRoutingKey = exchangeAndRoutingKeyNameGenerator

	publisher, err := amqp.NewPublisherWithConnection(cfg, watermill.NopLogger{}, conn)
	require.NoError(tb, err)
	tb.Cleanup(func() {
		_ = publisher.Close()
	})

	subscriber, err := amqp.NewSubscriberWithConnection(cfg, watermill.NopLogger{}, conn)
	require.NoError(tb, err)
	tb.Cleanup(func() {
		_ = subscriber.Close()
	})

	return publisher, subscriber
}

type registerer interface {
	Register(router *message.Router)
}

func runRouter(tb testing.TB, rs ...registerer) *message.Router {
	tb.Helper()
	return runRouterWithContext(tb, context.Background(), rs...)
}

func runRouterWithContext(tb testing.TB, ctx context.Context, rs ...registerer) *message.Router {
	tb.Helper()

	router, err := message.NewRouter(message.RouterConfig{CloseTimeout: 10 * time.Second}, watermill.NopLogger{})
	require.NoError(tb, err)
	tb.Cleanup(func() {
		_ = router.Close()
	})

	for _, r := range rs {
		r.Register(router)
	}

	go func() {
		require.NoError(tb, router.Run(ctx))
	}()
	<-router.Running()

	return router
}

type middlewareRegisterer struct {
	h message.HandlerMiddleware
}

var _ registerer = middlewareRegisterer{}

func (m middlewareRegisterer) Register(router *message.Router) {
	router.AddMiddleware(m.h)
}

func newWaitForHandlerMiddleware(handlerName string) (middlewareRegisterer, <-chan struct{}) {
	ch := make(chan struct{})
	return middlewareRegisterer{
		h: func(h message.HandlerFunc) message.HandlerFunc {
			return func(msg *message.Message) ([]*message.Message, error) {
				if message.HandlerNameFromCtx(msg.Context()) != handlerName {
					return h(msg)
				}

				result, err := h(msg)
				if len(result) == 0 && err == nil {
					close(ch)
				}
				return result, err
			}
		},
	}, ch
}

func newPubSub(tb testing.TB) *gochannel.GoChannel {
	tb.Helper()

	pubSub := gochannel.NewGoChannel(gochannel.Config{Persistent: true}, watermill.NopLogger{})
	tb.Cleanup(func() {
		_ = pubSub.Close()
	})

	return pubSub
}
