package msg

import (
	"context"
	"fmt"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/msg/internal/model"
	"github.com/ThreeDotsLabs/watermill"

	"github.com/ThreeDotsLabs/watermill/message"
)

//counterfeiter:generate -o internal/mock/tribe_service.gen.go . TribeService
type TribeService interface {
	Refresh(ctx context.Context, key, url string) (int64, error)
	UpdateDominance(ctx context.Context, serverKey string, numPlayerVillages int64) error
}

//counterfeiter:generate -o internal/mock/tribe_snapshot_service.gen.go . TribeSnapshotService
type TribeSnapshotService interface {
	Create(ctx context.Context, key string, date time.Time) error
}

type TribeConsumer struct {
	marshaler   Marshaler
	publisher   message.Publisher
	subscriber  message.Subscriber
	logger      watermill.LoggerAdapter
	svc         TribeService
	snapshotSvc TribeSnapshotService
}

func NewTribeConsumer(
	marshaler Marshaler,
	publisher message.Publisher,
	subscriber message.Subscriber,
	logger watermill.LoggerAdapter,
	svc TribeService,
	snapshotSvc TribeSnapshotService,
) *TribeConsumer {
	return &TribeConsumer{
		marshaler:   marshaler,
		publisher:   publisher,
		subscriber:  subscriber,
		logger:      logger,
		svc:         svc,
		snapshotSvc: snapshotSvc,
	}
}

const (
	tribesEventRefreshed        = "tribes.event.refreshed"
	tribesCmdCreateSnapshots    = "tribes.cmd.create_snapshots"
	tribesEventSnapshotsCreated = "tribes.event.snapshots_created"
)

func (t *TribeConsumer) Register(router *message.Router) {
	router.AddHandler(
		"TribeConsumer.refresh",
		serversEventRefreshed,
		t.subscriber,
		tribesEventRefreshed,
		t.publisher,
		t.refresh,
	)

	router.AddNoPublisherHandler(
		"TribeConsumer.updateDominance",
		villagesEventRefreshed,
		t.subscriber,
		t.updateDominance,
	)

	router.AddHandler(
		"TribeConsumer.createSnapshots",
		tribesCmdCreateSnapshots,
		t.subscriber,
		tribesEventSnapshotsCreated,
		t.publisher,
		t.createSnapshots,
	)
}

func (t *TribeConsumer) refresh(msg *message.Message) ([]*message.Message, error) {
	var payload model.ServerRefreshedEvPayload
	if err := t.marshaler.Unmarshal(msg, &payload); err != nil {
		t.logger.Error("invalid payload, dropping message", err, watermill.LogFields{
			"handler_name": message.HandlerNameFromCtx(msg.Context()),
		})
		return nil, nil
	}

	if !payload.Open {
		t.logger.Debug("server is closed, dropping message", watermill.LogFields{
			"handler_name": message.HandlerNameFromCtx(msg.Context()),
			"server_key":   payload.Key,
			"version_code": payload.VersionCode,
		})
		return nil, nil
	}

	numTribes, err := t.svc.Refresh(msg.Context(), payload.Key, payload.URL)
	if err != nil {
		return nil, fmt.Errorf("TribeService.Refresh: %w", err)
	}

	ev, err := t.marshaler.Marshal(model.TribesRefreshedEvPayload{
		Key:         payload.Key,
		URL:         payload.URL,
		VersionCode: payload.VersionCode,
		NumTribes:   numTribes,
	})
	if err != nil {
		return nil, fmt.Errorf("marshaler.Marshal: %w", err)
	}

	return []*message.Message{ev}, nil
}

func (t *TribeConsumer) updateDominance(msg *message.Message) error {
	var payload model.VillagesRefreshedEvPayload
	if err := t.marshaler.Unmarshal(msg, &payload); err != nil {
		t.logger.Error("invalid payload, dropping message", err, watermill.LogFields{
			"handler_name": message.HandlerNameFromCtx(msg.Context()),
		})
		return nil
	}

	if err := t.svc.UpdateDominance(msg.Context(), payload.Key, payload.NumPlayerVillages); err != nil {
		return fmt.Errorf("TribeService.UpdateDominance: %w", err)
	}

	return nil
}

func (t *TribeConsumer) createSnapshots(msg *message.Message) ([]*message.Message, error) {
	var payload model.CreateSnapshotsCmdPayload
	if err := t.marshaler.Unmarshal(msg, &payload); err != nil {
		t.logger.Error("invalid payload, dropping message", err, watermill.LogFields{
			"handler_name": message.HandlerNameFromCtx(msg.Context()),
		})
		return nil, nil
	}

	if err := t.snapshotSvc.Create(msg.Context(), payload.Key, payload.Date); err != nil {
		return nil, fmt.Errorf("TribeSnapshotService.Create: %w", err)
	}

	ev, err := t.marshaler.Marshal(model.SnapshotsCreatedEvPayload{
		Key:         payload.Key,
		VersionCode: payload.VersionCode,
	})
	if err != nil {
		return nil, fmt.Errorf("marshaler.Marshal: %w", err)
	}

	return []*message.Message{ev}, nil
}
