package msg_test

import (
	"context"
	"testing"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"github.com/ThreeDotsLabs/watermill"
	"github.com/google/uuid"

	"gitea.dwysokinski.me/twhelp/core/internal/msg/internal/model"

	"github.com/ThreeDotsLabs/watermill/message/subscriber"

	"gitea.dwysokinski.me/twhelp/core/internal/msg"
	"gitea.dwysokinski.me/twhelp/core/internal/msg/internal/mock"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

func TestTribeConsumer_refresh(t *testing.T) {
	t.Parallel()

	t.Run("OK", func(t *testing.T) {
		t.Parallel()

		marshaler := msg.JSONMarshaler{NewUUID: uuid.NewString}
		pubSub := newPubSub(t)

		var numTribes int64 = 12344
		tribeSvc := &mock.FakeTribeService{}
		tribeSvc.RefreshReturns(numTribes, nil)

		runRouter(t, msg.NewTribeConsumer(
			marshaler,
			pubSub,
			pubSub,
			watermill.NopLogger{},
			tribeSvc,
			&mock.FakePlayerSnapshotService{},
		))

		msgs, err := pubSub.Subscribe(context.Background(), "tribes.event.refreshed")
		require.NoError(t, err)

		serverRefreshedPayload := model.ServerRefreshedEvPayload{
			Key:         "pl151",
			URL:         "https://pl151.plemiona.pl",
			Open:        true,
			VersionCode: "pl",
		}
		ev, err := marshaler.Marshal(serverRefreshedPayload)
		require.NoError(t, err)
		require.NoError(t, pubSub.Publish("servers.event.refreshed", ev))

		receivedMsgs, _ := subscriber.BulkRead(msgs, 1, 10*time.Second)
		require.Len(t, receivedMsgs, 1)
		var tribesRefreshedPayload model.TribesRefreshedEvPayload
		assert.NoError(t, marshaler.Unmarshal(receivedMsgs[0], &tribesRefreshedPayload))
		assert.Equal(t, serverRefreshedPayload.Key, tribesRefreshedPayload.Key)
		assert.Equal(t, serverRefreshedPayload.URL, tribesRefreshedPayload.URL)
		assert.Equal(t, serverRefreshedPayload.VersionCode, tribesRefreshedPayload.VersionCode)
		assert.Equal(t, numTribes, tribesRefreshedPayload.NumTribes)
	})

	t.Run("OK: only open servers should be updated", func(t *testing.T) {
		t.Parallel()

		ctx, cancel := context.WithTimeout(context.Background(), time.Second)
		defer cancel()

		marshaler := msg.JSONMarshaler{NewUUID: uuid.NewString}
		pubSub := newPubSub(t)
		tribeSvc := &mock.FakeTribeService{}

		waitMiddleware, wait := newWaitForHandlerMiddleware("TribeConsumer.refresh")
		runRouter(
			t,
			waitMiddleware,
			msg.NewTribeConsumer(
				marshaler,
				pubSub,
				pubSub,
				watermill.NopLogger{},
				tribeSvc,
				&mock.FakePlayerSnapshotService{},
			),
		)

		ev, err := marshaler.Marshal(model.ServerRefreshedEvPayload{Open: false})
		require.NoError(t, err)
		require.NoError(t, pubSub.Publish("servers.event.refreshed", ev))
		ev.SetContext(ctx)

		select {
		case <-wait:
		case <-ctx.Done():
			t.Fatal("timeout")
		}

		assert.Equal(t, 0, tribeSvc.RefreshCallCount())
	})
}

func TestTribeConsumer_updateDominance(t *testing.T) {
	t.Parallel()

	ctx, cancel := context.WithTimeout(context.Background(), time.Second)
	defer cancel()

	marshaler := msg.JSONMarshaler{NewUUID: uuid.NewString}
	pubSub := newPubSub(t)

	tribeSvc := &mock.FakeTribeService{}
	tribeSvc.UpdateDominanceReturns(nil)

	waitMiddleware, wait := newWaitForHandlerMiddleware("TribeConsumer.updateDominance")
	runRouter(
		t,
		waitMiddleware,
		msg.NewTribeConsumer(
			marshaler,
			pubSub,
			pubSub,
			watermill.NopLogger{},
			tribeSvc,
			&mock.FakePlayerSnapshotService{},
		),
	)

	payload := model.VillagesRefreshedEvPayload{
		Key:                  "pl151",
		URL:                  "https://pl151.plemiona.pl",
		VersionCode:          "pl",
		NumVillages:          1500,
		NumPlayerVillages:    1400,
		NumBarbarianVillages: 100,
		NumBonusVillages:     300,
	}
	ev, err := marshaler.Marshal(payload)
	require.NoError(t, err)
	require.NoError(t, pubSub.Publish("villages.event.refreshed", ev))
	ev.SetContext(ctx)

	select {
	case <-wait:
	case <-ctx.Done():
		t.Fatal("timeout")
	}

	require.Equal(t, 1, tribeSvc.UpdateDominanceCallCount())
	_, serverKey, numPlayerVillages := tribeSvc.UpdateDominanceArgsForCall(0)
	assert.Equal(t, payload.Key, serverKey)
	assert.Equal(t, payload.NumPlayerVillages, numPlayerVillages)
}

func TestTribeConsumer_createSnapshots(t *testing.T) {
	t.Parallel()

	marshaler := msg.JSONMarshaler{NewUUID: uuid.NewString}
	pubSub := newPubSub(t)

	runRouter(t, msg.NewTribeConsumer(
		marshaler,
		pubSub,
		pubSub,
		watermill.NopLogger{},
		&mock.FakeTribeService{},
		&mock.FakeTribeSnapshotService{},
	))

	msgs, err := pubSub.Subscribe(context.Background(), "tribes.event.snapshots_created")
	require.NoError(t, err)

	createSnapshotsPayload := domain.CreateSnapshotsCmdPayload{
		Key:         "pl151",
		VersionCode: "pl",
		Date:        time.Now(),
	}
	cmd, err := marshaler.Marshal(createSnapshotsPayload)
	require.NoError(t, err)
	require.NoError(t, pubSub.Publish("tribes.cmd.create_snapshots", cmd))

	receivedMsgs, _ := subscriber.BulkRead(msgs, 1, 10*time.Second)
	require.Len(t, receivedMsgs, 1)
	var snapshotsCreatedPayload model.SnapshotsCreatedEvPayload
	assert.NoError(t, marshaler.Unmarshal(receivedMsgs[0], &snapshotsCreatedPayload))
	assert.Equal(t, createSnapshotsPayload.Key, snapshotsCreatedPayload.Key)
	assert.Equal(t, createSnapshotsPayload.VersionCode, snapshotsCreatedPayload.VersionCode)
}
