package msg

import (
	"context"
	"fmt"

	"gitea.dwysokinski.me/twhelp/core/internal/msg/internal/model"
	"github.com/ThreeDotsLabs/watermill"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"

	"github.com/ThreeDotsLabs/watermill/message"
)

//counterfeiter:generate -o internal/mock/village_service.gen.go . VillageService
type VillageService interface {
	Refresh(ctx context.Context, key, url string) (domain.RefreshVillagesResult, error)
}

type VillageConsumer struct {
	marshaler  Marshaler
	publisher  message.Publisher
	subscriber message.Subscriber
	logger     watermill.LoggerAdapter
	svc        VillageService
}

func NewVillageConsumer(
	marshaler Marshaler,
	publisher message.Publisher,
	subscriber message.Subscriber,
	logger watermill.LoggerAdapter,
	svc VillageService,
) *VillageConsumer {
	return &VillageConsumer{
		marshaler:  marshaler,
		publisher:  publisher,
		subscriber: subscriber,
		logger:     logger,
		svc:        svc,
	}
}

const villagesEventRefreshed = "villages.event.refreshed"

func (v *VillageConsumer) Register(router *message.Router) {
	router.AddHandler(
		"VillageConsumer.refresh",
		serversEventRefreshed,
		v.subscriber,
		villagesEventRefreshed,
		v.publisher,
		v.refresh,
	)
}

func (v *VillageConsumer) refresh(msg *message.Message) ([]*message.Message, error) {
	var payload model.ServerRefreshedEvPayload
	if err := v.marshaler.Unmarshal(msg, &payload); err != nil {
		v.logger.Error("invalid payload, dropping message", err, watermill.LogFields{
			"handler_name": message.HandlerNameFromCtx(msg.Context()),
		})
		return nil, nil
	}

	if !payload.Open {
		v.logger.Debug("server is closed, dropping message", watermill.LogFields{
			"handler_name": message.HandlerNameFromCtx(msg.Context()),
			"server_key":   payload.Key,
			"version_code": payload.VersionCode,
		})
		return nil, nil
	}

	res, err := v.svc.Refresh(msg.Context(), payload.Key, payload.URL)
	if err != nil {
		return nil, fmt.Errorf("VillageService.Refresh: %w", err)
	}

	ev, err := v.marshaler.Marshal(model.VillagesRefreshedEvPayload{
		Key:                  payload.Key,
		URL:                  payload.URL,
		VersionCode:          payload.VersionCode,
		NumVillages:          res.NumVillages,
		NumPlayerVillages:    res.NumPlayerVillages,
		NumBarbarianVillages: res.NumBarbarianVillages,
		NumBonusVillages:     res.NumBonusVillages,
	})
	if err != nil {
		return nil, fmt.Errorf("marshaler.Marshal: %w", err)
	}

	return []*message.Message{ev}, nil
}
