package msg_test

import (
	"context"
	"testing"
	"time"

	"github.com/ThreeDotsLabs/watermill"
	"github.com/google/uuid"

	"gitea.dwysokinski.me/twhelp/core/internal/msg/internal/model"

	"github.com/ThreeDotsLabs/watermill/message/subscriber"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"gitea.dwysokinski.me/twhelp/core/internal/msg"
	"gitea.dwysokinski.me/twhelp/core/internal/msg/internal/mock"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

func TestVillageConsumer_refresh(t *testing.T) {
	t.Parallel()

	t.Run("OK", func(t *testing.T) {
		t.Parallel()

		marshaler := msg.JSONMarshaler{NewUUID: uuid.NewString}
		pubSub := newPubSub(t)

		refreshResult := domain.RefreshVillagesResult{
			NumVillages:          1500,
			NumPlayerVillages:    1400,
			NumBarbarianVillages: 100,
			NumBonusVillages:     50,
		}
		villageSvc := &mock.FakeVillageService{}
		villageSvc.RefreshReturns(refreshResult, nil)

		runRouter(t, msg.NewVillageConsumer(marshaler, pubSub, pubSub, watermill.NopLogger{}, villageSvc))

		msgs, err := pubSub.Subscribe(context.Background(), "villages.event.refreshed")
		require.NoError(t, err)

		serverRefreshedPayload := model.ServerRefreshedEvPayload{
			Key:         "pl151",
			URL:         "https://pl151.plemiona.pl",
			Open:        true,
			VersionCode: "pl",
		}
		ev, err := marshaler.Marshal(serverRefreshedPayload)
		require.NoError(t, err)
		require.NoError(t, pubSub.Publish("servers.event.refreshed", ev))

		receivedMsgs, _ := subscriber.BulkRead(msgs, 1, 10*time.Second)
		require.Len(t, receivedMsgs, 1)
		var villagesRefreshedPayload model.VillagesRefreshedEvPayload
		assert.NoError(t, marshaler.Unmarshal(receivedMsgs[0], &villagesRefreshedPayload))
		assert.Equal(t, serverRefreshedPayload.Key, villagesRefreshedPayload.Key)
		assert.Equal(t, serverRefreshedPayload.URL, villagesRefreshedPayload.URL)
		assert.Equal(t, serverRefreshedPayload.VersionCode, villagesRefreshedPayload.VersionCode)
		assert.Equal(t, refreshResult.NumVillages, villagesRefreshedPayload.NumVillages)
		assert.Equal(t, refreshResult.NumPlayerVillages, villagesRefreshedPayload.NumPlayerVillages)
		assert.Equal(t, refreshResult.NumBarbarianVillages, villagesRefreshedPayload.NumBarbarianVillages)
		assert.Equal(t, refreshResult.NumBonusVillages, villagesRefreshedPayload.NumBonusVillages)
	})

	t.Run("OK: only open servers should be updated", func(t *testing.T) {
		t.Parallel()

		ctx, cancel := context.WithTimeout(context.Background(), time.Second)
		defer cancel()

		marshaler := msg.JSONMarshaler{NewUUID: uuid.NewString}
		pubSub := newPubSub(t)
		villageSvc := &mock.FakeVillageService{}

		waitMiddleware, wait := newWaitForHandlerMiddleware("VillageConsumer.refresh")
		runRouter(
			t,
			waitMiddleware,
			msg.NewVillageConsumer(marshaler, pubSub, pubSub, watermill.NopLogger{}, villageSvc),
		)

		ev, err := marshaler.Marshal(model.ServerRefreshedEvPayload{Open: false})
		require.NoError(t, err)
		require.NoError(t, pubSub.Publish("servers.event.refreshed", ev))
		ev.SetContext(ctx)

		select {
		case <-wait:
		case <-ctx.Done():
			t.Fatal("timeout")
		}

		assert.Equal(t, 0, villageSvc.RefreshCallCount())
	})
}
