package rest_test

import (
	"context"
	"errors"
	"fmt"
	"net/http"
	"net/url"
	"strconv"
	"testing"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/mock"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
	"github.com/google/go-cmp/cmp/cmpopts"

	"github.com/google/go-cmp/cmp"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
)

func TestEnnoblement_list(t *testing.T) {
	t.Parallel()

	now := time.Now()
	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	server := domain.Server{
		Key:         "pl151",
		URL:         "https://pl151.plemiona.pl",
		Open:        true,
		Special:     false,
		VersionCode: "pl",
	}

	tests := []struct {
		name               string
		setup              func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, ennoblementSvc *mock.FakeEnnoblementService)
		versionCode        string
		serverKey          string
		queryParams        url.Values
		expectedStatus     int
		target             any
		expectedResponse   any
		expectedTotalCount string
	}{
		{
			name: "OK: without params",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, ennoblementSvc *mock.FakeEnnoblementService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				ennoblementSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListEnnoblementsParams,
				) ([]domain.EnnoblementWithRelations, int64, error) {
					expectedParams := domain.ListEnnoblementsParams{
						ServerKeys: []string{server.Key},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					ennoblements := []domain.EnnoblementWithRelations{
						{
							Ennoblement: domain.Ennoblement{
								ID: 1234,
								BaseEnnoblement: domain.BaseEnnoblement{
									VillageID:  123,
									NewOwnerID: 124,
									NewTribeID: 125,
									OldOwnerID: 126,
									OldTribeID: 127,
									Points:     128,
									CreatedAt:  now,
								},
								ServerKey: server.Key,
							},
							Village: domain.VillageMeta{
								ID:         123,
								Name:       "village name",
								Continent:  "K44",
								X:          450,
								Y:          450,
								ProfileURL: "profile-123",
							},
							NewOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         124,
									Name:       "new owner name",
									ProfileURL: "profile-124",
								},
							},
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         125,
									Name:       "new tribe name",
									Tag:        "new tribe tag",
									ProfileURL: "profile-125",
								},
							},
							OldOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         126,
									Name:       "old owner name",
									ProfileURL: "profile-126",
								},
							},
							OldTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         127,
									Name:       "old tribe name",
									Tag:        "old tribe tag",
									ProfileURL: "profile-127",
								},
							},
						},
						{
							Ennoblement: domain.Ennoblement{
								ID: 12345,
								BaseEnnoblement: domain.BaseEnnoblement{
									VillageID:  311,
									NewOwnerID: 312,
									NewTribeID: 313,
									OldOwnerID: 0,
									OldTribeID: 0,
									Points:     129,
									CreatedAt:  now,
								},
								ServerKey: server.Key,
							},
							Village: domain.VillageMeta{
								ID:         311,
								Name:       "village name",
								Continent:  "K55",
								X:          550,
								Y:          550,
								ProfileURL: "profile-311",
							},
							NewOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         312,
									Name:       "new owner name",
									ProfileURL: "profile-312",
								},
							},
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         313,
									Name:       "new tribe name",
									Tag:        "new tribe tag",
									ProfileURL: "profile-313",
								},
							},
						},
					}
					return ennoblements, int64(len(ennoblements)), nil
				})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			expectedStatus: http.StatusOK,
			target:         &model.ListEnnoblementsResp{},
			expectedResponse: &model.ListEnnoblementsResp{
				Data: []model.Ennoblement{
					{
						ID:     1234,
						Points: 128,
						Village: model.VillageMeta{
							ID:       123,
							FullName: "village name (450|450) K44",
							X:        450,
							Y:        450,
							Player: model.NullPlayerMeta{
								Valid: true,
								Player: model.PlayerMeta{
									ID:         126,
									Name:       "old owner name",
									ProfileURL: "profile-126",
									Tribe: model.NullTribeMeta{
										Valid: true,
										Tribe: model.TribeMeta{
											ID:         127,
											Name:       "old tribe name",
											Tag:        "old tribe tag",
											ProfileURL: "profile-127",
										},
									},
								},
							},
							Continent:  "K44",
							ProfileURL: "profile-123",
						},
						NewOwner: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         124,
								Name:       "new owner name",
								ProfileURL: "profile-124",
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         125,
										Name:       "new tribe name",
										Tag:        "new tribe tag",
										ProfileURL: "profile-125",
									},
								},
							},
						},
						CreatedAt: now,
					},
					{
						ID:     12345,
						Points: 129,
						Village: model.VillageMeta{
							ID:         311,
							FullName:   "village name (550|550) K55",
							ProfileURL: "profile-311",
							Continent:  "K55",
							X:          550,
							Y:          550,
						},
						NewOwner: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         312,
								Name:       "new owner name",
								ProfileURL: "profile-312",
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         313,
										Name:       "new tribe name",
										Tag:        "new tribe tag",
										ProfileURL: "profile-313",
									},
								},
							},
						},
						CreatedAt: now,
					},
				},
			},
			expectedTotalCount: "2",
		},
		{
			name: "OK: limit=1,offset=100,since=now-30h,before=now,sort=[createdAt:desc]",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, ennoblementSvc *mock.FakeEnnoblementService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				ennoblementSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListEnnoblementsParams,
				) ([]domain.EnnoblementWithRelations, int64, error) {
					expectedParams := domain.ListEnnoblementsParams{
						ServerKeys: []string{server.Key},
						CreatedAtGTE: domain.NullTime{
							Valid: true,
							Time:  now.Add(-30 * time.Hour).Truncate(time.Second),
						},
						CreatedAtLTE: domain.NullTime{
							Valid: true,
							Time:  now.Truncate(time.Second),
						},
						Sort: []domain.EnnoblementSort{
							{By: domain.EnnoblementSortByCreatedAt, Direction: domain.SortDirectionDESC},
						},
						Pagination: domain.Pagination{
							Limit:  1,
							Offset: 100,
						},
					}
					if diff := cmp.Diff(params, expectedParams, cmpopts.IgnoreUnexported(time.Time{})); diff != "" {
						return nil, 0, errors.New("invalid params: " + diff)
					}

					ennoblements := []domain.EnnoblementWithRelations{
						{
							Ennoblement: domain.Ennoblement{
								ID: 12345,
								BaseEnnoblement: domain.BaseEnnoblement{
									VillageID:  311,
									NewOwnerID: 312,
									NewTribeID: 313,
									OldOwnerID: 0,
									OldTribeID: 0,
									Points:     129,
									CreatedAt:  now,
								},
								ServerKey: server.Key,
							},
							Village: domain.VillageMeta{
								ID:         311,
								Name:       "village name",
								Continent:  "K44",
								X:          450,
								Y:          450,
								ProfileURL: "profile-311",
							},
							NewOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         312,
									Name:       "new owner name",
									ProfileURL: "profile-312",
								},
							},
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         313,
									Name:       "new tribe name",
									Tag:        "new tribe tag",
									ProfileURL: "profile-313",
								},
							},
						},
					}
					return ennoblements, int64(len(ennoblements)), nil
				})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			expectedStatus: http.StatusOK,
			queryParams: url.Values{
				"limit":  []string{"1"},
				"offset": []string{"100"},
				"sort":   []string{"createdAt:DESC"},
				"since":  []string{now.Add(-30 * time.Hour).Format(time.RFC3339)},
				"before": []string{now.Format(time.RFC3339)},
			},
			target: &model.ListEnnoblementsResp{},
			expectedResponse: &model.ListEnnoblementsResp{
				Data: []model.Ennoblement{
					{
						ID:     12345,
						Points: 129,
						Village: model.VillageMeta{
							ID:         311,
							FullName:   "village name (450|450) K44",
							Continent:  "K44",
							X:          450,
							Y:          450,
							ProfileURL: "profile-311",
						},
						NewOwner: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         312,
								Name:       "new owner name",
								ProfileURL: "profile-312",
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         313,
										Name:       "new tribe name",
										Tag:        "new tribe tag",
										ProfileURL: "profile-313",
									},
								},
							},
						},
						CreatedAt: now,
					},
				},
			},
			expectedTotalCount: "1",
		},
		{
			name: "ERR: since is not a valid timestamp",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, ennoblementSvc *mock.FakeEnnoblementService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			queryParams: url.Values{
				"since": []string{"???"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "since: parsing time \"???\" as \"2006-01-02T15:04:05Z07:00\": cannot parse \"???\" as \"2006\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: before is not a valid timestamp",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, ennoblementSvc *mock.FakeEnnoblementService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			queryParams: url.Values{
				"before": []string{"???"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "before: parsing time \"???\" as \"2006-01-02T15:04:05Z07:00\": cannot parse \"???\" as \"2006\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: limit is not a valid int32",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, ennoblementSvc *mock.FakeEnnoblementService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			queryParams: url.Values{
				"limit": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "limit: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: offset is not a valid int32",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, ennoblementSvc *mock.FakeEnnoblementService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			queryParams: url.Values{
				"offset": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "offset: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - invalid format",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, ennoblementSvc *mock.FakeEnnoblementService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc", "test"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: parsing \"test\": invalid syntax, expected field:direction",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported field",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, ennoblementSvc *mock.FakeEnnoblementService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc", "test:asc"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: unsupported sort by: \"test\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported direction",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, ennoblementSvc *mock.FakeEnnoblementService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc2"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[0]: unsupported sort direction: \"asc2\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: version not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, ennoblementSvc *mock.FakeEnnoblementService) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			serverKey:      server.Key,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: server not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, ennoblementSvc *mock.FakeEnnoblementService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{}, domain.ServerNotFoundError{Key: server.Key + "2"})
			},
			versionCode:    version.Code,
			serverKey:      server.Key + "2",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("server (key=%s) not found", server.Key+"2"),
				},
			},
			expectedTotalCount: "",
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			ennoblementSvc := &mock.FakeEnnoblementService{}
			tt.setup(versionSvc, serverSvc, ennoblementSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
				withEnnoblementService(ennoblementSvc),
			)

			target := fmt.Sprintf(
				"/v1/versions/%s/servers/%s/ennoblements?%s",
				tt.versionCode,
				tt.serverKey,
				tt.queryParams.Encode(),
			)
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertTotalCount(t, resp.Header, tt.expectedTotalCount)
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}

func TestEnnoblement_listPlayer(t *testing.T) {
	t.Parallel()

	now := time.Now()
	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	server := domain.Server{
		Key:         "pl151",
		URL:         "https://pl151.plemiona.pl",
		Open:        true,
		Special:     false,
		VersionCode: "pl",
	}
	player := domain.PlayerWithRelations{
		Player: domain.Player{
			BasePlayer: domain.BasePlayer{
				ID:         124,
				Name:       "player 124",
				ProfileURL: "player-124",
			},
		},
	}
	playerID := strconv.FormatInt(player.ID, 10)

	tests := []struct {
		name  string
		setup func(
			versionSvc *mock.FakeVersionService,
			serverSvc *mock.FakeServerService,
			playerSvc *mock.FakePlayerService,
			ennoblementSvc *mock.FakeEnnoblementService,
		)
		versionCode        string
		serverKey          string
		playerID           string
		queryParams        url.Values
		expectedStatus     int
		target             any
		expectedResponse   any
		expectedTotalCount string
	}{
		{
			name: "OK: without params",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
				ennoblementSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListEnnoblementsParams,
				) ([]domain.EnnoblementWithRelations, int64, error) {
					expectedParams := domain.ListEnnoblementsParams{
						ServerKeys: []string{server.Key},
						PlayerID: domain.NullInt64{
							Int64: player.ID,
							Valid: true,
						},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					ennoblements := []domain.EnnoblementWithRelations{
						{
							Ennoblement: domain.Ennoblement{
								ID: 1234,
								BaseEnnoblement: domain.BaseEnnoblement{
									VillageID:  123,
									NewOwnerID: 124,
									NewTribeID: 125,
									OldOwnerID: 126,
									OldTribeID: 127,
									Points:     128,
									CreatedAt:  now,
								},
								ServerKey: server.Key,
							},
							Village: domain.VillageMeta{
								ID:         123,
								Name:       "village name",
								Continent:  "K44",
								X:          450,
								Y:          450,
								ProfileURL: "profile-123",
							},
							OldOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         player.ID,
									Name:       player.Name,
									ProfileURL: player.ProfileURL,
								},
							},
							OldTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         125,
									Name:       "tribe 125",
									Tag:        "tribe tag 125",
									ProfileURL: "profile-125",
								},
							},
							NewOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         126,
									Name:       "player 126",
									ProfileURL: "profile-126",
								},
							},
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         127,
									Name:       "tribe 127",
									Tag:        "tribe tag 127",
									ProfileURL: "profile-127",
								},
							},
						},
						{
							Ennoblement: domain.Ennoblement{
								ID: 12345,
								BaseEnnoblement: domain.BaseEnnoblement{
									VillageID:  311,
									NewOwnerID: 312,
									NewTribeID: 313,
									OldOwnerID: 0,
									OldTribeID: 0,
									Points:     129,
									CreatedAt:  now,
								},
								ServerKey: server.Key,
							},
							Village: domain.VillageMeta{
								ID:         311,
								Name:       "village name",
								Continent:  "K55",
								X:          550,
								Y:          550,
								ProfileURL: "profile-311",
							},
							NewOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         player.ID,
									Name:       player.Name,
									ProfileURL: player.ProfileURL,
								},
							},
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         125,
									Name:       "tribe 125",
									Tag:        "tribe tag 125",
									ProfileURL: "profile-125",
								},
							},
						},
					}
					return ennoblements, int64(len(ennoblements)), nil
				})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			playerID:       playerID,
			expectedStatus: http.StatusOK,
			target:         &model.ListEnnoblementsResp{},
			expectedResponse: &model.ListEnnoblementsResp{
				Data: []model.Ennoblement{
					{
						ID:     1234,
						Points: 128,
						Village: model.VillageMeta{
							ID:         123,
							FullName:   "village name (450|450) K44",
							Continent:  "K44",
							X:          450,
							Y:          450,
							ProfileURL: "profile-123",
							Player: model.NullPlayerMeta{
								Valid: true,
								Player: model.PlayerMeta{
									ID:         player.ID,
									Name:       player.Name,
									ProfileURL: player.ProfileURL,
									Tribe: model.NullTribeMeta{
										Valid: true,
										Tribe: model.TribeMeta{
											ID:         125,
											Name:       "tribe 125",
											Tag:        "tribe tag 125",
											ProfileURL: "profile-125",
										},
									},
								},
							},
						},
						NewOwner: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         126,
								Name:       "player 126",
								ProfileURL: "profile-126",
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         127,
										Name:       "tribe 127",
										Tag:        "tribe tag 127",
										ProfileURL: "profile-127",
									},
								},
							},
						},
						CreatedAt: now,
					},
					{
						ID:     12345,
						Points: 129,
						Village: model.VillageMeta{
							ID:         311,
							FullName:   "village name (550|550) K55",
							ProfileURL: "profile-311",
							Continent:  "K55",
							X:          550,
							Y:          550,
						},
						NewOwner: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         player.ID,
								Name:       player.Name,
								ProfileURL: player.ProfileURL,
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         125,
										Name:       "tribe 125",
										Tag:        "tribe tag 125",
										ProfileURL: "profile-125",
									},
								},
							},
						},
						CreatedAt: now,
					},
				},
			},
			expectedTotalCount: "2",
		},
		{
			name: "OK: limit=1,offset=100,since=now-30h,before=now,sort=[createdAt:desc]",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
				ennoblementSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListEnnoblementsParams,
				) ([]domain.EnnoblementWithRelations, int64, error) {
					expectedParams := domain.ListEnnoblementsParams{
						ServerKeys: []string{server.Key},
						CreatedAtGTE: domain.NullTime{
							Valid: true,
							Time:  now.Add(-30 * time.Hour).Truncate(time.Second),
						},
						CreatedAtLTE: domain.NullTime{
							Valid: true,
							Time:  now.Truncate(time.Second),
						},
						Sort: []domain.EnnoblementSort{
							{By: domain.EnnoblementSortByCreatedAt, Direction: domain.SortDirectionDESC},
						},
						PlayerID: domain.NullInt64{
							Int64: player.ID,
							Valid: true,
						},
						Pagination: domain.Pagination{
							Limit:  1,
							Offset: 100,
						},
					}
					if diff := cmp.Diff(params, expectedParams, cmpopts.IgnoreUnexported(time.Time{})); diff != "" {
						return nil, 0, errors.New("invalid params: " + diff)
					}

					ennoblements := []domain.EnnoblementWithRelations{
						{
							Ennoblement: domain.Ennoblement{
								ID: 12345,
								BaseEnnoblement: domain.BaseEnnoblement{
									VillageID:  311,
									NewOwnerID: 312,
									NewTribeID: 313,
									OldOwnerID: 0,
									OldTribeID: 0,
									Points:     129,
									CreatedAt:  now,
								},
								ServerKey: server.Key,
							},
							Village: domain.VillageMeta{
								ID:         311,
								Name:       "village name",
								Continent:  "K44",
								X:          450,
								Y:          450,
								ProfileURL: "profile-311",
							},
							NewOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         player.ID,
									Name:       player.Name,
									ProfileURL: player.ProfileURL,
								},
							},
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         313,
									Name:       "new tribe name",
									Tag:        "new tribe tag",
									ProfileURL: "profile-313",
								},
							},
						},
					}
					return ennoblements, int64(len(ennoblements)), nil
				})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			playerID:       playerID,
			expectedStatus: http.StatusOK,
			queryParams: url.Values{
				"limit":  []string{"1"},
				"offset": []string{"100"},
				"sort":   []string{"createdAt:DESC"},
				"since":  []string{now.Add(-30 * time.Hour).Format(time.RFC3339)},
				"before": []string{now.Format(time.RFC3339)},
			},
			target: &model.ListEnnoblementsResp{},
			expectedResponse: &model.ListEnnoblementsResp{
				Data: []model.Ennoblement{
					{
						ID:     12345,
						Points: 129,
						Village: model.VillageMeta{
							ID:         311,
							FullName:   "village name (450|450) K44",
							Continent:  "K44",
							X:          450,
							Y:          450,
							ProfileURL: "profile-311",
						},
						NewOwner: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         player.ID,
								Name:       player.Name,
								ProfileURL: player.ProfileURL,
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         313,
										Name:       "new tribe name",
										Tag:        "new tribe tag",
										ProfileURL: "profile-313",
									},
								},
							},
						},
						CreatedAt: now,
					},
				},
			},
			expectedTotalCount: "1",
		},
		{
			name: "ERR: since is not a valid timestamp",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"since": []string{"???"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "since: parsing time \"???\" as \"2006-01-02T15:04:05Z07:00\": cannot parse \"???\" as \"2006\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: before is not a valid timestamp",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"before": []string{"???"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "before: parsing time \"???\" as \"2006-01-02T15:04:05Z07:00\": cannot parse \"???\" as \"2006\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: limit is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"limit": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "limit: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: offset is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"offset": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "offset: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - invalid format",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc", "test"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: parsing \"test\": invalid syntax, expected field:direction",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported field",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc", "test:asc"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: unsupported sort by: \"test\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported direction",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc2"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[0]: unsupported sort direction: \"asc2\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: player id is not a valid int64",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			playerID:       "asdf",
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "playerId: strconv.ParseInt: parsing \"asdf\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: version not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			serverKey:      server.Key,
			playerID:       playerID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: server not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{}, domain.ServerNotFoundError{Key: server.Key + "2"})
			},
			versionCode:    version.Code,
			serverKey:      server.Key + "2",
			playerID:       playerID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("server (key=%s) not found", server.Key+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: player not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(domain.PlayerWithRelations{}, domain.PlayerNotFoundError{ID: 1241})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			playerID:       playerID + "1",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("player (id=%s) not found", playerID+"1"),
				},
			},
			expectedTotalCount: "",
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			ennoblementSvc := &mock.FakeEnnoblementService{}
			playerSvc := &mock.FakePlayerService{}
			tt.setup(versionSvc, serverSvc, playerSvc, ennoblementSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
				withPlayerService(playerSvc),
				withEnnoblementService(ennoblementSvc),
			)

			target := fmt.Sprintf(
				"/v1/versions/%s/servers/%s/players/%s/ennoblements?%s",
				tt.versionCode,
				tt.serverKey,
				tt.playerID,
				tt.queryParams.Encode(),
			)
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertTotalCount(t, resp.Header, tt.expectedTotalCount)
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}

func TestEnnoblement_listVillage(t *testing.T) {
	t.Parallel()

	now := time.Now()
	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	server := domain.Server{
		Key:         "pl151",
		URL:         "https://pl151.plemiona.pl",
		Open:        true,
		Special:     false,
		VersionCode: "pl",
	}
	village := domain.VillageWithRelations{
		Village: domain.Village{
			BaseVillage: domain.BaseVillage{
				ID:         124,
				Name:       "village 124",
				ProfileURL: "village-124",
				Continent:  "K44",
				X:          450,
				Y:          450,
			},
		},
	}
	villageID := strconv.FormatInt(village.ID, 10)

	tests := []struct {
		name  string
		setup func(
			versionSvc *mock.FakeVersionService,
			serverSvc *mock.FakeServerService,
			villageSvc *mock.FakeVillageService,
			ennoblementSvc *mock.FakeEnnoblementService,
		)
		versionCode        string
		serverKey          string
		villageID          string
		queryParams        url.Values
		expectedStatus     int
		target             any
		expectedResponse   any
		expectedTotalCount string
	}{
		{
			name: "OK: without params",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				villageSvc *mock.FakeVillageService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				villageSvc.GetByServerKeyAndIDWithRelationsReturns(village, nil)
				ennoblementSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListEnnoblementsParams,
				) ([]domain.EnnoblementWithRelations, int64, error) {
					expectedParams := domain.ListEnnoblementsParams{
						ServerKeys: []string{server.Key},
						VillageID: domain.NullInt64{
							Int64: village.ID,
							Valid: true,
						},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					ennoblements := []domain.EnnoblementWithRelations{
						{
							Ennoblement: domain.Ennoblement{
								ID: 1234,
								BaseEnnoblement: domain.BaseEnnoblement{
									VillageID:  123,
									NewOwnerID: 124,
									NewTribeID: 125,
									OldOwnerID: 126,
									OldTribeID: 127,
									Points:     128,
									CreatedAt:  now,
								},
								ServerKey: server.Key,
							},
							Village: domain.VillageMeta{
								ID:         village.ID,
								Name:       village.Name,
								ProfileURL: village.ProfileURL,
								Continent:  village.Continent,
								X:          village.X,
								Y:          village.Y,
							},
							OldOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         1114,
									Name:       "player-1114",
									ProfileURL: "profile-1114",
								},
							},
							OldTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         125,
									Name:       "tribe 125",
									Tag:        "tribe tag 125",
									ProfileURL: "profile-125",
								},
							},
							NewOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         126,
									Name:       "player 126",
									ProfileURL: "profile-126",
								},
							},
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         127,
									Name:       "tribe 127",
									Tag:        "tribe tag 127",
									ProfileURL: "profile-127",
								},
							},
						},
						{
							Ennoblement: domain.Ennoblement{
								ID: 12345,
								BaseEnnoblement: domain.BaseEnnoblement{
									VillageID:  311,
									NewOwnerID: 312,
									NewTribeID: 313,
									OldOwnerID: 0,
									OldTribeID: 0,
									Points:     129,
									CreatedAt:  now,
								},
								ServerKey: server.Key,
							},
							Village: domain.VillageMeta{
								ID:         village.ID,
								Name:       village.Name,
								ProfileURL: village.ProfileURL,
								Continent:  village.Continent,
								X:          village.X,
								Y:          village.Y,
							},
							NewOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         1114,
									Name:       "player-1114",
									ProfileURL: "profile-1114",
								},
							},
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         125,
									Name:       "tribe 125",
									Tag:        "tribe tag 125",
									ProfileURL: "profile-125",
								},
							},
						},
					}
					return ennoblements, int64(len(ennoblements)), nil
				})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			villageID:      villageID,
			expectedStatus: http.StatusOK,
			target:         &model.ListEnnoblementsResp{},
			expectedResponse: &model.ListEnnoblementsResp{
				Data: []model.Ennoblement{
					{
						ID:     1234,
						Points: 128,
						Village: model.VillageMeta{
							ID:         village.ID,
							FullName:   village.FullName(),
							ProfileURL: village.ProfileURL,
							Continent:  village.Continent,
							X:          village.X,
							Y:          village.Y,
							Player: model.NullPlayerMeta{
								Valid: true,
								Player: model.PlayerMeta{
									ID:         1114,
									Name:       "player-1114",
									ProfileURL: "profile-1114",
									Tribe: model.NullTribeMeta{
										Valid: true,
										Tribe: model.TribeMeta{
											ID:         125,
											Name:       "tribe 125",
											Tag:        "tribe tag 125",
											ProfileURL: "profile-125",
										},
									},
								},
							},
						},
						NewOwner: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         126,
								Name:       "player 126",
								ProfileURL: "profile-126",
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         127,
										Name:       "tribe 127",
										Tag:        "tribe tag 127",
										ProfileURL: "profile-127",
									},
								},
							},
						},
						CreatedAt: now,
					},
					{
						ID:     12345,
						Points: 129,
						Village: model.VillageMeta{
							ID:         village.ID,
							FullName:   village.FullName(),
							ProfileURL: village.ProfileURL,
							Continent:  village.Continent,
							X:          village.X,
							Y:          village.Y,
						},
						NewOwner: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         1114,
								Name:       "player-1114",
								ProfileURL: "profile-1114",
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         125,
										Name:       "tribe 125",
										Tag:        "tribe tag 125",
										ProfileURL: "profile-125",
									},
								},
							},
						},
						CreatedAt: now,
					},
				},
			},
			expectedTotalCount: "2",
		},
		{
			name: "OK: limit=1,offset=100,since=now-30h,before=now,sort=[createdAt:desc]",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				villageSvc *mock.FakeVillageService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				villageSvc.GetByServerKeyAndIDWithRelationsReturns(village, nil)
				ennoblementSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListEnnoblementsParams,
				) ([]domain.EnnoblementWithRelations, int64, error) {
					expectedParams := domain.ListEnnoblementsParams{
						ServerKeys: []string{server.Key},
						CreatedAtGTE: domain.NullTime{
							Valid: true,
							Time:  now.Add(-30 * time.Hour).Truncate(time.Second),
						},
						CreatedAtLTE: domain.NullTime{
							Valid: true,
							Time:  now.Truncate(time.Second),
						},
						Sort: []domain.EnnoblementSort{
							{By: domain.EnnoblementSortByCreatedAt, Direction: domain.SortDirectionDESC},
						},
						VillageID: domain.NullInt64{
							Int64: village.ID,
							Valid: true,
						},
						Pagination: domain.Pagination{
							Limit:  1,
							Offset: 100,
						},
					}
					if diff := cmp.Diff(params, expectedParams, cmpopts.IgnoreUnexported(time.Time{})); diff != "" {
						return nil, 0, errors.New("invalid params: " + diff)
					}

					ennoblements := []domain.EnnoblementWithRelations{
						{
							Ennoblement: domain.Ennoblement{
								ID: 12345,
								BaseEnnoblement: domain.BaseEnnoblement{
									VillageID:  311,
									NewOwnerID: 312,
									NewTribeID: 313,
									OldOwnerID: 0,
									OldTribeID: 0,
									Points:     129,
									CreatedAt:  now,
								},
								ServerKey: server.Key,
							},
							Village: domain.VillageMeta{
								ID:         village.ID,
								Name:       village.Name,
								ProfileURL: village.ProfileURL,
								Continent:  village.Continent,
								X:          village.X,
								Y:          village.Y,
							},
							NewOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         1115,
									Name:       "name-1115",
									ProfileURL: "profile-1115",
								},
							},
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         313,
									Name:       "new tribe name",
									Tag:        "new tribe tag",
									ProfileURL: "profile-313",
								},
							},
						},
					}
					return ennoblements, int64(len(ennoblements)), nil
				})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			villageID:      villageID,
			expectedStatus: http.StatusOK,
			queryParams: url.Values{
				"limit":  []string{"1"},
				"offset": []string{"100"},
				"sort":   []string{"createdAt:DESC"},
				"since":  []string{now.Add(-30 * time.Hour).Format(time.RFC3339)},
				"before": []string{now.Format(time.RFC3339)},
			},
			target: &model.ListEnnoblementsResp{},
			expectedResponse: &model.ListEnnoblementsResp{
				Data: []model.Ennoblement{
					{
						ID:     12345,
						Points: 129,
						Village: model.VillageMeta{
							ID:         village.ID,
							FullName:   village.FullName(),
							ProfileURL: village.ProfileURL,
							Continent:  village.Continent,
							X:          village.X,
							Y:          village.Y,
						},
						NewOwner: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         1115,
								Name:       "name-1115",
								ProfileURL: "profile-1115",
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         313,
										Name:       "new tribe name",
										Tag:        "new tribe tag",
										ProfileURL: "profile-313",
									},
								},
							},
						},
						CreatedAt: now,
					},
				},
			},
			expectedTotalCount: "1",
		},
		{
			name: "ERR: since is not a valid timestamp",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				villageSvc *mock.FakeVillageService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				villageSvc.GetByServerKeyAndIDWithRelationsReturns(village, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			villageID:   villageID,
			queryParams: url.Values{
				"since": []string{"???"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "since: parsing time \"???\" as \"2006-01-02T15:04:05Z07:00\": cannot parse \"???\" as \"2006\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: before is not a valid timestamp",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				villageSvc *mock.FakeVillageService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				villageSvc.GetByServerKeyAndIDWithRelationsReturns(village, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			villageID:   villageID,
			queryParams: url.Values{
				"before": []string{"???"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "before: parsing time \"???\" as \"2006-01-02T15:04:05Z07:00\": cannot parse \"???\" as \"2006\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: limit is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				villageSvc *mock.FakeVillageService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				villageSvc.GetByServerKeyAndIDWithRelationsReturns(village, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			villageID:   villageID,
			queryParams: url.Values{
				"limit": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "limit: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: offset is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				villageSvc *mock.FakeVillageService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				villageSvc.GetByServerKeyAndIDWithRelationsReturns(village, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			villageID:   villageID,
			queryParams: url.Values{
				"offset": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "offset: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - invalid format",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				villageSvc *mock.FakeVillageService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				villageSvc.GetByServerKeyAndIDWithRelationsReturns(village, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			villageID:   villageID,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc", "test"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: parsing \"test\": invalid syntax, expected field:direction",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported field",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				villageSvc *mock.FakeVillageService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				villageSvc.GetByServerKeyAndIDWithRelationsReturns(village, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			villageID:   villageID,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc", "test:asc"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: unsupported sort by: \"test\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported direction",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				villageSvc *mock.FakeVillageService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				villageSvc.GetByServerKeyAndIDWithRelationsReturns(village, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			villageID:   villageID,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc2"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[0]: unsupported sort direction: \"asc2\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: village id is not a valid int64",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				villageSvc *mock.FakeVillageService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			villageID:      "asdf",
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "villageId: strconv.ParseInt: parsing \"asdf\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: version not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				villageSvc *mock.FakeVillageService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			serverKey:      server.Key,
			villageID:      villageID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: server not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				villageSvc *mock.FakeVillageService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{}, domain.ServerNotFoundError{Key: server.Key + "2"})
			},
			versionCode:    version.Code,
			serverKey:      server.Key + "2",
			villageID:      villageID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("server (key=%s) not found", server.Key+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: player not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				villageSvc *mock.FakeVillageService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				villageSvc.GetByServerKeyAndIDWithRelationsReturns(domain.VillageWithRelations{}, domain.VillageNotFoundError{ID: 1241})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			villageID:      villageID + "1",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("village (id=%s) not found", villageID+"1"),
				},
			},
			expectedTotalCount: "",
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			ennoblementSvc := &mock.FakeEnnoblementService{}
			villageSvc := &mock.FakeVillageService{}
			tt.setup(versionSvc, serverSvc, villageSvc, ennoblementSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
				withVillageService(villageSvc),
				withEnnoblementService(ennoblementSvc),
			)

			target := fmt.Sprintf(
				"/v1/versions/%s/servers/%s/villages/%s/ennoblements?%s",
				tt.versionCode,
				tt.serverKey,
				tt.villageID,
				tt.queryParams.Encode(),
			)
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertTotalCount(t, resp.Header, tt.expectedTotalCount)
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}

func TestEnnoblement_listTribe(t *testing.T) {
	t.Parallel()

	now := time.Now()
	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	server := domain.Server{
		Key:         "pl151",
		URL:         "https://pl151.plemiona.pl",
		Open:        true,
		Special:     false,
		VersionCode: "pl",
	}
	tribe := domain.Tribe{
		BaseTribe: domain.BaseTribe{
			ID:         124,
			Name:       "tribe 124",
			Tag:        "tag 124",
			ProfileURL: "tribe-124",
		},
	}
	tribeID := strconv.FormatInt(tribe.ID, 10)

	tests := []struct {
		name  string
		setup func(
			versionSvc *mock.FakeVersionService,
			serverSvc *mock.FakeServerService,
			tribeSvc *mock.FakeTribeService,
			ennoblementSvc *mock.FakeEnnoblementService,
		)
		versionCode        string
		serverKey          string
		tribeID            string
		queryParams        url.Values
		expectedStatus     int
		target             any
		expectedResponse   any
		expectedTotalCount string
	}{
		{
			name: "OK: without params",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
				ennoblementSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListEnnoblementsParams,
				) ([]domain.EnnoblementWithRelations, int64, error) {
					expectedParams := domain.ListEnnoblementsParams{
						ServerKeys: []string{server.Key},
						TribeID: domain.NullInt64{
							Int64: tribe.ID,
							Valid: true,
						},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					ennoblements := []domain.EnnoblementWithRelations{
						{
							Ennoblement: domain.Ennoblement{
								ID: 1234,
								BaseEnnoblement: domain.BaseEnnoblement{
									VillageID:  123,
									NewOwnerID: 124,
									NewTribeID: 125,
									OldOwnerID: 126,
									OldTribeID: 127,
									Points:     128,
									CreatedAt:  now,
								},
								ServerKey: server.Key,
							},
							Village: domain.VillageMeta{
								ID:         123,
								Name:       "village name",
								Continent:  "K44",
								X:          450,
								Y:          450,
								ProfileURL: "profile-123",
							},
							NewOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         124,
									Name:       "new owner name",
									ProfileURL: "profile-124",
								},
							},
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         125,
									Name:       "new tribe name",
									Tag:        "new tribe tag",
									ProfileURL: "profile-125",
								},
							},
							OldOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         126,
									Name:       "old owner name",
									ProfileURL: "profile-126",
								},
							},
							OldTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         tribe.ID,
									Name:       tribe.Name,
									Tag:        tribe.Tag,
									ProfileURL: tribe.ProfileURL,
								},
							},
						},
						{
							Ennoblement: domain.Ennoblement{
								ID: 12345,
								BaseEnnoblement: domain.BaseEnnoblement{
									VillageID:  311,
									NewOwnerID: 312,
									NewTribeID: 313,
									OldOwnerID: 0,
									OldTribeID: 0,
									Points:     129,
									CreatedAt:  now,
								},
								ServerKey: server.Key,
							},
							Village: domain.VillageMeta{
								ID:         311,
								Name:       "village name",
								Continent:  "K55",
								X:          550,
								Y:          550,
								ProfileURL: "profile-311",
							},
							NewOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         312,
									Name:       "new owner name",
									ProfileURL: "profile-312",
								},
							},
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         tribe.ID,
									Name:       tribe.Name,
									Tag:        tribe.Tag,
									ProfileURL: tribe.ProfileURL,
								},
							},
						},
					}
					return ennoblements, int64(len(ennoblements)), nil
				})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			tribeID:        tribeID,
			expectedStatus: http.StatusOK,
			target:         &model.ListEnnoblementsResp{},
			expectedResponse: &model.ListEnnoblementsResp{
				Data: []model.Ennoblement{
					{
						ID:     1234,
						Points: 128,
						Village: model.VillageMeta{
							ID:       123,
							FullName: "village name (450|450) K44",
							X:        450,
							Y:        450,
							Player: model.NullPlayerMeta{
								Valid: true,
								Player: model.PlayerMeta{
									ID:         126,
									Name:       "old owner name",
									ProfileURL: "profile-126",
									Tribe: model.NullTribeMeta{
										Valid: true,
										Tribe: model.TribeMeta{
											ID:         tribe.ID,
											Name:       tribe.Name,
											Tag:        tribe.Tag,
											ProfileURL: tribe.ProfileURL,
										},
									},
								},
							},
							Continent:  "K44",
							ProfileURL: "profile-123",
						},
						NewOwner: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         124,
								Name:       "new owner name",
								ProfileURL: "profile-124",
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         125,
										Name:       "new tribe name",
										Tag:        "new tribe tag",
										ProfileURL: "profile-125",
									},
								},
							},
						},
						CreatedAt: now,
					},
					{
						ID:     12345,
						Points: 129,
						Village: model.VillageMeta{
							ID:         311,
							FullName:   "village name (550|550) K55",
							ProfileURL: "profile-311",
							Continent:  "K55",
							X:          550,
							Y:          550,
						},
						NewOwner: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         312,
								Name:       "new owner name",
								ProfileURL: "profile-312",
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         tribe.ID,
										Name:       tribe.Name,
										Tag:        tribe.Tag,
										ProfileURL: tribe.ProfileURL,
									},
								},
							},
						},
						CreatedAt: now,
					},
				},
			},
			expectedTotalCount: "2",
		},
		{
			name: "OK: limit=1,offset=100,since=now-30h,before=now,sort=[createdAt:desc]",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
				ennoblementSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListEnnoblementsParams,
				) ([]domain.EnnoblementWithRelations, int64, error) {
					expectedParams := domain.ListEnnoblementsParams{
						ServerKeys: []string{server.Key},
						CreatedAtGTE: domain.NullTime{
							Valid: true,
							Time:  now.Add(-30 * time.Hour).Truncate(time.Second),
						},
						CreatedAtLTE: domain.NullTime{
							Valid: true,
							Time:  now.Truncate(time.Second),
						},
						Sort: []domain.EnnoblementSort{
							{By: domain.EnnoblementSortByCreatedAt, Direction: domain.SortDirectionDESC},
						},
						TribeID: domain.NullInt64{
							Int64: tribe.ID,
							Valid: true,
						},
						Pagination: domain.Pagination{
							Limit:  1,
							Offset: 100,
						},
					}
					if diff := cmp.Diff(params, expectedParams, cmpopts.IgnoreUnexported(time.Time{})); diff != "" {
						return nil, 0, errors.New("invalid params: " + diff)
					}

					ennoblements := []domain.EnnoblementWithRelations{
						{
							Ennoblement: domain.Ennoblement{
								ID: 1234,
								BaseEnnoblement: domain.BaseEnnoblement{
									VillageID:  123,
									NewOwnerID: 124,
									NewTribeID: 125,
									OldOwnerID: 126,
									OldTribeID: 127,
									Points:     128,
									CreatedAt:  now,
								},
								ServerKey: server.Key,
							},
							Village: domain.VillageMeta{
								ID:         123,
								Name:       "village name",
								Continent:  "K44",
								X:          450,
								Y:          450,
								ProfileURL: "profile-123",
							},
							NewOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         124,
									Name:       "new owner name",
									ProfileURL: "profile-124",
								},
							},
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         125,
									Name:       "new tribe name",
									Tag:        "new tribe tag",
									ProfileURL: "profile-125",
								},
							},
							OldOwner: domain.NullPlayerMeta{
								Valid: true,
								Player: domain.PlayerMeta{
									ID:         126,
									Name:       "old owner name",
									ProfileURL: "profile-126",
								},
							},
							OldTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         tribe.ID,
									Name:       tribe.Name,
									Tag:        tribe.Tag,
									ProfileURL: tribe.ProfileURL,
								},
							},
						},
					}
					return ennoblements, int64(len(ennoblements)), nil
				})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			tribeID:        tribeID,
			expectedStatus: http.StatusOK,
			queryParams: url.Values{
				"limit":  []string{"1"},
				"offset": []string{"100"},
				"sort":   []string{"createdAt:DESC"},
				"since":  []string{now.Add(-30 * time.Hour).Format(time.RFC3339)},
				"before": []string{now.Format(time.RFC3339)},
			},
			target: &model.ListEnnoblementsResp{},
			expectedResponse: &model.ListEnnoblementsResp{
				Data: []model.Ennoblement{
					{
						ID:     1234,
						Points: 128,
						Village: model.VillageMeta{
							ID:       123,
							FullName: "village name (450|450) K44",
							X:        450,
							Y:        450,
							Player: model.NullPlayerMeta{
								Valid: true,
								Player: model.PlayerMeta{
									ID:         126,
									Name:       "old owner name",
									ProfileURL: "profile-126",
									Tribe: model.NullTribeMeta{
										Valid: true,
										Tribe: model.TribeMeta{
											ID:         tribe.ID,
											Name:       tribe.Name,
											Tag:        tribe.Tag,
											ProfileURL: tribe.ProfileURL,
										},
									},
								},
							},
							Continent:  "K44",
							ProfileURL: "profile-123",
						},
						NewOwner: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         124,
								Name:       "new owner name",
								ProfileURL: "profile-124",
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         125,
										Name:       "new tribe name",
										Tag:        "new tribe tag",
										ProfileURL: "profile-125",
									},
								},
							},
						},
						CreatedAt: now,
					},
				},
			},
			expectedTotalCount: "1",
		},
		{
			name: "ERR: since is not a valid timestamp",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"since": []string{"???"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "since: parsing time \"???\" as \"2006-01-02T15:04:05Z07:00\": cannot parse \"???\" as \"2006\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: before is not a valid timestamp",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"before": []string{"???"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "before: parsing time \"???\" as \"2006-01-02T15:04:05Z07:00\": cannot parse \"???\" as \"2006\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: limit is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"limit": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "limit: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: offset is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"offset": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "offset: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - invalid format",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc", "test"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: parsing \"test\": invalid syntax, expected field:direction",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported field",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc", "test:asc"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: unsupported sort by: \"test\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported direction",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc2"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[0]: unsupported sort direction: \"asc2\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: tribe id is not a valid int64",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			tribeID:        "asdf",
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "tribeId: strconv.ParseInt: parsing \"asdf\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: version not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			serverKey:      server.Key,
			tribeID:        tribeID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: server not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{}, domain.ServerNotFoundError{Key: server.Key + "2"})
			},
			versionCode:    version.Code,
			serverKey:      server.Key + "2",
			tribeID:        tribeID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("server (key=%s) not found", server.Key+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: tribe not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				ennoblementSvc *mock.FakeEnnoblementService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(domain.Tribe{}, domain.TribeNotFoundError{ID: 1241})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			tribeID:        tribeID + "1",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("tribe (id=%s) not found", tribeID+"1"),
				},
			},
			expectedTotalCount: "",
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			ennoblementSvc := &mock.FakeEnnoblementService{}
			tribeSvc := &mock.FakeTribeService{}
			tt.setup(versionSvc, serverSvc, tribeSvc, ennoblementSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
				withTribeService(tribeSvc),
				withEnnoblementService(ennoblementSvc),
			)

			target := fmt.Sprintf(
				"/v1/versions/%s/servers/%s/tribes/%s/ennoblements?%s",
				tt.versionCode,
				tt.serverKey,
				tt.tribeID,
				tt.queryParams.Encode(),
			)
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertTotalCount(t, resp.Header, tt.expectedTotalCount)
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}
