package model_test

import (
	"encoding/json"
	"testing"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
	"github.com/stretchr/testify/assert"
)

func TestNullTime_MarshalJSON(t *testing.T) {
	t.Parallel()

	tests := []struct {
		name         string
		t            model.NullTime
		expectedJSON string
	}{
		{
			name: "OK: null 1",
			t: model.NullTime{
				Time:  time.Time{},
				Valid: false,
			},
			expectedJSON: "null",
		},
		{
			name: "OK: null 2",
			t: model.NullTime{
				Time:  time.Now(),
				Valid: false,
			},
			expectedJSON: "null",
		},
		{
			name: "OK: 2022-07-30T14:13:12.0000005Z",
			t: model.NullTime{
				Time:  time.Date(2022, time.July, 30, 14, 13, 12, 500, time.UTC),
				Valid: true,
			},
			expectedJSON: `"2022-07-30T14:13:12.0000005Z"`,
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			b, err := json.Marshal(tt.t)
			assert.NoError(t, err)
			assert.JSONEq(t, tt.expectedJSON, string(b))
		})
	}
}

func TestNullTime_UnmarshalJSON(t *testing.T) {
	t.Parallel()

	tests := []struct {
		name                    string
		json                    string
		expectedT               model.NullTime
		expectedJSONSyntaxError bool
	}{
		{
			name: "OK: null",
			json: "null",
			expectedT: model.NullTime{
				Time:  time.Time{},
				Valid: false,
			},
		},
		{
			name: "OK: 2022-07-30T14:13:12.0000005Z",
			json: `"2022-07-30T14:13:12.0000005Z"`,
			expectedT: model.NullTime{
				Time:  time.Date(2022, time.July, 30, 14, 13, 12, 500, time.UTC),
				Valid: true,
			},
		},
		{
			name:                    "ERR: invalid time 1",
			json:                    "2022-07-30T14:13:12.0000005Z",
			expectedT:               model.NullTime{},
			expectedJSONSyntaxError: true,
		},
		{
			name:                    "ERR: invalid time 2",
			json:                    "hello world",
			expectedT:               model.NullTime{},
			expectedJSONSyntaxError: true,
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			var target model.NullTime
			err := json.Unmarshal([]byte(tt.json), &target)
			if tt.expectedJSONSyntaxError {
				var syntaxError *json.SyntaxError
				assert.ErrorAs(t, err, &syntaxError)
			} else {
				assert.NoError(t, err)
			}
			assert.Equal(t, tt.expectedT, target)
		})
	}
}

func assertNullTime(tb testing.TB, t time.Time, nullT model.NullTime) {
	tb.Helper()

	if t.IsZero() {
		assert.Zero(tb, nullT)
		return
	}

	assert.Equal(tb, t, nullT.Time)
	assert.True(tb, nullT.Valid)
}
