package model

import (
	"encoding/json"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
)

type Player struct {
	ID             int64         `json:"id"`
	Name           string        `json:"name"`
	NumVillages    int64         `json:"numVillages"`
	Points         int64         `json:"points"`
	Rank           int64         `json:"rank"`
	Tribe          NullTribeMeta `json:"tribe" extensions:"x-nullable"`
	RankAtt        int64         `json:"rankAtt"`
	ScoreAtt       int64         `json:"scoreAtt"`
	RankDef        int64         `json:"rankDef"`
	ScoreDef       int64         `json:"scoreDef"`
	RankSup        int64         `json:"rankSup"`
	ScoreSup       int64         `json:"scoreSup"`
	RankTotal      int64         `json:"rankTotal"`
	ScoreTotal     int64         `json:"scoreTotal"`
	ProfileURL     string        `json:"profileUrl"`
	BestRank       int64         `json:"bestRank"`
	BestRankAt     time.Time     `json:"bestRankAt" format:"date-time"`
	MostPoints     int64         `json:"mostPoints"`
	MostPointsAt   time.Time     `json:"mostPointsAt" format:"date-time"`
	MostVillages   int64         `json:"mostVillages"`
	MostVillagesAt time.Time     `json:"mostVillagesAt" format:"date-time"`
	LastActivityAt time.Time     `json:"lastActivityAt" format:"date-time"`
	CreatedAt      time.Time     `json:"createdAt" format:"date-time"`
	DeletedAt      NullTime      `json:"deletedAt" swaggertype:"string" format:"date-time" extensions:"x-nullable"`
} // @name Player

func NewPlayer(p domain.PlayerWithRelations) Player {
	var t NullTribeMeta
	if p.Tribe.Valid {
		t = NullTribeMeta{
			Valid: true,
			Tribe: NewTribeMeta(p.Tribe.Tribe),
		}
	}

	return Player{
		ID:             p.ID,
		Name:           p.Name,
		NumVillages:    p.NumVillages,
		Points:         p.Points,
		Rank:           p.Rank,
		Tribe:          t,
		RankAtt:        p.RankAtt,
		ScoreAtt:       p.ScoreAtt,
		RankDef:        p.RankDef,
		ScoreDef:       p.ScoreDef,
		RankSup:        p.RankSup,
		ScoreSup:       p.ScoreSup,
		RankTotal:      p.RankTotal,
		ScoreTotal:     p.ScoreTotal,
		ProfileURL:     p.ProfileURL,
		BestRank:       p.BestRank,
		BestRankAt:     p.BestRankAt,
		MostPoints:     p.MostPoints,
		MostPointsAt:   p.MostPointsAt,
		MostVillages:   p.MostVillages,
		MostVillagesAt: p.MostVillagesAt,
		LastActivityAt: p.LastActivityAt,
		CreatedAt:      p.CreatedAt,
		DeletedAt: NullTime{
			Time:  p.DeletedAt,
			Valid: !p.DeletedAt.IsZero(),
		},
	}
}

type PlayerWithServer struct {
	Player
	Server ServerMeta `json:"server"`
} // @name PlayerWithServer

func NewPlayerWithServer(p domain.PlayerWithRelations) PlayerWithServer {
	return PlayerWithServer{
		Player: NewPlayer(p),
		Server: NewServerMeta(p.Server),
	}
}

// PlayerMeta represents basic player information
// @Description Basic player information
type PlayerMeta struct {
	ID         int64         `json:"id"`
	Name       string        `json:"name"`
	ProfileURL string        `json:"profileUrl"`
	Tribe      NullTribeMeta `json:"tribe" extensions:"x-nullable"`
} // @name PlayerMeta

func NewPlayerMeta(p domain.PlayerMetaWithRelations) PlayerMeta {
	var t NullTribeMeta
	if p.Tribe.Valid {
		t = NullTribeMeta{
			Valid: true,
			Tribe: NewTribeMeta(p.Tribe.Tribe),
		}
	}
	return PlayerMeta{
		ID:         p.ID,
		Name:       p.Name,
		ProfileURL: p.ProfileURL,
		Tribe:      t,
	}
}

type NullPlayerMeta struct {
	Player PlayerMeta
	Valid  bool
}

func (p NullPlayerMeta) MarshalJSON() ([]byte, error) {
	if !p.Valid {
		return []byte("null"), nil
	}

	return json.Marshal(p.Player)
}

func (p *NullPlayerMeta) UnmarshalJSON(data []byte) error {
	// Ignore null, like in the main JSON package.
	if string(data) == "null" {
		return nil
	}

	if err := json.Unmarshal(data, &p.Player); err != nil {
		return err
	}

	p.Valid = true

	return nil
}

type ListPlayersResp struct {
	Data []Player `json:"data"`
} // @name ListPlayersResp

func NewListPlayersResp(players []domain.PlayerWithRelations) ListPlayersResp {
	resp := ListPlayersResp{
		Data: make([]Player, 0, len(players)),
	}
	for _, p := range players {
		resp.Data = append(resp.Data, NewPlayer(p))
	}
	return resp
}

type ListPlayerOtherServersResp struct {
	Data []PlayerWithServer `json:"data"`
} // @name ListPlayerOtherServersResp

func NewListPlayerOtherServersResp(players []domain.PlayerWithRelations) ListPlayerOtherServersResp {
	resp := ListPlayerOtherServersResp{
		Data: make([]PlayerWithServer, 0, len(players)),
	}
	for _, p := range players {
		resp.Data = append(resp.Data, NewPlayerWithServer(p))
	}
	return resp
}

type GetPlayerResp struct {
	Data Player `json:"data"`
} // @name GetPlayerResp

func NewGetPlayerResp(v domain.PlayerWithRelations) GetPlayerResp {
	return GetPlayerResp{
		Data: NewPlayer(v),
	}
}
