package model_test

import (
	"encoding/json"
	"testing"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
	"github.com/stretchr/testify/assert"
)

func TestNewPlayer_NewPlayerWithServer(t *testing.T) {
	player := domain.PlayerWithRelations{
		Player: domain.Player{
			BasePlayer: domain.BasePlayer{
				OpponentsDefeated: domain.OpponentsDefeated{
					RankAtt:    8,
					ScoreAtt:   7,
					RankDef:    6,
					ScoreDef:   5,
					RankSup:    4,
					ScoreSup:   3,
					RankTotal:  2,
					ScoreTotal: 1,
				},
				ID:          997,
				Name:        "name 997",
				NumVillages: 5,
				Points:      4,
				Rank:        2,
				TribeID:     0,
				ProfileURL:  "profile-997",
			},
			BestRank:       111,
			BestRankAt:     time.Now(),
			MostPoints:     1112,
			MostPointsAt:   time.Now().Add(-5 * time.Minute),
			MostVillages:   1113,
			MostVillagesAt: time.Now().Add(time.Minute),
			LastActivityAt: time.Now().Add(time.Second),
			ServerKey:      "pl151",
			CreatedAt:      time.Now(),
		},
		Server: domain.ServerMeta{
			Key:  "pl151",
			URL:  "https://pl151.plemiona.pl",
			Open: true,
		},
	}
	assertPlayer(t, player, model.NewPlayer(player))
	assertPlayerWithServer(t, player, model.NewPlayerWithServer(player))

	player = domain.PlayerWithRelations{
		Player: domain.Player{
			BasePlayer: domain.BasePlayer{
				OpponentsDefeated: domain.OpponentsDefeated{
					RankAtt:    8,
					ScoreAtt:   7,
					RankDef:    6,
					ScoreDef:   5,
					RankSup:    4,
					ScoreSup:   3,
					RankTotal:  2,
					ScoreTotal: 1,
				},
				ID:          997,
				Name:        "name 997",
				NumVillages: 5,
				Points:      4,
				Rank:        2,
				TribeID:     1234,
				ProfileURL:  "profile-997",
			},
			BestRank:       1117,
			BestRankAt:     time.Now(),
			MostPoints:     1115,
			MostPointsAt:   time.Now().Add(-5 * time.Second),
			MostVillages:   1114,
			MostVillagesAt: time.Now().Add(time.Hour),
			LastActivityAt: time.Now().Add(time.Minute),
			CreatedAt:      time.Now(),
			ServerKey:      "pl151",
		},
		Tribe: domain.NullTribeMeta{
			Valid: true,
			Tribe: domain.TribeMeta{
				ID:         1234,
				Name:       "name 997",
				Tag:        "tag 997",
				ProfileURL: "profile-997",
			},
		},
		Server: domain.ServerMeta{
			Key:  "pl151",
			URL:  "https://pl151.plemiona.pl",
			Open: true,
		},
	}
	assertPlayer(t, player, model.NewPlayer(player))
	assertPlayerWithServer(t, player, model.NewPlayerWithServer(player))
}

func TestNewPlayerMeta(t *testing.T) {
	playerWithoutTribe := domain.PlayerMetaWithRelations{
		PlayerMeta: domain.PlayerMeta{
			ID:         997,
			Name:       "name 997",
			ProfileURL: "profile-997",
		},
	}
	assertPlayerMeta(t, playerWithoutTribe, model.NewPlayerMeta(playerWithoutTribe))

	playerWithTribe := domain.PlayerMetaWithRelations{
		PlayerMeta: domain.PlayerMeta{
			ID:         997,
			Name:       "name 997",
			ProfileURL: "profile-997",
		},
		Tribe: domain.NullTribeMeta{
			Valid: true,
			Tribe: domain.TribeMeta{
				ID:         1234,
				Name:       "name 997",
				Tag:        "tag 997",
				ProfileURL: "profile-997",
			},
		},
	}
	assertPlayerMeta(t, playerWithTribe, model.NewPlayerMeta(playerWithTribe))
}

func TestNullPlayerMeta_MarshalJSON(t *testing.T) {
	t.Parallel()

	tests := []struct {
		name         string
		player       model.NullPlayerMeta
		expectedJSON string
	}{
		{
			name: "OK: null 1",
			player: model.NullPlayerMeta{
				Player: model.PlayerMeta{},
				Valid:  false,
			},
			expectedJSON: "null",
		},
		{
			name: "OK: null 2",
			player: model.NullPlayerMeta{
				Player: model.PlayerMeta{ID: 1234},
				Valid:  false,
			},
			expectedJSON: "null",
		},
		{
			name: "OK: valid struct",
			player: model.NullPlayerMeta{
				Player: model.PlayerMeta{
					ID:         997,
					Name:       "name 997",
					ProfileURL: "profile-997",
					Tribe: model.NullTribeMeta{
						Valid: true,
						Tribe: model.TribeMeta{
							ID:         1234,
							Name:       "name 997",
							ProfileURL: "profile-997",
							Tag:        "tag 997",
						},
					},
				},
				Valid: true,
			},
			//nolint:lll
			expectedJSON: `{"id":997,"name":"name 997","profileUrl":"profile-997","tribe":{"id":1234,"name":"name 997","profileUrl":"profile-997","tag":"tag 997"}}`,
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			b, err := json.Marshal(tt.player)
			assert.NoError(t, err)
			assert.JSONEq(t, tt.expectedJSON, string(b))
		})
	}
}

func TestNullPlayerMeta_UnmarshalJSON(t *testing.T) {
	t.Parallel()

	tests := []struct {
		name                    string
		json                    string
		expectedPlayer          model.NullPlayerMeta
		expectedJSONSyntaxError bool
	}{
		{
			name: "OK: null",
			json: "null",
			expectedPlayer: model.NullPlayerMeta{
				Player: model.PlayerMeta{},
				Valid:  false,
			},
		},
		{
			name: "OK: valid struct",
			//nolint:lll
			json: `{"id":997,"name":"name 997","profileUrl":"profile-997","tribe":{"id":1234,"name":"name 997","profileUrl":"profile-997","tag":"tag 997"}}`,
			expectedPlayer: model.NullPlayerMeta{
				Player: model.PlayerMeta{
					ID:         997,
					Name:       "name 997",
					ProfileURL: "profile-997",
					Tribe: model.NullTribeMeta{
						Valid: true,
						Tribe: model.TribeMeta{
							ID:         1234,
							Name:       "name 997",
							ProfileURL: "profile-997",
							Tag:        "tag 997",
						},
					},
				},
				Valid: true,
			},
		},
		{
			name:                    "ERR: invalid tribe 1",
			json:                    "2022-07-30T14:13:12.0000005Z",
			expectedPlayer:          model.NullPlayerMeta{},
			expectedJSONSyntaxError: true,
		},
		{
			name:                    "ERR: invalid tribe 2",
			json:                    "hello world",
			expectedPlayer:          model.NullPlayerMeta{},
			expectedJSONSyntaxError: true,
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			var target model.NullPlayerMeta
			err := json.Unmarshal([]byte(tt.json), &target)
			if tt.expectedJSONSyntaxError {
				var syntaxError *json.SyntaxError
				assert.ErrorAs(t, err, &syntaxError)
			} else {
				assert.NoError(t, err)
			}
			assert.Equal(t, tt.expectedPlayer, target)
		})
	}
}

func TestNewListPlayersResp_NewListPlayerOtherServersResp(t *testing.T) {
	t.Parallel()

	players := []domain.PlayerWithRelations{
		{
			Player: domain.Player{
				BasePlayer: domain.BasePlayer{
					OpponentsDefeated: domain.OpponentsDefeated{
						RankAtt:    8,
						ScoreAtt:   7,
						RankDef:    6,
						ScoreDef:   5,
						RankSup:    4,
						ScoreSup:   3,
						RankTotal:  2,
						ScoreTotal: 1,
					},
					ID:          997,
					Name:        "name 997",
					NumVillages: 5,
					Points:      4,
					Rank:        2,
					TribeID:     1234,
					ProfileURL:  "profile-997",
				},
				BestRank:       1117,
				BestRankAt:     time.Now(),
				MostPoints:     1115,
				MostPointsAt:   time.Now().Add(-5 * time.Second),
				MostVillages:   1114,
				MostVillagesAt: time.Now().Add(time.Hour),
				LastActivityAt: time.Now().Add(time.Minute),
				ServerKey:      "pl151",
				CreatedAt:      time.Now(),
			},
			Tribe: domain.NullTribeMeta{
				Valid: true,
				Tribe: domain.TribeMeta{
					ID:         1234,
					Name:       "name 997",
					Tag:        "tag 997",
					ProfileURL: "profile-997",
				},
			},
			Server: domain.ServerMeta{
				Key:  "pl151",
				URL:  "https://pl151.plemiona.pl",
				Open: true,
			},
		},
		{
			Player: domain.Player{
				BasePlayer: domain.BasePlayer{
					OpponentsDefeated: domain.OpponentsDefeated{
						RankAtt:    1,
						ScoreAtt:   2,
						RankDef:    3,
						ScoreDef:   4,
						RankSup:    5,
						ScoreSup:   6,
						RankTotal:  7,
						ScoreTotal: 8,
					},
					ID:          998,
					Name:        "name 998",
					NumVillages: 2,
					Points:      3,
					Rank:        4,
					TribeID:     0,
					ProfileURL:  "profile-998",
				},
				BestRank:       111,
				BestRankAt:     time.Now(),
				MostPoints:     1112,
				MostPointsAt:   time.Now().Add(-5 * time.Minute),
				MostVillages:   1113,
				MostVillagesAt: time.Now().Add(time.Minute),
				LastActivityAt: time.Now().Add(time.Second),
				ServerKey:      "pl151",
				CreatedAt:      time.Now(),
			},
			Tribe: domain.NullTribeMeta{
				Valid: false,
			},
			Server: domain.ServerMeta{
				Key:  "pl151",
				URL:  "https://pl151.plemiona.pl",
				Open: true,
			},
		},
	}

	resp1 := model.NewListPlayersResp(players)
	assert.Len(t, resp1.Data, len(players))
	for i, player := range resp1.Data {
		assertPlayer(t, players[i], player)
	}

	resp2 := model.NewListPlayerOtherServersResp(players)
	assert.Len(t, resp2.Data, len(players))
	for i, player := range resp2.Data {
		assertPlayerWithServer(t, players[i], player)
	}
}

func TestNewGetPlayerResp(t *testing.T) {
	t.Parallel()

	player := domain.PlayerWithRelations{
		Player: domain.Player{
			BasePlayer: domain.BasePlayer{
				OpponentsDefeated: domain.OpponentsDefeated{
					RankAtt:    8,
					ScoreAtt:   7,
					RankDef:    6,
					ScoreDef:   5,
					RankSup:    4,
					ScoreSup:   3,
					RankTotal:  2,
					ScoreTotal: 1,
				},
				ID:          997,
				Name:        "name 997",
				NumVillages: 5,
				Points:      4,
				Rank:        2,
				ProfileURL:  "profile-997",
			},
			BestRank:       1117,
			BestRankAt:     time.Now(),
			MostPoints:     1115,
			MostPointsAt:   time.Now().Add(-5 * time.Second),
			MostVillages:   1114,
			MostVillagesAt: time.Now().Add(time.Hour),
			LastActivityAt: time.Now().Add(time.Minute),
			ServerKey:      "pl151",
			CreatedAt:      time.Now(),
		},
	}
	assertPlayer(t, player, model.NewGetPlayerResp(player).Data)
}

func assertPlayer(tb testing.TB, dp domain.PlayerWithRelations, rp model.Player) {
	tb.Helper()

	assert.Equal(tb, dp.ID, rp.ID)
	assert.Equal(tb, dp.Name, rp.Name)
	assert.Equal(tb, dp.NumVillages, rp.NumVillages)
	assert.Equal(tb, dp.Points, rp.Points)
	assert.Equal(tb, dp.Rank, rp.Rank)
	assert.Equal(tb, dp.RankAtt, rp.RankAtt)
	assert.Equal(tb, dp.ScoreAtt, rp.ScoreAtt)
	assert.Equal(tb, dp.RankDef, rp.RankDef)
	assert.Equal(tb, dp.ScoreDef, rp.ScoreDef)
	assert.Equal(tb, dp.RankSup, rp.RankSup)
	assert.Equal(tb, dp.ScoreSup, rp.ScoreSup)
	assert.Equal(tb, dp.RankTotal, rp.RankTotal)
	assert.Equal(tb, dp.ScoreTotal, rp.ScoreTotal)
	assert.Equal(tb, dp.ProfileURL, rp.ProfileURL)
	assert.Equal(tb, dp.BestRank, rp.BestRank)
	assert.Equal(tb, dp.BestRankAt, rp.BestRankAt)
	assert.Equal(tb, dp.MostPoints, rp.MostPoints)
	assert.Equal(tb, dp.MostPointsAt, rp.MostPointsAt)
	assert.Equal(tb, dp.MostVillages, rp.MostVillages)
	assert.Equal(tb, dp.MostVillagesAt, rp.MostVillagesAt)
	assert.Equal(tb, dp.LastActivityAt, rp.LastActivityAt)
	assert.Equal(tb, dp.CreatedAt, rp.CreatedAt)
	assertNullTribeMeta(tb, dp.Tribe, rp.Tribe)
	assertNullTime(tb, dp.DeletedAt, rp.DeletedAt)
}

func assertPlayerWithServer(tb testing.TB, dp domain.PlayerWithRelations, rp model.PlayerWithServer) {
	tb.Helper()

	assertPlayer(tb, dp, rp.Player)
	assertServerMeta(tb, dp.Server, rp.Server)
}

func assertPlayerMeta(tb testing.TB, dp domain.PlayerMetaWithRelations, rp model.PlayerMeta) {
	tb.Helper()

	assert.Equal(tb, dp.ID, rp.ID)
	assert.Equal(tb, dp.Name, rp.Name)
	assert.Equal(tb, dp.ProfileURL, rp.ProfileURL)
	assertNullTribeMeta(tb, dp.Tribe, rp.Tribe)
}

func assertNullPlayerMeta(tb testing.TB, dp domain.NullPlayerMetaWithRelations, rp model.NullPlayerMeta) {
	tb.Helper()

	if !dp.Valid {
		assert.False(tb, rp.Valid)
		assert.Zero(tb, rp.Player)
		return
	}

	assert.True(tb, rp.Valid)
	assertPlayerMeta(tb, dp.Player, rp.Player)
}
