package model

import (
	"encoding/json"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
)

type Tribe struct {
	ID             int64     `json:"id"`
	Name           string    `json:"name"`
	Tag            string    `json:"tag"`
	NumMembers     int64     `json:"numMembers"`
	NumVillages    int64     `json:"numVillages"`
	Points         int64     `json:"points"`
	AllPoints      int64     `json:"allPoints"`
	Rank           int64     `json:"rank"`
	Dominance      float64   `json:"dominance"`
	RankAtt        int64     `json:"rankAtt"`
	ScoreAtt       int64     `json:"scoreAtt"`
	RankDef        int64     `json:"rankDef"`
	ScoreDef       int64     `json:"scoreDef"`
	RankTotal      int64     `json:"rankTotal"`
	ScoreTotal     int64     `json:"scoreTotal"`
	ProfileURL     string    `json:"profileUrl"`
	BestRank       int64     `json:"bestRank"`
	BestRankAt     time.Time `json:"bestRankAt" format:"date-time"`
	MostPoints     int64     `json:"mostPoints"`
	MostPointsAt   time.Time `json:"mostPointsAt" format:"date-time"`
	MostVillages   int64     `json:"mostVillages"`
	MostVillagesAt time.Time `json:"mostVillagesAt" format:"date-time"`
	CreatedAt      time.Time `json:"createdAt" format:"date-time"`
	DeletedAt      NullTime  `json:"deletedAt" swaggertype:"string" format:"date-time" extensions:"x-nullable"`
} // @name Tribe

func NewTribe(t domain.Tribe) Tribe {
	return Tribe{
		ID:             t.ID,
		Name:           t.Name,
		Tag:            t.Tag,
		NumMembers:     t.NumMembers,
		NumVillages:    t.NumVillages,
		Points:         t.Points,
		AllPoints:      t.AllPoints,
		Rank:           t.Rank,
		Dominance:      t.Dominance,
		RankAtt:        t.RankAtt,
		ScoreAtt:       t.ScoreAtt,
		RankDef:        t.RankDef,
		ScoreDef:       t.ScoreDef,
		RankTotal:      t.RankTotal,
		ScoreTotal:     t.ScoreTotal,
		ProfileURL:     t.ProfileURL,
		BestRank:       t.BestRank,
		BestRankAt:     t.BestRankAt,
		MostPoints:     t.MostPoints,
		MostPointsAt:   t.MostPointsAt,
		MostVillages:   t.MostVillages,
		MostVillagesAt: t.MostVillagesAt,
		CreatedAt:      t.CreatedAt,
		DeletedAt: NullTime{
			Time:  t.DeletedAt,
			Valid: !t.DeletedAt.IsZero(),
		},
	}
}

// TribeMeta represents basic tribe information
// @Description Basic tribe information
type TribeMeta struct {
	ID         int64  `json:"id"`
	Name       string `json:"name"`
	Tag        string `json:"tag"`
	ProfileURL string `json:"profileUrl"`
} // @name TribeMeta

func NewTribeMeta(t domain.TribeMeta) TribeMeta {
	return TribeMeta{
		ID:         t.ID,
		Name:       t.Name,
		ProfileURL: t.ProfileURL,
		Tag:        t.Tag,
	}
}

type NullTribeMeta struct {
	Tribe TribeMeta
	Valid bool
}

func (t NullTribeMeta) MarshalJSON() ([]byte, error) {
	if !t.Valid {
		return []byte("null"), nil
	}

	return json.Marshal(t.Tribe)
}

func (t *NullTribeMeta) UnmarshalJSON(data []byte) error {
	// Ignore null, like in the main JSON package.
	if string(data) == "null" {
		return nil
	}

	if err := json.Unmarshal(data, &t.Tribe); err != nil {
		return err
	}

	t.Valid = true

	return nil
}

type ListTribesResp struct {
	Data []Tribe `json:"data"`
} // @name ListTribesResp

func NewListTribesResp(tribes []domain.Tribe) ListTribesResp {
	resp := ListTribesResp{
		Data: make([]Tribe, 0, len(tribes)),
	}
	for _, t := range tribes {
		resp.Data = append(resp.Data, NewTribe(t))
	}
	return resp
}

type GetTribeResp struct {
	Data Tribe `json:"data"`
} // @name GetTribeResp

func NewGetTribeResp(v domain.Tribe) GetTribeResp {
	return GetTribeResp{
		Data: NewTribe(v),
	}
}
