package model_test

import (
	"encoding/json"
	"testing"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
	"github.com/stretchr/testify/assert"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
)

func TestNewTribe(t *testing.T) {
	t.Parallel()

	tribe := domain.Tribe{
		BaseTribe: domain.BaseTribe{
			OpponentsDefeated: domain.OpponentsDefeated{
				RankAtt:    8,
				ScoreAtt:   7,
				RankDef:    6,
				ScoreDef:   5,
				RankSup:    4,
				ScoreSup:   3,
				RankTotal:  2,
				ScoreTotal: 1,
			},
			ID:          997,
			Name:        "name 997",
			Tag:         "tag 997",
			NumMembers:  6,
			NumVillages: 5,
			Points:      4,
			AllPoints:   3,
			Rank:        2,
			ProfileURL:  "profile-997",
		},
		BestRank:       111,
		BestRankAt:     time.Now(),
		MostPoints:     1112,
		MostPointsAt:   time.Now().Add(-5 * time.Minute),
		MostVillages:   1113,
		MostVillagesAt: time.Now().Add(time.Minute),
		ServerKey:      "pl151",
		Dominance:      12.5,
		CreatedAt:      time.Now(),
	}
	assertTribe(t, tribe, model.NewTribe(tribe))
}

func TestNewTribeMeta(t *testing.T) {
	t.Parallel()

	meta := domain.TribeMeta{
		ID:         123,
		Name:       "Name",
		Tag:        "Tag",
		ProfileURL: "ProfileURL",
	}
	assertTribeMeta(t, meta, model.NewTribeMeta(meta))
}

func TestNullTribeMeta_MarshalJSON(t *testing.T) {
	t.Parallel()

	tests := []struct {
		name         string
		tribe        model.NullTribeMeta
		expectedJSON string
	}{
		{
			name: "OK: null 1",
			tribe: model.NullTribeMeta{
				Tribe: model.TribeMeta{},
				Valid: false,
			},
			expectedJSON: "null",
		},
		{
			name: "OK: null 2",
			tribe: model.NullTribeMeta{
				Tribe: model.TribeMeta{ID: 1234},
				Valid: false,
			},
			expectedJSON: "null",
		},
		{
			name: "OK: valid struct",
			tribe: model.NullTribeMeta{
				Tribe: model.TribeMeta{
					ID:         997,
					Name:       "name 997",
					Tag:        "tag 997",
					ProfileURL: "profile-997",
				},
				Valid: true,
			},
			expectedJSON: `{"id":997,"name":"name 997","tag":"tag 997","profileUrl":"profile-997"}`,
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			b, err := json.Marshal(tt.tribe)
			assert.NoError(t, err)
			assert.JSONEq(t, tt.expectedJSON, string(b))
		})
	}
}

func TestNullTribeMeta_UnmarshalJSON(t *testing.T) {
	t.Parallel()

	tests := []struct {
		name                    string
		json                    string
		expectedTribe           model.NullTribeMeta
		expectedJSONSyntaxError bool
	}{
		{
			name: "OK: null",
			json: "null",
			expectedTribe: model.NullTribeMeta{
				Tribe: model.TribeMeta{},
				Valid: false,
			},
		},
		{
			name: "OK: valid struct",
			//nolint:lll
			json: `{"id":997,"name":"name 997","tag":"tag 997","profileUrl":"profile-997"}`,
			expectedTribe: model.NullTribeMeta{
				Tribe: model.TribeMeta{
					ID:         997,
					Name:       "name 997",
					Tag:        "tag 997",
					ProfileURL: "profile-997",
				},
				Valid: true,
			},
		},
		{
			name:                    "ERR: invalid tribe 1",
			json:                    "2022-07-30T14:13:12.0000005Z",
			expectedTribe:           model.NullTribeMeta{},
			expectedJSONSyntaxError: true,
		},
		{
			name:                    "ERR: invalid tribe 2",
			json:                    "hello world",
			expectedTribe:           model.NullTribeMeta{},
			expectedJSONSyntaxError: true,
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			var target model.NullTribeMeta
			err := json.Unmarshal([]byte(tt.json), &target)
			if tt.expectedJSONSyntaxError {
				var syntaxError *json.SyntaxError
				assert.ErrorAs(t, err, &syntaxError)
			} else {
				assert.NoError(t, err)
			}
			assert.Equal(t, tt.expectedTribe, target)
		})
	}
}

func TestNewListTribesResp(t *testing.T) {
	t.Parallel()

	tribes := []domain.Tribe{
		{
			BaseTribe: domain.BaseTribe{
				OpponentsDefeated: domain.OpponentsDefeated{
					RankAtt:    8,
					ScoreAtt:   7,
					RankDef:    6,
					ScoreDef:   5,
					RankSup:    4,
					ScoreSup:   3,
					RankTotal:  2,
					ScoreTotal: 1,
				},
				ID:          997,
				Name:        "name 997",
				Tag:         "tag 997",
				NumMembers:  6,
				NumVillages: 5,
				Points:      4,
				AllPoints:   3,
				Rank:        2,
				ProfileURL:  "profile-997",
			},
			BestRank:       111,
			BestRankAt:     time.Now(),
			MostPoints:     1112,
			MostPointsAt:   time.Now().Add(-5 * time.Minute),
			MostVillages:   1113,
			MostVillagesAt: time.Now().Add(time.Minute),
			ServerKey:      "pl151",
			Dominance:      12.5,
			CreatedAt:      time.Now(),
		},
		{
			BaseTribe: domain.BaseTribe{
				OpponentsDefeated: domain.OpponentsDefeated{
					RankAtt:    1,
					ScoreAtt:   2,
					RankDef:    3,
					ScoreDef:   4,
					RankSup:    5,
					ScoreSup:   6,
					RankTotal:  7,
					ScoreTotal: 9,
				},
				ID:          998,
				Name:        "name 998",
				Tag:         "tag 998",
				NumMembers:  1,
				NumVillages: 2,
				Points:      3,
				AllPoints:   4,
				Rank:        5,
				ProfileURL:  "profile-998",
			},
			BestRank:       1113,
			BestRankAt:     time.Now().Add(time.Second),
			MostPoints:     1114,
			MostPointsAt:   time.Now().Add(-10 * time.Minute),
			MostVillages:   1115,
			MostVillagesAt: time.Now().Add(time.Hour),
			ServerKey:      "pl163",
			Dominance:      12.55,
			CreatedAt:      time.Now().Add(-5000 * time.Hour),
			DeletedAt:      time.Now(),
		},
	}
	resp := model.NewListTribesResp(tribes)
	assert.Len(t, resp.Data, len(tribes))
	for i, tribe := range resp.Data {
		assertTribe(t, tribes[i], tribe)
	}
}

func TestNewGetTribeResp(t *testing.T) {
	t.Parallel()

	tribe := domain.Tribe{
		BaseTribe: domain.BaseTribe{
			OpponentsDefeated: domain.OpponentsDefeated{
				RankAtt:    8,
				ScoreAtt:   7,
				RankDef:    6,
				ScoreDef:   5,
				RankSup:    4,
				ScoreSup:   3,
				RankTotal:  2,
				ScoreTotal: 1,
			},
			ID:          997,
			Name:        "name 997",
			Tag:         "tag 997",
			NumMembers:  6,
			NumVillages: 5,
			Points:      4,
			AllPoints:   3,
			Rank:        2,
			ProfileURL:  "profile-997",
		},
		BestRank:       111,
		BestRankAt:     time.Now(),
		MostPoints:     1112,
		MostPointsAt:   time.Now().Add(-5 * time.Minute),
		MostVillages:   1113,
		MostVillagesAt: time.Now().Add(time.Minute),
		ServerKey:      "pl151",
		Dominance:      12.5,
		CreatedAt:      time.Now(),
	}
	assertTribe(t, tribe, model.NewGetTribeResp(tribe).Data)
}

func assertTribe(tb testing.TB, dt domain.Tribe, rt model.Tribe) {
	tb.Helper()

	assert.Equal(tb, dt.ID, rt.ID)
	assert.Equal(tb, dt.Name, rt.Name)
	assert.Equal(tb, dt.Tag, rt.Tag)
	assert.Equal(tb, dt.NumMembers, rt.NumMembers)
	assert.Equal(tb, dt.NumVillages, rt.NumVillages)
	assert.Equal(tb, dt.Points, rt.Points)
	assert.Equal(tb, dt.AllPoints, rt.AllPoints)
	assert.Equal(tb, dt.Rank, rt.Rank)
	assert.Equal(tb, dt.Dominance, rt.Dominance)
	assert.Equal(tb, dt.RankAtt, rt.RankAtt)
	assert.Equal(tb, dt.ScoreAtt, rt.ScoreAtt)
	assert.Equal(tb, dt.RankDef, rt.RankDef)
	assert.Equal(tb, dt.ScoreDef, rt.ScoreDef)
	assert.Equal(tb, dt.RankTotal, rt.RankTotal)
	assert.Equal(tb, dt.ScoreTotal, rt.ScoreTotal)
	assert.Equal(tb, dt.ProfileURL, rt.ProfileURL)
	assert.Equal(tb, dt.BestRank, rt.BestRank)
	assert.Equal(tb, dt.BestRankAt, rt.BestRankAt)
	assert.Equal(tb, dt.MostPoints, rt.MostPoints)
	assert.Equal(tb, dt.MostPointsAt, rt.MostPointsAt)
	assert.Equal(tb, dt.MostVillages, rt.MostVillages)
	assert.Equal(tb, dt.MostVillagesAt, rt.MostVillagesAt)
	assert.Equal(tb, dt.CreatedAt, rt.CreatedAt)
	assertNullTime(tb, dt.DeletedAt, rt.DeletedAt)
}

func assertTribeMeta(tb testing.TB, dt domain.TribeMeta, rt model.TribeMeta) {
	tb.Helper()

	assert.Equal(tb, dt.ID, rt.ID)
	assert.Equal(tb, dt.Name, rt.Name)
	assert.Equal(tb, dt.Tag, rt.Tag)
	assert.Equal(tb, dt.ProfileURL, rt.ProfileURL)
}

func assertNullTribeMeta(tb testing.TB, dt domain.NullTribeMeta, rt model.NullTribeMeta) {
	tb.Helper()

	if !dt.Valid {
		assert.False(tb, rt.Valid)
		assert.Zero(tb, rt.Tribe)
		return
	}

	assert.True(tb, rt.Valid)
	assertTribeMeta(tb, dt.Tribe, rt.Tribe)
}
