package model_test

import (
	"testing"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
	"github.com/stretchr/testify/assert"
)

func TestNewVillage(t *testing.T) {
	t.Parallel()

	barbarianVillage := domain.VillageWithRelations{
		Village: domain.Village{
			BaseVillage: domain.BaseVillage{
				ID:         1234,
				Name:       "name",
				Points:     1,
				X:          2,
				Y:          3,
				Continent:  "K11",
				Bonus:      1,
				PlayerID:   0,
				ProfileURL: "profile-1234",
			},
			ServerKey: "pl151",
			CreatedAt: time.Now(),
		},
		Player: domain.NullPlayerMetaWithRelations{},
	}
	assertVillage(t, barbarianVillage, model.NewVillage(barbarianVillage))

	playerVillage := domain.VillageWithRelations{
		Village: domain.Village{
			BaseVillage: domain.BaseVillage{
				ID:         1234,
				Name:       "name",
				Points:     1,
				X:          2,
				Y:          3,
				Continent:  "K11",
				Bonus:      1,
				PlayerID:   997,
				ProfileURL: "profile-1234",
			},
			ServerKey: "pl151",
			CreatedAt: time.Now(),
		},
		Player: domain.NullPlayerMetaWithRelations{
			Valid: true,
			Player: domain.PlayerMetaWithRelations{
				PlayerMeta: domain.PlayerMeta{
					ID:         997,
					Name:       "name 997",
					ProfileURL: "profile-997",
				},
				Tribe: domain.NullTribeMeta{
					Valid: true,
					Tribe: domain.TribeMeta{
						ID:         1234,
						Name:       "name 997",
						Tag:        "tag 997",
						ProfileURL: "profile-1234",
					},
				},
			},
		},
	}
	assertVillage(t, playerVillage, model.NewVillage(playerVillage))
}

func TestNewVillageMeta(t *testing.T) {
	t.Parallel()

	village := domain.VillageMetaWithRelations{
		VillageMeta: domain.VillageMeta{
			ID:         1234,
			Name:       "name",
			X:          2,
			Y:          3,
			Continent:  "K11",
			ProfileURL: "profile-1234",
		},
		Player: domain.NullPlayerMetaWithRelations{
			Valid: true,
			Player: domain.PlayerMetaWithRelations{
				PlayerMeta: domain.PlayerMeta{
					ID:         1234,
					Name:       "name-1234",
					ProfileURL: "profile-1234",
				},
				Tribe: domain.NullTribeMeta{
					Valid: true,
					Tribe: domain.TribeMeta{
						ID:         1555,
						Name:       "name-1555",
						Tag:        "tag-1555",
						ProfileURL: "profile-1555",
					},
				},
			},
		},
	}
	assertVillageMeta(t, village, model.NewVillageMeta(village))
}

func TestNewListVillagesResp(t *testing.T) {
	t.Parallel()

	villages := []domain.VillageWithRelations{
		{
			Village: domain.Village{
				BaseVillage: domain.BaseVillage{
					ID:         1234,
					Name:       "name",
					Points:     1,
					X:          2,
					Y:          3,
					Continent:  "K11",
					Bonus:      1,
					PlayerID:   997,
					ProfileURL: "profile-1234",
				},
				ServerKey: "pl151",
				CreatedAt: time.Now(),
			},
			Player: domain.NullPlayerMetaWithRelations{
				Valid: true,
				Player: domain.PlayerMetaWithRelations{
					PlayerMeta: domain.PlayerMeta{
						ID:         997,
						Name:       "name 997",
						ProfileURL: "profile-997",
					},
					Tribe: domain.NullTribeMeta{
						Valid: true,
						Tribe: domain.TribeMeta{
							ID:         1234,
							Name:       "name 997",
							Tag:        "tag 997",
							ProfileURL: "profile-1234",
						},
					},
				},
			},
		},
		{
			Village: domain.Village{
				BaseVillage: domain.BaseVillage{
					ID:         1234,
					Name:       "name",
					Points:     1,
					X:          2,
					Y:          3,
					Continent:  "K11",
					Bonus:      1,
					PlayerID:   0,
					ProfileURL: "profile-1234",
				},
				ServerKey: "pl151",
				CreatedAt: time.Now(),
			},
		},
	}
	resp := model.NewListVillagesResp(villages)
	assert.Len(t, resp.Data, len(villages))
	for i, village := range resp.Data {
		assertVillage(t, villages[i], village)
	}
}

func TestNewGetVillageResp(t *testing.T) {
	t.Parallel()

	village := domain.VillageWithRelations{
		Village: domain.Village{
			BaseVillage: domain.BaseVillage{
				ID:         1234,
				Name:       "name",
				Points:     1,
				X:          2,
				Y:          3,
				Continent:  "K11",
				Bonus:      1,
				PlayerID:   0,
				ProfileURL: "profile-1234",
			},
			ServerKey: "pl151",
			CreatedAt: time.Now(),
		},

		Player: domain.NullPlayerMetaWithRelations{},
	}
	assertVillage(t, village, model.NewGetVillageResp(village).Data)
}

func assertVillage(tb testing.TB, dv domain.VillageWithRelations, rv model.Village) {
	tb.Helper()

	assert.Equal(tb, dv.ID, rv.ID)
	assert.Equal(tb, dv.Name, rv.Name)
	assert.Equal(tb, dv.Points, rv.Points)
	assert.Equal(tb, dv.X, rv.X)
	assert.Equal(tb, dv.Y, rv.Y)
	assert.Equal(tb, dv.Continent, rv.Continent)
	assert.Equal(tb, dv.Bonus, rv.Bonus)
	assert.Equal(tb, dv.ProfileURL, rv.ProfileURL)
	assert.Equal(tb, dv.CreatedAt, rv.CreatedAt)
	assertNullPlayerMeta(tb, dv.Player, rv.Player)
}

func assertVillageMeta(tb testing.TB, dv domain.VillageMetaWithRelations, rv model.VillageMeta) {
	tb.Helper()

	assert.Equal(tb, dv.ID, rv.ID)
	assert.Equal(tb, dv.FullName(), rv.FullName)
	assert.Equal(tb, dv.X, rv.X)
	assert.Equal(tb, dv.Y, rv.Y)
	assert.Equal(tb, dv.Continent, rv.Continent)
	assert.Equal(tb, dv.ProfileURL, rv.ProfileURL)
	assertNullPlayerMeta(tb, dv.Player, rv.Player)
}
