package rest

import (
	"context"
	"net/http"

	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
	"github.com/go-chi/chi/v5"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
)

type playerCtxKey struct{}

//counterfeiter:generate -o internal/mock/player_service.gen.go . PlayerService
type PlayerService interface {
	ListCountWithRelations(ctx context.Context, params domain.ListPlayersParams) ([]domain.PlayerWithRelations, int64, error)
	GetByServerKeyAndIDWithRelations(ctx context.Context, serverKey string, id int64) (domain.PlayerWithRelations, error)
}

type player struct {
	svc PlayerService
}

// @ID          listPlayers
// @Summary     List players
// @Description List all players
// @Tags        versions,servers,players
// @Produce     json
// @Success     200         {object}  model.ListPlayersResp
// @Success     400         {object}  model.ErrorResp
// @Success     404         {object}  model.ErrorResp
// @Failure     500         {object}  model.ErrorResp
// @Header      200         {integer} X-Total-Count "Total number of records"
// @Param       versionCode path      string        true  "Version code"
// @Param       serverKey   path      string        true  "Server key"
// @Param       deleted     query     boolean       false "true=only deleted players, false=only existing players, by default both existing and deleted players are returned"
// @Param       offset      query     int           false "specifies where to start a page" minimum(0) default(0)
// @Param       limit       query     int           false "page size"                       minimum(1) maximum(200) default(200)
// @Param       name        query     []string      false "player names, max 20"
// @Param       sort        query     []string      false "format: field:direction, default: [id:asc]" Enums(id:asc,id:desc,scoreAtt:asc,scoreAtt:desc,scoreDef:asc,scoreDef:desc,scoreSup:asc,scoreSup:desc,scoreTotal:asc,scoreTotal:desc,points:asc,points:desc,deletedAt:asc,deletedAt:desc)
// @Router      /versions/{versionCode}/servers/{serverKey}/players [get]
func (p *player) list(w http.ResponseWriter, r *http.Request) {
	var err error
	ctx := r.Context()
	srv, _ := serverFromContext(ctx)
	params := domain.ListPlayersParams{
		ServerKeys: []string{srv.Key},
	}
	query := queryParams[domain.PlayerSort]{
		values:  r.URL.Query(),
		newSort: domain.NewPlayerSort,
	}

	params.Names = query.strings("name")

	params.Deleted, err = query.nullBool("deleted")
	if err != nil {
		renderErr(w, err)
		return
	}

	params.Pagination, err = query.pagination()
	if err != nil {
		renderErr(w, err)
		return
	}

	params.Sort, err = query.sort()
	if err != nil {
		renderErr(w, err)
		return
	}

	players, count, err := p.svc.ListCountWithRelations(ctx, params)
	if err != nil {
		renderErr(w, err)
		return
	}

	setTotalCountHeader(w.Header(), count)
	renderJSON(w, http.StatusOK, model.NewListPlayersResp(players))
}

// @ID          getPlayer
// @Summary     Get a player
// @Description Get a player
// @Tags        versions,servers,players
// @Produce     json
// @Success     200         {object} model.GetPlayerResp
// @Failure     400         {object} model.ErrorResp
// @Failure     404         {object} model.ErrorResp
// @Failure     500         {object} model.ErrorResp
// @Param       versionCode path     string  true "Version code"
// @Param       serverKey   path     string  true "Server key"
// @Param       playerId    path     integer true "Player ID"
// @Router      /versions/{versionCode}/servers/{serverKey}/players/{playerId} [get]
func (p *player) getByID(w http.ResponseWriter, r *http.Request) {
	pl, _ := playerFromContext(r.Context())
	renderJSON(w, http.StatusOK, model.NewGetPlayerResp(pl))
}

// @ID          listPlayerOtherServers
// @Summary     List other servers of a particular player
// @Description List other servers of a particular player
// @Tags        versions,servers,players
// @Produce     json
// @Success     200         {object} model.ListPlayerOtherServersResp
// @Failure     400         {object} model.ErrorResp
// @Failure     404         {object} model.ErrorResp
// @Failure     500         {object} model.ErrorResp
// @Param       versionCode path     string  true  "Version code"
// @Param       serverKey   path     string  true  "Server key"
// @Param       playerId    path     integer true  "Player ID"
// @Param       offset      query    int     false "specifies where to start a page" minimum(0) default(0)
// @Param       limit       query    int     false "page size"                       minimum(1) maximum(200) default(200)
// @Router      /versions/{versionCode}/servers/{serverKey}/players/{playerId}/other-servers [get]
func (p *player) listOtherServers(w http.ResponseWriter, r *http.Request) {
	var err error
	ctx := r.Context()
	srv, _ := serverFromContext(ctx)
	pl, _ := playerFromContext(ctx)
	params := domain.ListPlayersParams{
		VersionCodes:  []string{srv.VersionCode},
		ServerKeysNIN: []string{srv.Key},
		IDs:           []int64{pl.ID},
	}
	query := queryParams[domain.PlayerSort]{
		values: r.URL.Query(),
	}

	params.Pagination, err = query.pagination()
	if err != nil {
		renderErr(w, err)
		return
	}

	players, count, err := p.svc.ListCountWithRelations(ctx, params)
	if err != nil {
		renderErr(w, err)
		return
	}

	setTotalCountHeader(w.Header(), count)
	renderJSON(w, http.StatusOK, model.NewListPlayerOtherServersResp(players))
}

// @ID          listTribeMembers
// @Summary     List tribe members
// @Description List all tribe members
// @Tags        versions,servers,players,tribes
// @Produce     json
// @Success     200         {object}  model.ListPlayersResp
// @Success     400         {object}  model.ErrorResp
// @Success     404         {object}  model.ErrorResp
// @Failure     500         {object}  model.ErrorResp
// @Header      200         {integer} X-Total-Count "Total number of records"
// @Param       versionCode path      string        true  "Version code"
// @Param       serverKey   path      string        true  "Server key"
// @Param       tribeId     path      integer       true  "Tribe ID"
// @Param       offset      query     int           false "specifies where to start a page"            minimum(0) default(0)
// @Param       limit       query     int           false "page size"                                  minimum(1) maximum(200) default(200)
// @Param       sort        query     []string      false "format: field:direction, default: [id:asc]" Enums(id:asc,id:desc,scoreAtt:asc,scoreAtt:desc,scoreDef:asc,scoreDef:desc,scoreSup:asc,scoreSup:desc,scoreTotal:asc,scoreTotal:desc,points:asc,points:desc,deletedAt:asc,deletedAt:desc)
// @Router      /versions/{versionCode}/servers/{serverKey}/tribes/{tribeId}/members [get]
func (p *player) listTribeMembers(w http.ResponseWriter, r *http.Request) {
	var err error
	ctx := r.Context()
	srv, _ := serverFromContext(ctx)
	trb, _ := tribeFromContext(ctx)
	params := domain.ListPlayersParams{
		ServerKeys: []string{srv.Key},
		TribeID: domain.NullInt64{
			Valid: true,
			Int64: trb.ID,
		},
	}
	query := queryParams[domain.PlayerSort]{
		values:  r.URL.Query(),
		newSort: domain.NewPlayerSort,
	}

	params.Pagination, err = query.pagination()
	if err != nil {
		renderErr(w, err)
		return
	}

	params.Sort, err = query.sort()
	if err != nil {
		renderErr(w, err)
		return
	}

	players, count, err := p.svc.ListCountWithRelations(ctx, params)
	if err != nil {
		renderErr(w, err)
		return
	}

	setTotalCountHeader(w.Header(), count)
	renderJSON(w, http.StatusOK, model.NewListPlayersResp(players))
}

func playerMiddleware(svc PlayerService) func(http.Handler) http.Handler {
	return func(next http.Handler) http.Handler {
		return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
			ctx := r.Context()
			srv, _ := serverFromContext(ctx)
			routeCtx := chi.RouteContext(ctx)

			playerID, err := urlParamInt64(routeCtx, "playerId")
			if err != nil {
				renderErr(w, err)
				return
			}

			pl, err := svc.GetByServerKeyAndIDWithRelations(ctx, srv.Key, playerID)
			if err != nil {
				renderErr(w, err)
				return
			}

			next.ServeHTTP(w, r.WithContext(playerToContext(ctx, pl)))
		})
	}
}

func playerToContext(ctx context.Context, pl domain.PlayerWithRelations) context.Context {
	return context.WithValue(ctx, playerCtxKey{}, pl)
}

func playerFromContext(ctx context.Context) (domain.PlayerWithRelations, bool) {
	pl, ok := ctx.Value(playerCtxKey{}).(domain.PlayerWithRelations)
	return pl, ok
}
