package rest

import (
	"context"
	"net/http"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
)

//counterfeiter:generate -o internal/mock/player_snapshot_service.gen.go . PlayerSnapshotService
type PlayerSnapshotService interface {
	ListCountWithRelations(ctx context.Context, params domain.ListPlayerSnapshotsParams) ([]domain.PlayerSnapshotWithRelations, int64, error)
}

type playerSnapshot struct {
	svc PlayerSnapshotService
}

// @ID          listPlayerSnapshotsDeprecated
// @Summary     Deprecated, use /versions/{versionCode}/servers/{serverKey}/players/{playerId}/snapshots instead.
// @Description Deprecated, use /versions/{versionCode}/servers/{serverKey}/players/{playerId}/snapshots instead.
// @Tags        versions,servers,players,snapshots
// @Produce     json
// @Success     200         {object}  model.ListPlayerSnapshotsResp
// @Success     400         {object}  model.ErrorResp
// @Success     404         {object}  model.ErrorResp
// @Failure     500         {object}  model.ErrorResp
// @Header      200         {integer} X-Total-Count "Total number of records"
// @Param       versionCode path      string        true  "Version code"
// @Param       serverKey   path      string        true  "Server key"
// @Param       playerId    path      integer       true  "Player ID"
// @Param       offset      query     int           false "specifies where to start a page"                     minimum(0) default(0)
// @Param       limit       query     int           false "page size"                                           minimum(1) maximum(100) default(100)
// @Param       sort        query     []string      false "format: field:direction, default: [date:asc,id:asc]" Enums(id:asc,id:desc,date:asc,date:desc)
// @Router      /versions/{versionCode}/servers/{serverKey}/players/{playerId}/history [get]
// @Deprecated
func (p *playerSnapshot) history(w http.ResponseWriter, r *http.Request) {
	p.list(w, r)
}

// @ID          listPlayerSnapshots
// @Summary     List snapshots of a particular player
// @Description List snapshots of a particular player
// @Tags        versions,servers,players,snapshots
// @Produce     json
// @Success     200         {object}  model.ListPlayerSnapshotsResp
// @Success     400         {object}  model.ErrorResp
// @Success     404         {object}  model.ErrorResp
// @Failure     500         {object}  model.ErrorResp
// @Header      200         {integer} X-Total-Count "Total number of records"
// @Param       versionCode path      string        true  "Version code"
// @Param       serverKey   path      string        true  "Server key"
// @Param       playerId    path      integer       true  "Player ID"
// @Param       offset      query     int           false "specifies where to start a page"                     minimum(0) default(0)
// @Param       limit       query     int           false "page size"                                           minimum(1) maximum(100) default(100)
// @Param       sort        query     []string      false "format: field:direction, default: [date:asc,id:asc]" Enums(id:asc,id:desc,date:asc,date:desc)
// @Router      /versions/{versionCode}/servers/{serverKey}/players/{playerId}/snapshots [get]
func (p *playerSnapshot) list(w http.ResponseWriter, r *http.Request) {
	var err error
	ctx := r.Context()
	srv, _ := serverFromContext(ctx)

	params := domain.ListPlayerSnapshotsParams{
		ServerKeys: []string{srv.Key},
	}
	query := queryParams[domain.PlayerSnapshotSort]{
		values:  r.URL.Query(),
		newSort: domain.NewPlayerSnapshotSort,
	}

	pl, _ := playerFromContext(r.Context())
	params.PlayerIDs = []int64{pl.ID}

	params.Pagination, err = query.pagination()
	if err != nil {
		renderErr(w, err)
		return
	}

	params.Sort, err = query.sort()
	if err != nil {
		renderErr(w, err)
		return
	}

	snapshots, count, err := p.svc.ListCountWithRelations(ctx, params)
	if err != nil {
		renderErr(w, err)
		return
	}

	setTotalCountHeader(w.Header(), count)
	renderJSON(w, http.StatusOK, model.NewListPlayerSnapshotsResp(snapshots))
}
