package rest_test

import (
	"context"
	"fmt"
	"net/http"
	"net/url"
	"strconv"
	"testing"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/mock"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
	"github.com/google/go-cmp/cmp"
)

func TestPlayerSnapshot_list(t *testing.T) {
	t.Parallel()

	now := time.Now()
	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	server := domain.Server{
		Key:         "pl151",
		URL:         "https://pl151.plemiona.pl",
		Open:        true,
		Special:     false,
		VersionCode: "pl",
	}
	player := domain.PlayerWithRelations{
		Player: domain.Player{
			BasePlayer: domain.BasePlayer{
				ID:         124,
				Name:       "player 124",
				ProfileURL: "player-124",
			},
		},
	}
	playerID := strconv.FormatInt(player.ID, 10)

	tests := []struct {
		name  string
		setup func(
			versionSvc *mock.FakeVersionService,
			serverSvc *mock.FakeServerService,
			playerSvc *mock.FakePlayerService,
			playerSnapshotSvc *mock.FakePlayerSnapshotService,
		)
		versionCode        string
		serverKey          string
		playerID           string
		queryParams        url.Values
		expectedStatus     int
		target             any
		expectedResponse   any
		expectedTotalCount string
	}{
		{
			name: "OK: without params",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				playerSnapshotSvc *mock.FakePlayerSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
				playerSnapshotSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListPlayerSnapshotsParams,
				) ([]domain.PlayerSnapshotWithRelations, int64, error) {
					expectedParams := domain.ListPlayerSnapshotsParams{
						ServerKeys: []string{server.Key},
						PlayerIDs:  []int64{player.ID},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					snapshots := []domain.PlayerSnapshotWithRelations{
						{
							PlayerSnapshot: domain.PlayerSnapshot{
								OpponentsDefeated: domain.OpponentsDefeated{
									RankAtt:    99,
									ScoreAtt:   100,
									RankDef:    101,
									ScoreDef:   102,
									RankSup:    103,
									ScoreSup:   104,
									RankTotal:  105,
									ScoreTotal: 106,
								},
								ID:          999,
								PlayerID:    player.ID,
								ServerKey:   server.Key,
								NumVillages: 100,
								Points:      101,
								Rank:        102,
								TribeID:     15,
								Date:        now,
								CreatedAt:   now.Add(-5 * time.Minute),
							},
							Tribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         15,
									Name:       "Name",
									Tag:        "Tag",
									ProfileURL: "profile-15",
								},
							},
						},
						{
							PlayerSnapshot: domain.PlayerSnapshot{
								OpponentsDefeated: domain.OpponentsDefeated{
									RankAtt:    999,
									ScoreAtt:   1000,
									RankDef:    1010,
									ScoreDef:   1020,
									RankSup:    1030,
									ScoreSup:   1040,
									RankTotal:  1050,
									ScoreTotal: 1060,
								},
								ID:          9990,
								PlayerID:    player.ID,
								ServerKey:   server.Key,
								NumVillages: 1000,
								Points:      1010,
								Rank:        1020,
								TribeID:     0,
								Date:        now.Add(-24 * time.Hour),
								CreatedAt:   now.Add(-36 * time.Hour),
							},
							Tribe: domain.NullTribeMeta{
								Valid: false,
							},
						},
					}
					return snapshots, int64(len(snapshots)), nil
				})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			playerID:       playerID,
			queryParams:    nil,
			expectedStatus: http.StatusOK,
			target:         &model.ListPlayerSnapshotsResp{},
			expectedResponse: &model.ListPlayerSnapshotsResp{
				Data: []model.PlayerSnapshot{
					{
						ID:          999,
						NumVillages: 100,
						Points:      101,
						Rank:        102,
						RankAtt:     99,
						ScoreAtt:    100,
						RankDef:     101,
						ScoreDef:    102,
						RankSup:     103,
						ScoreSup:    104,
						RankTotal:   105,
						ScoreTotal:  106,
						Tribe: model.NullTribeMeta{
							Tribe: model.TribeMeta{
								ID:         15,
								Name:       "Name",
								Tag:        "Tag",
								ProfileURL: "profile-15",
							},
							Valid: true,
						},
						Date: model.Date{Time: now.Truncate(24 * time.Hour)},
					},
					{
						ID:          9990,
						NumVillages: 1000,
						Points:      1010,
						Rank:        1020,
						RankAtt:     999,
						ScoreAtt:    1000,
						RankDef:     1010,
						ScoreDef:    1020,
						RankSup:     1030,
						ScoreSup:    1040,
						RankTotal:   1050,
						ScoreTotal:  1060,
						Date:        model.Date{Time: now.Add(-24 * time.Hour).Truncate(24 * time.Hour)},
					},
				},
			},
			expectedTotalCount: "2",
		},
		{
			name: "OK: limit=1,offset=100,sort=[date:desc,id:desc]",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				playerSnapshotSvc *mock.FakePlayerSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
				playerSnapshotSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListPlayerSnapshotsParams,
				) ([]domain.PlayerSnapshotWithRelations, int64, error) {
					expectedParams := domain.ListPlayerSnapshotsParams{
						ServerKeys: []string{server.Key},
						PlayerIDs:  []int64{player.ID},
						Pagination: domain.Pagination{
							Limit:  1,
							Offset: 100,
						},
						Sort: []domain.PlayerSnapshotSort{
							{By: domain.PlayerSnapshotSortByDate, Direction: domain.SortDirectionDESC},
							{By: domain.PlayerSnapshotSortByID, Direction: domain.SortDirectionDESC},
						},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					snapshots := []domain.PlayerSnapshotWithRelations{
						{
							PlayerSnapshot: domain.PlayerSnapshot{
								OpponentsDefeated: domain.OpponentsDefeated{
									RankAtt:    999,
									ScoreAtt:   1000,
									RankDef:    1010,
									ScoreDef:   1020,
									RankSup:    1030,
									ScoreSup:   1040,
									RankTotal:  1050,
									ScoreTotal: 1060,
								},
								ID:          9990,
								PlayerID:    player.ID,
								ServerKey:   server.Key,
								NumVillages: 1000,
								Points:      1010,
								Rank:        1020,
								TribeID:     0,
								Date:        now.Add(-24 * time.Hour),
								CreatedAt:   now.Add(-36 * time.Hour),
							},
							Tribe: domain.NullTribeMeta{
								Valid: false,
							},
						},
					}
					return snapshots, int64(len(snapshots)) + int64(params.Pagination.Offset), nil
				})
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"limit":  []string{"1"},
				"offset": []string{"100"},
				"sort":   []string{"date:desc", "id:desc"},
			},
			expectedStatus: http.StatusOK,
			target:         &model.ListPlayerSnapshotsResp{},
			expectedResponse: &model.ListPlayerSnapshotsResp{
				Data: []model.PlayerSnapshot{
					{
						ID:          9990,
						NumVillages: 1000,
						Points:      1010,
						Rank:        1020,
						RankAtt:     999,
						ScoreAtt:    1000,
						RankDef:     1010,
						ScoreDef:    1020,
						RankSup:     1030,
						ScoreSup:    1040,
						RankTotal:   1050,
						ScoreTotal:  1060,
						Date:        model.Date{Time: now.Add(-24 * time.Hour).Truncate(24 * time.Hour)},
					},
				},
			},
			expectedTotalCount: "101",
		},
		{
			name: "ERR: limit is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				playerSnapshotSvc *mock.FakePlayerSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"limit": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "limit: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: offset is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				playerSnapshotSvc *mock.FakePlayerSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"offset": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "offset: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - invalid format",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				playerSnapshotSvc *mock.FakePlayerSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"sort": []string{"date:asc", "test"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: parsing \"test\": invalid syntax, expected field:direction",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported field",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				playerSnapshotSvc *mock.FakePlayerSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"sort": []string{"date:asc", "test:asc"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: unsupported sort by: \"test\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported direction",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				playerSnapshotSvc *mock.FakePlayerSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"sort": []string{"date:asc2"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[0]: unsupported sort direction: \"asc2\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: player id is not a valid int64",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				playerSnapshotSvc *mock.FakePlayerSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			playerID:       "asdf",
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "playerId: strconv.ParseInt: parsing \"asdf\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: version not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				playerSnapshotSvc *mock.FakePlayerSnapshotService,
			) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			serverKey:      server.Key,
			playerID:       playerID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: server not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				playerSnapshotSvc *mock.FakePlayerSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{}, domain.ServerNotFoundError{Key: server.Key + "2"})
			},
			versionCode:    version.Code,
			serverKey:      server.Key + "2",
			playerID:       playerID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("server (key=%s) not found", server.Key+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: player not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				playerSnapshotSvc *mock.FakePlayerSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(domain.PlayerWithRelations{}, domain.PlayerNotFoundError{ID: 1241})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			playerID:       playerID + "1",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("player (id=%s) not found", playerID+"1"),
				},
			},
			expectedTotalCount: "",
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			playerSnapshotSvc := &mock.FakePlayerSnapshotService{}
			playerSvc := &mock.FakePlayerService{}
			tt.setup(versionSvc, serverSvc, playerSvc, playerSnapshotSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
				withPlayerSnapshotService(playerSnapshotSvc),
				withPlayerService(playerSvc),
			)

			target := fmt.Sprintf(
				"/v1/versions/%s/servers/%s/players/%s/snapshots?%s",
				tt.versionCode,
				tt.serverKey,
				tt.playerID,
				tt.queryParams.Encode(),
			)
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertTotalCount(t, resp.Header, tt.expectedTotalCount)
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}
