package rest

import (
	"errors"
	"fmt"
	"net/url"
	"strconv"
	"strings"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
)

const (
	sortKey = "sort"
)

type queryParams[Sort any] struct {
	values  url.Values
	newSort func(by string, direction string) (Sort, error)
}

func (q queryParams[_]) nullBool(key string) (domain.NullBool, error) {
	if !q.values.Has(key) {
		return domain.NullBool{}, nil
	}

	v, err := strconv.ParseBool(q.values.Get(key))
	if err != nil {
		return domain.NullBool{}, domain.ValidationError{Field: key, Err: err}
	}

	return domain.NullBool{
		Valid: true,
		Bool:  v,
	}, nil
}

func (q queryParams[_]) pagination() (domain.Pagination, error) {
	limit, err := q.nullInt32("limit")
	if err != nil {
		return domain.Pagination{}, err
	}

	offset, err := q.nullInt32("offset")
	if err != nil {
		return domain.Pagination{}, err
	}

	p := domain.Pagination{}

	if limit.Valid {
		p.Limit = limit.Int32
	}

	if offset.Valid {
		p.Offset = offset.Int32
	}

	return p, nil
}

func (q queryParams[_]) nullInt32(key string) (domain.NullInt32, error) {
	if !q.values.Has(key) {
		return domain.NullInt32{}, nil
	}

	v, err := strconv.ParseInt(q.values.Get(key), 10, 32)
	if err != nil {
		return domain.NullInt32{}, domain.ValidationError{Field: key, Err: err}
	}

	return domain.NullInt32{
		Valid: true,
		Int32: int32(v),
	}, nil
}

func (q queryParams[Sort]) sort() ([]Sort, error) {
	if q.newSort == nil {
		return nil, errors.New("newSort is nil, can't proceed")
	}

	vals := q.strings(sortKey)
	if len(vals) == 0 {
		return nil, nil
	}

	res := make([]Sort, 0, len(vals))
	for i, v := range vals {
		index := strings.IndexByte(v, ':')
		if index == -1 {
			return nil, domain.SliceValidationError{
				Field: sortKey,
				Index: i,
				Err:   fmt.Errorf("parsing \"%s\": invalid syntax, expected field:direction", v),
			}
		}

		s, err := q.newSort(v[:index], v[index+1:])
		if err != nil {
			return nil, domain.SliceValidationError{
				Field: sortKey,
				Index: i,
				Err:   err,
			}
		}

		res = append(res, s)
	}

	return res, nil
}

func (q queryParams[_]) nullTime(key string) (domain.NullTime, error) {
	if !q.values.Has(key) {
		return domain.NullTime{}, nil
	}

	t, err := time.Parse(time.RFC3339, q.values.Get(key))
	if err != nil {
		return domain.NullTime{}, domain.ValidationError{
			Field: key,
			Err:   err,
		}
	}

	return domain.NullTime{
		Valid: true,
		Time:  t,
	}, nil
}

func (q queryParams[_]) coords(key string) ([]domain.Coords, error) {
	vals := q.strings(key)
	if len(vals) == 0 {
		return nil, nil
	}

	coords := make([]domain.Coords, 0, len(vals))
	for i, v := range vals {
		c, err := domain.NewCoords(v)
		if err != nil {
			return nil, domain.SliceValidationError{
				Field: key,
				Index: i,
				Err:   err,
			}
		}
		coords = append(coords, c)
	}

	return coords, nil
}

func (q queryParams[_]) strings(key string) []string {
	return q.values[key]
}
