package rest

import (
	"encoding/json"
	"errors"
	"net/http"
	"strconv"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
	httpSwagger "github.com/swaggo/http-swagger/v2"

	"github.com/go-chi/cors"

	"github.com/go-chi/chi/v5"

	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/docs"
)

//go:generate counterfeiter -generate

const headerXTotalCount = "X-Total-Count"

func New(
	versionSvc VersionService,
	serverSvc ServerService,
	tribeSvc TribeService,
	playerSvc PlayerService,
	villageSvc VillageService,
	ennoblementSvc EnnoblementService,
	playerSnapshotSvc PlayerSnapshotService,
	tribeSnapshotSvc TribeSnapshotService,
	tribeChangeSvc TribeChangeService,
	opts ...Option,
) *chi.Mux {
	cfg := newConfig(opts...)

	// handlers
	versionHandler := &version{
		svc: versionSvc,
	}
	serverHandler := &server{
		svc: serverSvc,
	}
	tribeHandler := &tribe{
		svc: tribeSvc,
	}
	playerHandler := &player{
		svc: playerSvc,
	}
	playerSnapshotHandler := &playerSnapshot{
		svc: playerSnapshotSvc,
	}
	tribeSnapshotHandler := &tribeSnapshot{
		svc: tribeSnapshotSvc,
	}
	villageHandler := &village{
		svc: villageSvc,
	}
	ennoblementHandler := &ennoblement{
		svc: ennoblementSvc,
	}
	tribeChangeHandler := &tribeChange{
		svc: tribeChangeSvc,
	}

	router := chi.NewRouter()

	if cfg.cors.Enabled {
		router.Use(cors.Handler(cors.Options{
			AllowedOrigins:   cfg.cors.AllowedOrigins,
			AllowCredentials: cfg.cors.AllowCredentials,
			AllowedMethods:   cfg.cors.AllowedMethods,
			AllowedHeaders:   []string{"Origin", "Content-Length", "Content-Type"},
			ExposedHeaders:   []string{headerXTotalCount, "Content-Type", "Content-Length"},
			MaxAge:           cfg.cors.MaxAge,
		}))
	}

	router.NotFound(routeNotFoundHandler)
	router.MethodNotAllowed(methodNotAllowedHandler)

	router.Route("/v1", func(r chi.Router) {
		if cfg.swagger.Enabled {
			docs.SwaggerInfo.Host = cfg.swagger.Host
			docs.SwaggerInfo.Schemes = cfg.swagger.Schemes
			r.Get("/swagger/*", httpSwagger.Handler(httpSwagger.DocExpansion("none")))
		}

		r.Route("/versions", func(r chi.Router) {
			r.Get("/", versionHandler.list)
			r.With(versionMiddleware(versionSvc)).Route("/{versionCode}", func(r chi.Router) {
				r.Get("/", versionHandler.getByCode)
				r.Route("/servers", func(r chi.Router) {
					r.Get("/", serverHandler.list)
					r.With(serverMiddleware(serverSvc)).Route("/{serverKey}", func(r chi.Router) {
						r.Get("/", serverHandler.getByKey)
						r.Get("/config", serverHandler.getConfig)
						r.Get("/building-info", serverHandler.getBuildingInfo)
						r.Get("/unit-info", serverHandler.getUnitInfo)
						r.Route("/tribes", func(r chi.Router) {
							r.Get("/", tribeHandler.list)
							r.With(tribeMiddleware(tribeSvc)).Route("/{tribeId}", func(r chi.Router) {
								r.Get("/", tribeHandler.getByID)
								r.Get("/members", playerHandler.listTribeMembers)
								r.Get("/ennoblements", ennoblementHandler.listTribe)
								r.Get("/history", tribeSnapshotHandler.history)
								r.Get("/snapshots", tribeSnapshotHandler.list)
								r.Get("/tribe-changes", tribeChangeHandler.listTribe)
							})
						})
						r.Route("/players", func(r chi.Router) {
							r.Get("/", playerHandler.list)
							r.With(playerMiddleware(playerSvc)).Route("/{playerId}", func(r chi.Router) {
								r.Get("/", playerHandler.getByID)
								r.Get("/ennoblements", ennoblementHandler.listPlayer)
								r.Get("/history", playerSnapshotHandler.history)
								r.Get("/snapshots", playerSnapshotHandler.list)
								r.Get("/tribe-changes", tribeChangeHandler.listPlayer)
								r.Get("/other-servers", playerHandler.listOtherServers)
								r.Get("/villages", villageHandler.listPlayer)
							})
						})
						r.Route("/villages", func(r chi.Router) {
							r.Get("/", villageHandler.list)
							r.With(villageMiddleware(villageSvc)).Route("/{villageId}", func(r chi.Router) {
								r.Get("/", villageHandler.getByID)
								r.Get("/ennoblements", ennoblementHandler.listVillage)
							})
						})
						r.Get("/ennoblements", ennoblementHandler.list)
					})
				})
			})
		})
	})

	return router
}

func routeNotFoundHandler(w http.ResponseWriter, _ *http.Request) {
	renderJSON(w, http.StatusNotFound, model.ErrorResp{
		Error: model.APIError{
			Code:    "route-not-found",
			Message: "route not found",
		},
	})
}

func methodNotAllowedHandler(w http.ResponseWriter, _ *http.Request) {
	renderJSON(w, http.StatusMethodNotAllowed, model.ErrorResp{
		Error: model.APIError{
			Code:    "method-not-allowed",
			Message: "method not allowed",
		},
	})
}

type internalServerError struct {
	err error
}

func (e internalServerError) Error() string {
	return e.err.Error()
}

func (e internalServerError) UserError() string {
	return "internal server error"
}

func (e internalServerError) Code() domain.ErrorCode {
	return domain.ErrorCodeUnknown
}

func (e internalServerError) Unwrap() error {
	return e.err
}

func renderErr(w http.ResponseWriter, err error) {
	var userError domain.Error
	if !errors.As(err, &userError) {
		userError = internalServerError{err: err}
	}
	renderJSON(w, errorCodeToHTTPStatus(userError.Code()), model.ErrorResp{
		Error: model.APIError{
			Code:    userError.Code().String(),
			Message: userError.UserError(),
		},
	})
}

func errorCodeToHTTPStatus(code domain.ErrorCode) int {
	switch code {
	case domain.ErrorCodeValidationError:
		return http.StatusBadRequest
	case domain.ErrorCodeEntityNotFound:
		return http.StatusNotFound
	case domain.ErrorCodeUnknown:
		fallthrough
	default:
		return http.StatusInternalServerError
	}
}

func renderJSON(w http.ResponseWriter, status int, data any) {
	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(status)
	_ = json.NewEncoder(w).Encode(data)
}

func setTotalCountHeader(header http.Header, total int64) {
	header.Set(headerXTotalCount, strconv.FormatInt(total, 10))
}

func urlParamInt64(ctx *chi.Context, key string) (int64, error) {
	v, err := strconv.ParseInt(ctx.URLParam(key), 10, 64)
	if err != nil {
		return 0, domain.ValidationError{
			Field: key,
			Err:   err,
		}
	}
	return v, nil
}
