package rest_test

import (
	"encoding/json"
	"io"
	"net/http"
	"net/http/httptest"
	"strings"
	"testing"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/router/rest"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
	"github.com/google/uuid"

	"github.com/go-chi/chi/v5"
	"github.com/google/go-cmp/cmp"
	"github.com/google/go-cmp/cmp/cmpopts"
	"github.com/stretchr/testify/assert"
)

func TestRouteNotFound(t *testing.T) {
	t.Parallel()

	expectedResp := &model.ErrorResp{
		Error: model.APIError{
			Code:    "route-not-found",
			Message: "route not found",
		},
	}

	resp := doRequest(newRouter(), http.MethodGet, "/v1/"+uuid.NewString(), nil)
	defer resp.Body.Close()
	assertJSONResponse(t, resp, http.StatusNotFound, expectedResp, &model.ErrorResp{})
}

func TestMethodNotAllowed(t *testing.T) {
	t.Parallel()

	expectedResp := &model.ErrorResp{
		Error: model.APIError{
			Code:    "method-not-allowed",
			Message: "method not allowed",
		},
	}

	resp := doRequest(newRouter(), http.MethodPost, "/v1/versions", nil)
	defer resp.Body.Close()
	assertJSONResponse(t, resp, http.StatusMethodNotAllowed, expectedResp, &model.ErrorResp{})
}

func TestCORS(t *testing.T) {
	t.Parallel()

	tests := []struct {
		name           string
		cfg            rest.CORSConfig
		reqHeaders     map[string]string
		resHeaders     map[string]string
		expectedStatus int
	}{
		{
			name: "Disabled",
			cfg: rest.CORSConfig{
				Enabled:          false,
				AllowedOrigins:   []string{"*"},
				AllowedMethods:   []string{"GET", "POST"},
				AllowCredentials: false,
				MaxAge:           600,
			},
			reqHeaders: map[string]string{
				"Origin":                        "https://tribalwarshelp.com",
				"Access-Control-Request-Method": http.MethodGet,
			},
			resHeaders: map[string]string{
				"Content-Type": "application/json",
			},
			expectedStatus: http.StatusMethodNotAllowed,
		},
		{
			name: "*",
			cfg: rest.CORSConfig{
				Enabled:          true,
				AllowedOrigins:   []string{"*"},
				AllowedMethods:   []string{"GET", "POST"},
				AllowCredentials: false,
				MaxAge:           600,
			},
			reqHeaders: map[string]string{
				"Origin":                        "https://tribalwarshelp.com",
				"Access-Control-Request-Method": http.MethodGet,
			},
			resHeaders: map[string]string{
				"Access-Control-Allow-Origin":  "*",
				"Access-Control-Allow-Methods": http.MethodGet,
				"Access-Control-Max-Age":       "600",
				"Vary":                         "Origin Access-Control-Request-Method Access-Control-Request-Headers",
			},
			expectedStatus: http.StatusOK,
		},
		{
			name: "https://tribalwarshelp.com",
			cfg: rest.CORSConfig{
				Enabled:          true,
				AllowedOrigins:   []string{"https://tribalwarshelp.com"},
				AllowedMethods:   []string{"GET", "POST"},
				AllowCredentials: true,
				MaxAge:           300,
			},
			reqHeaders: map[string]string{
				"Origin":                        "https://tribalwarshelp.com",
				"Access-Control-Request-Method": http.MethodGet,
			},
			resHeaders: map[string]string{
				"Access-Control-Allow-Origin":      "https://tribalwarshelp.com",
				"Access-Control-Allow-Credentials": "true",
				"Access-Control-Allow-Methods":     http.MethodGet,
				"Access-Control-Max-Age":           "300",
				"Vary":                             "Origin Access-Control-Request-Method Access-Control-Request-Headers",
			},
			expectedStatus: http.StatusOK,
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			r := newRouter(withOption(rest.WithCORSConfig(tt.cfg)))

			req := httptest.NewRequest(http.MethodOptions, "/v1/versions", nil)
			for k, v := range tt.reqHeaders {
				req.Header.Set(k, v)
			}

			resp := doCustomRequest(r, req)
			defer resp.Body.Close()
			assert.Equal(t, tt.expectedStatus, resp.StatusCode)
			assert.Len(t, resp.Header, len(tt.resHeaders))
			for k, expected := range tt.resHeaders {
				assert.Equal(t, expected, strings.Join(resp.Header.Values(k), " "))
			}
		})
	}
}

func TestSwagger(t *testing.T) {
	t.Parallel()

	tests := []struct {
		name                string
		target              string
		expectedContentType string
	}{
		{
			name:                "/v1/swagger/index.html",
			target:              "/v1/swagger/index.html",
			expectedContentType: "text/html; charset=utf-8",
		},
		{
			name:                "/v1/swagger/doc.json",
			target:              "/v1/swagger/doc.json",
			expectedContentType: "application/json; charset=utf-8",
		},
	}
	router := newRouter(withOption(rest.WithSwaggerConfig(rest.SwaggerConfig{
		Enabled: true,
	})))

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			resp := doRequest(router, http.MethodGet, tt.target, nil)
			defer resp.Body.Close()
			assert.Equal(t, http.StatusOK, resp.StatusCode)
			assert.Equal(t, tt.expectedContentType, resp.Header.Get("Content-Type"))
		})
	}
}

type routerConfig struct {
	versionSvc        rest.VersionService
	serverSvc         rest.ServerService
	tribeSvc          rest.TribeService
	playerSvc         rest.PlayerService
	villageSvc        rest.VillageService
	ennoblementSvc    rest.EnnoblementService
	playerSnapshotSvc rest.PlayerSnapshotService
	tribeSnapshotSvc  rest.TribeSnapshotService
	tribeChangeSvc    rest.TribeChangeService
	opts              []rest.Option
}

type routerOption func(cfg *routerConfig)

func withVersionService(svc rest.VersionService) routerOption {
	return func(cfg *routerConfig) {
		cfg.versionSvc = svc
	}
}

func withServerService(svc rest.ServerService) routerOption {
	return func(cfg *routerConfig) {
		cfg.serverSvc = svc
	}
}

func withTribeService(svc rest.TribeService) routerOption {
	return func(cfg *routerConfig) {
		cfg.tribeSvc = svc
	}
}

func withPlayerService(svc rest.PlayerService) routerOption {
	return func(cfg *routerConfig) {
		cfg.playerSvc = svc
	}
}

func withVillageService(svc rest.VillageService) routerOption {
	return func(cfg *routerConfig) {
		cfg.villageSvc = svc
	}
}

func withEnnoblementService(svc rest.EnnoblementService) routerOption {
	return func(cfg *routerConfig) {
		cfg.ennoblementSvc = svc
	}
}

func withPlayerSnapshotService(svc rest.PlayerSnapshotService) routerOption {
	return func(cfg *routerConfig) {
		cfg.playerSnapshotSvc = svc
	}
}

func withTribeSnapshotService(svc rest.TribeSnapshotService) routerOption {
	return func(cfg *routerConfig) {
		cfg.tribeSnapshotSvc = svc
	}
}

func withTribeChangeService(svc rest.TribeChangeService) routerOption {
	return func(cfg *routerConfig) {
		cfg.tribeChangeSvc = svc
	}
}

func withOption(opt rest.Option) routerOption {
	return func(cfg *routerConfig) {
		cfg.opts = append(cfg.opts, opt)
	}
}

func newRouter(opts ...routerOption) chi.Router {
	cfg := &routerConfig{}
	for _, opt := range opts {
		opt(cfg)
	}
	return rest.New(
		cfg.versionSvc,
		cfg.serverSvc,
		cfg.tribeSvc,
		cfg.playerSvc,
		cfg.villageSvc,
		cfg.ennoblementSvc,
		cfg.playerSnapshotSvc,
		cfg.tribeSnapshotSvc,
		cfg.tribeChangeSvc,
		cfg.opts...,
	)
}

func doRequest(mux chi.Router, method, target string, body io.Reader) *http.Response {
	return doCustomRequest(mux, httptest.NewRequest(method, target, body))
}

func doCustomRequest(mux chi.Router, r *http.Request) *http.Response {
	rr := httptest.NewRecorder()
	mux.ServeHTTP(rr, r)
	return rr.Result()
}

func assertJSONResponse(tb testing.TB, resp *http.Response, expectedStatus int, expected any, target any) {
	tb.Helper()

	assert.Equal(tb, "application/json", resp.Header.Get("Content-Type"))
	assert.Equal(tb, expectedStatus, resp.StatusCode)
	assert.NoError(tb, json.NewDecoder(resp.Body).Decode(target))
	assert.Empty(tb, cmp.Diff(expected, target, cmpopts.IgnoreUnexported(time.Time{})))
}

func assertTotalCount(tb testing.TB, header http.Header, expected string) {
	tb.Helper()

	assert.Equal(tb, expected, header.Get("X-Total-Count"))
}
