package rest

import (
	"context"
	"net/http"

	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"

	"github.com/go-chi/chi/v5"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
)

type serverCtxKey struct{}

//counterfeiter:generate -o internal/mock/server_service.gen.go . ServerService
type ServerService interface {
	ListCount(ctx context.Context, params domain.ListServersParams) ([]domain.Server, int64, error)
	GetNormalByVersionCodeAndKey(ctx context.Context, versionCode, key string) (domain.Server, error)
}

type server struct {
	svc ServerService
}

// @ID          listServers
// @Summary     List servers
// @Description List all servers
// @Tags        versions,servers
// @Produce     json
// @Success     200         {object}  model.ListServersResp
// @Success     400         {object}  model.ErrorResp
// @Success     404         {object}  model.ErrorResp
// @Failure     500         {object}  model.ErrorResp
// @Header      200         {integer} X-Total-Count "Total number of records"
// @Param       versionCode path      string        true  "Version code"
// @Param       open        query     boolean       false "true=only open servers, false=only closed servers, by default both open and closed servers are returned"
// @Param       offset      query     int           false "specifies where to start a page" minimum(0) default(0)
// @Param       limit       query     int           false "page size"                       minimum(1) maximum(500) default(500)
// @Router      /versions/{versionCode}/servers [get]
func (s *server) list(w http.ResponseWriter, r *http.Request) {
	var err error
	ctx := r.Context()
	ver, _ := versionFromContext(r.Context())
	params := domain.ListServersParams{
		Special: domain.NullBool{
			Valid: true,
			Bool:  false,
		},
		VersionCodes: []string{ver.Code},
	}
	query := queryParams[any]{
		values: r.URL.Query(),
	}

	params.Open, err = query.nullBool("open")
	if err != nil {
		renderErr(w, err)
		return
	}

	params.Pagination, err = query.pagination()
	if err != nil {
		renderErr(w, err)
		return
	}

	servers, count, err := s.svc.ListCount(ctx, params)
	if err != nil {
		renderErr(w, err)
		return
	}

	setTotalCountHeader(w.Header(), count)
	renderJSON(w, http.StatusOK, model.NewListServersResp(servers))
}

// @ID          getServer
// @Summary     Get a server
// @Description Get a server
// @Tags        versions,servers
// @Produce     json
// @Success     200         {object} model.GetServerResp
// @Failure     404         {object} model.ErrorResp
// @Failure     500         {object} model.ErrorResp
// @Param       versionCode path     string true "Version code"
// @Param       serverKey   path     string true "Server key"
// @Router      /versions/{versionCode}/servers/{serverKey} [get]
func (s *server) getByKey(w http.ResponseWriter, r *http.Request) {
	srv, _ := serverFromContext(r.Context())
	renderJSON(w, http.StatusOK, model.NewGetServerResp(srv))
}

// @ID          getServerConfig
// @Summary     Get server config
// @Description Get server config
// @Tags        versions,servers
// @Produce     json
// @Success     200         {object} model.GetServerConfigResp
// @Failure     404         {object} model.ErrorResp
// @Failure     500         {object} model.ErrorResp
// @Param       versionCode path     string true "Version code"
// @Param       serverKey   path     string true "Server key"
// @Router      /versions/{versionCode}/servers/{serverKey}/config [get]
func (s *server) getConfig(w http.ResponseWriter, r *http.Request) {
	srv, _ := serverFromContext(r.Context())
	renderJSON(w, http.StatusOK, model.NewGetServerConfigResp(srv.Config))
}

// @ID          getBuildingInfo
// @Summary     Get building info
// @Description Get building info
// @Tags        versions,servers
// @Produce     json
// @Success     200         {object} model.GetBuildingInfoResp
// @Failure     404         {object} model.ErrorResp
// @Failure     500         {object} model.ErrorResp
// @Param       versionCode path     string true "Version code"
// @Param       serverKey   path     string true "Server key"
// @Router      /versions/{versionCode}/servers/{serverKey}/building-info [get]
func (s *server) getBuildingInfo(w http.ResponseWriter, r *http.Request) {
	srv, _ := serverFromContext(r.Context())
	renderJSON(w, http.StatusOK, model.NewGetBuildingInfoResp(srv.BuildingInfo))
}

// @ID          getUnitInfo
// @Summary     Get unit info
// @Description Get unit info
// @Tags        versions,servers
// @Produce     json
// @Success     200         {object} model.GetUnitInfoResp
// @Failure     404         {object} model.ErrorResp
// @Failure     500         {object} model.ErrorResp
// @Param       versionCode path     string true "Version code"
// @Param       serverKey   path     string true "Server key"
// @Router      /versions/{versionCode}/servers/{serverKey}/unit-info [get]
func (s *server) getUnitInfo(w http.ResponseWriter, r *http.Request) {
	srv, _ := serverFromContext(r.Context())
	renderJSON(w, http.StatusOK, model.NewGetUnitInfoResp(srv.UnitInfo))
}

func serverMiddleware(svc ServerService) func(http.Handler) http.Handler {
	return func(next http.Handler) http.Handler {
		return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
			ctx := r.Context()
			ver, _ := versionFromContext(ctx)
			routeCtx := chi.RouteContext(ctx)

			srv, err := svc.GetNormalByVersionCodeAndKey(ctx, ver.Code, routeCtx.URLParam("serverKey"))
			if err != nil {
				renderErr(w, err)
				return
			}

			next.ServeHTTP(w, r.WithContext(serverToContext(ctx, srv)))
		})
	}
}

func serverToContext(ctx context.Context, srv domain.Server) context.Context {
	return context.WithValue(ctx, serverCtxKey{}, srv)
}

func serverFromContext(ctx context.Context) (domain.Server, bool) {
	srv, ok := ctx.Value(serverCtxKey{}).(domain.Server)
	return srv, ok
}
