package rest_test

import (
	"context"
	"fmt"
	"net/http"
	"net/url"
	"testing"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/mock"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
	"github.com/google/go-cmp/cmp"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
)

func TestServer_list(t *testing.T) {
	t.Parallel()

	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	now := time.Now()

	tests := []struct {
		name               string
		setup              func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService)
		versionCode        string
		queryParams        url.Values
		expectedStatus     int
		target             any
		expectedResponse   any
		expectedTotalCount string
	}{
		{
			name: "OK: without params",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.ListCountCalls(func(_ context.Context, params domain.ListServersParams) ([]domain.Server, int64, error) {
					expectedParams := domain.ListServersParams{
						Open: domain.NullBool{
							Valid: false,
							Bool:  false,
						},
						Special: domain.NullBool{
							Valid: true,
							Bool:  false,
						},
						VersionCodes: []string{version.Code},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					servers := []domain.Server{
						{
							Key:                  "pl151",
							URL:                  "https://pl151.plemiona.pl",
							Open:                 true,
							Special:              false,
							NumPlayers:           1234,
							NumTribes:            1235,
							NumVillages:          1236,
							NumPlayerVillages:    1237,
							NumBarbarianVillages: 1238,
							NumBonusVillages:     1239,
							Config: domain.ServerConfig{
								Speed: 0.25,
							},
							BuildingInfo: domain.BuildingInfo{
								Main: domain.Building{BuildTime: 0.33334},
							},
							UnitInfo: domain.UnitInfo{
								Axe: domain.Unit{Attack: 123},
							},
							CreatedAt:                now.Add(-20 * time.Hour),
							PlayerDataUpdatedAt:      time.Time{},
							TribeDataUpdatedAt:       now,
							VillageDataUpdatedAt:     now.Add(-5 * time.Hour),
							EnnoblementDataUpdatedAt: time.Time{},
							VersionCode:              "pl",
						},
						{
							Key:                      "pl152",
							URL:                      "https://pl152.plemiona.pl",
							Open:                     true,
							Special:                  false,
							NumPlayers:               2234,
							NumTribes:                2235,
							NumVillages:              2236,
							NumPlayerVillages:        1237,
							NumBarbarianVillages:     1238,
							NumBonusVillages:         1239,
							Config:                   domain.ServerConfig{},
							BuildingInfo:             domain.BuildingInfo{},
							UnitInfo:                 domain.UnitInfo{},
							CreatedAt:                now.Add(-20 * time.Hour),
							PlayerDataUpdatedAt:      now,
							TribeDataUpdatedAt:       now,
							VillageDataUpdatedAt:     now.Add(-5 * time.Hour),
							EnnoblementDataUpdatedAt: now.Add(-1 * time.Minute),
							VersionCode:              "pl",
						},
					}

					return servers, int64(len(servers)), nil
				})
			},
			versionCode:    version.Code,
			expectedStatus: http.StatusOK,
			target:         &model.ListServersResp{},
			expectedResponse: &model.ListServersResp{
				Data: []model.Server{
					{
						Key:                  "pl151",
						URL:                  "https://pl151.plemiona.pl",
						Open:                 true,
						NumPlayers:           1234,
						NumTribes:            1235,
						NumVillages:          1236,
						NumPlayerVillages:    1237,
						NumBarbarianVillages: 1238,
						NumBonusVillages:     1239,
						CreatedAt:            now.Add(-20 * time.Hour),
						PlayerDataUpdatedAt:  model.NullTime{},
						TribeDataUpdatedAt: model.NullTime{
							Valid: true,
							Time:  now,
						},
						VillageDataUpdatedAt: model.NullTime{
							Valid: true,
							Time:  now.Add(-5 * time.Hour),
						},
						EnnoblementDataUpdatedAt: model.NullTime{},
					},
					{
						Key:                  "pl152",
						URL:                  "https://pl152.plemiona.pl",
						Open:                 true,
						NumPlayers:           2234,
						NumTribes:            2235,
						NumVillages:          2236,
						NumPlayerVillages:    1237,
						NumBarbarianVillages: 1238,
						NumBonusVillages:     1239,
						CreatedAt:            now.Add(-20 * time.Hour),
						PlayerDataUpdatedAt: model.NullTime{
							Valid: true,
							Time:  now,
						},
						TribeDataUpdatedAt: model.NullTime{
							Valid: true,
							Time:  now,
						},
						VillageDataUpdatedAt: model.NullTime{
							Valid: true,
							Time:  now.Add(-5 * time.Hour),
						},
						EnnoblementDataUpdatedAt: model.NullTime{
							Valid: true,
							Time:  now.Add(-1 * time.Minute),
						},
					},
				},
			},
			expectedTotalCount: "2",
		},
		{
			name: "OK: open=true,limit=500,offset=25",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.ListCountCalls(func(_ context.Context, params domain.ListServersParams) ([]domain.Server, int64, error) {
					expectedParams := domain.ListServersParams{
						Open: domain.NullBool{
							Valid: true,
							Bool:  true,
						},
						Special: domain.NullBool{
							Valid: true,
							Bool:  false,
						},
						VersionCodes: []string{version.Code},
						Pagination: domain.Pagination{
							Limit:  500,
							Offset: 25,
						},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					servers := []domain.Server{
						{
							Key:                  "pl151",
							URL:                  "https://pl151.plemiona.pl",
							Open:                 true,
							Special:              false,
							NumPlayers:           1234,
							NumTribes:            1235,
							NumVillages:          1236,
							NumPlayerVillages:    1237,
							NumBarbarianVillages: 1238,
							NumBonusVillages:     1239,
							Config: domain.ServerConfig{
								Speed: 0.25,
							},
							BuildingInfo: domain.BuildingInfo{
								Main: domain.Building{BuildTime: 0.33334},
							},
							UnitInfo: domain.UnitInfo{
								Axe: domain.Unit{Attack: 123},
							},
							CreatedAt:                now.Add(-20 * time.Hour),
							PlayerDataUpdatedAt:      time.Time{},
							TribeDataUpdatedAt:       now,
							VillageDataUpdatedAt:     now.Add(-5 * time.Hour),
							EnnoblementDataUpdatedAt: time.Time{},
							VersionCode:              "pl",
						},
						{
							Key:                      "pl152",
							URL:                      "https://pl152.plemiona.pl",
							Open:                     true,
							Special:                  false,
							NumPlayers:               2234,
							NumTribes:                2235,
							NumVillages:              2236,
							NumPlayerVillages:        1237,
							NumBarbarianVillages:     1238,
							NumBonusVillages:         1239,
							Config:                   domain.ServerConfig{},
							BuildingInfo:             domain.BuildingInfo{},
							UnitInfo:                 domain.UnitInfo{},
							CreatedAt:                now.Add(-20 * time.Hour),
							PlayerDataUpdatedAt:      now,
							TribeDataUpdatedAt:       now,
							VillageDataUpdatedAt:     now.Add(-5 * time.Hour),
							EnnoblementDataUpdatedAt: now.Add(-1 * time.Minute),
							VersionCode:              "pl",
						},
					}

					return servers, int64(len(servers)) + 25, nil
				})
			},
			versionCode: version.Code,
			queryParams: url.Values{
				"offset": []string{"25"},
				"open":   []string{"true"},
				"limit":  []string{"500"},
			},
			expectedStatus: http.StatusOK,
			target:         &model.ListServersResp{},
			expectedResponse: &model.ListServersResp{
				Data: []model.Server{
					{
						Key:                  "pl151",
						URL:                  "https://pl151.plemiona.pl",
						Open:                 true,
						NumPlayers:           1234,
						NumTribes:            1235,
						NumVillages:          1236,
						NumPlayerVillages:    1237,
						NumBarbarianVillages: 1238,
						NumBonusVillages:     1239,
						CreatedAt:            now.Add(-20 * time.Hour),
						PlayerDataUpdatedAt:  model.NullTime{},
						TribeDataUpdatedAt: model.NullTime{
							Valid: true,
							Time:  now,
						},
						VillageDataUpdatedAt: model.NullTime{
							Valid: true,
							Time:  now.Add(-5 * time.Hour),
						},
						EnnoblementDataUpdatedAt: model.NullTime{},
					},
					{
						Key:                  "pl152",
						URL:                  "https://pl152.plemiona.pl",
						Open:                 true,
						NumPlayers:           2234,
						NumTribes:            2235,
						NumVillages:          2236,
						NumPlayerVillages:    1237,
						NumBarbarianVillages: 1238,
						NumBonusVillages:     1239,
						CreatedAt:            now.Add(-20 * time.Hour),
						PlayerDataUpdatedAt: model.NullTime{
							Valid: true,
							Time:  now,
						},
						TribeDataUpdatedAt: model.NullTime{
							Valid: true,
							Time:  now,
						},
						VillageDataUpdatedAt: model.NullTime{
							Valid: true,
							Time:  now.Add(-5 * time.Hour),
						},
						EnnoblementDataUpdatedAt: model.NullTime{
							Valid: true,
							Time:  now.Add(-1 * time.Minute),
						},
					},
				},
			},
			expectedTotalCount: "27",
		},
		{
			name: "OK: open=false,limit=1",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.ListCountCalls(func(_ context.Context, params domain.ListServersParams) ([]domain.Server, int64, error) {
					expectedParams := domain.ListServersParams{
						Open: domain.NullBool{
							Valid: true,
							Bool:  false,
						},
						Special: domain.NullBool{
							Valid: true,
							Bool:  false,
						},
						VersionCodes: []string{version.Code},
						Pagination: domain.Pagination{
							Limit:  1,
							Offset: 0,
						},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					servers := []domain.Server{
						{
							Key:                  "pl151",
							URL:                  "https://pl151.plemiona.pl",
							Open:                 true,
							Special:              false,
							NumPlayers:           1234,
							NumTribes:            1235,
							NumVillages:          1236,
							NumPlayerVillages:    1237,
							NumBarbarianVillages: 1238,
							NumBonusVillages:     1239,
							Config: domain.ServerConfig{
								Speed: 0.25,
							},
							BuildingInfo: domain.BuildingInfo{
								Main: domain.Building{BuildTime: 0.33334},
							},
							UnitInfo: domain.UnitInfo{
								Axe: domain.Unit{Attack: 123},
							},
							CreatedAt:                now.Add(-20 * time.Hour),
							PlayerDataUpdatedAt:      time.Time{},
							TribeDataUpdatedAt:       now,
							VillageDataUpdatedAt:     now.Add(-5 * time.Hour),
							EnnoblementDataUpdatedAt: time.Time{},
							VersionCode:              "pl",
						},
					}

					return servers, int64(len(servers)), nil
				})
			},
			versionCode: version.Code,
			queryParams: url.Values{
				"open":  []string{"false"},
				"limit": []string{"1"},
			},
			expectedStatus: http.StatusOK,
			target:         &model.ListServersResp{},
			expectedResponse: &model.ListServersResp{
				Data: []model.Server{
					{
						Key:                  "pl151",
						URL:                  "https://pl151.plemiona.pl",
						Open:                 true,
						NumPlayers:           1234,
						NumTribes:            1235,
						NumVillages:          1236,
						NumPlayerVillages:    1237,
						NumBarbarianVillages: 1238,
						NumBonusVillages:     1239,
						CreatedAt:            now.Add(-20 * time.Hour),
						PlayerDataUpdatedAt:  model.NullTime{},
						TribeDataUpdatedAt: model.NullTime{
							Valid: true,
							Time:  now,
						},
						VillageDataUpdatedAt: model.NullTime{
							Valid: true,
							Time:  now.Add(-5 * time.Hour),
						},
						EnnoblementDataUpdatedAt: model.NullTime{},
					},
				},
			},
			expectedTotalCount: "1",
		},
		{
			name: "ERR: limit is not a valid int32",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(version, nil)
			},
			versionCode: version.Code,
			queryParams: url.Values{
				"limit": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "limit: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: offset is not a valid int32",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(version, nil)
			},
			versionCode: version.Code,
			queryParams: url.Values{
				"offset": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "offset: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: open is not a valid boolean",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(version, nil)
			},
			versionCode: version.Code,
			queryParams: url.Values{
				"open": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "open: strconv.ParseBool: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: version not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
			expectedTotalCount: "",
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			tt.setup(versionSvc, serverSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
			)

			target := "/v1/versions/" + tt.versionCode + "/servers?" + tt.queryParams.Encode()
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertTotalCount(t, resp.Header, tt.expectedTotalCount)
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}

func TestServer_getByKey(t *testing.T) {
	t.Parallel()

	now := time.Now()
	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	tests := []struct {
		name             string
		setup            func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService)
		versionCode      string
		key              string
		expectedStatus   int
		target           any
		expectedResponse any
	}{
		{
			name: "OK",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{
					Key:                  "pl151",
					URL:                  "https://pl151.plemiona.pl",
					Open:                 false,
					Special:              false,
					NumPlayers:           1234,
					NumTribes:            1235,
					NumVillages:          1236,
					NumPlayerVillages:    1237,
					NumBarbarianVillages: 1238,
					NumBonusVillages:     1239,
					Config: domain.ServerConfig{
						Speed: 0.25,
					},
					BuildingInfo: domain.BuildingInfo{
						Snob: domain.Building{BuildTime: 1234.444},
					},
					UnitInfo: domain.UnitInfo{
						Sword: domain.Unit{BuildTime: 1111122.2222},
					},
					CreatedAt:                now.Add(-20 * time.Hour),
					PlayerDataUpdatedAt:      time.Time{},
					TribeDataUpdatedAt:       now,
					VillageDataUpdatedAt:     now.Add(-5 * time.Hour),
					EnnoblementDataUpdatedAt: now.Add(-15 * time.Hour),
					VersionCode:              "pl",
				}, nil)
			},
			versionCode:    version.Code,
			key:            "pl151",
			expectedStatus: http.StatusOK,
			target:         &model.GetServerResp{},
			expectedResponse: &model.GetServerResp{
				Data: model.Server{
					Key:                  "pl151",
					URL:                  "https://pl151.plemiona.pl",
					Open:                 false,
					NumPlayers:           1234,
					NumTribes:            1235,
					NumVillages:          1236,
					NumPlayerVillages:    1237,
					NumBarbarianVillages: 1238,
					NumBonusVillages:     1239,
					CreatedAt:            now.Add(-20 * time.Hour),
					PlayerDataUpdatedAt:  model.NullTime{},
					TribeDataUpdatedAt: model.NullTime{
						Valid: true,
						Time:  now,
					},
					VillageDataUpdatedAt: model.NullTime{
						Valid: true,
						Time:  now.Add(-5 * time.Hour),
					},
					EnnoblementDataUpdatedAt: model.NullTime{
						Valid: true,
						Time:  now.Add(-15 * time.Hour),
					},
				},
			},
		},
		{
			name: "ERR: version not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			key:            "pl151",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
		},
		{
			name: "ERR: server not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{}, domain.ServerNotFoundError{Key: "pl1512"})
			},
			versionCode:    version.Code,
			key:            "pl1512",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: "server (key=pl1512) not found",
				},
			},
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			tt.setup(versionSvc, serverSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
			)

			target := "/v1/versions/" + tt.versionCode + "/servers/" + tt.key
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}

func TestServer_getConfig(t *testing.T) {
	t.Parallel()

	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	tests := []struct {
		name             string
		setup            func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService)
		versionCode      string
		key              string
		expectedStatus   int
		target           any
		expectedResponse any
	}{
		{
			name: "OK",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{
					Config: domain.ServerConfig{
						Speed:     2,
						UnitSpeed: 0.5,
						Moral:     1,
						Build: domain.ServerConfigBuild{
							Destroy: 0,
						},
						Misc: domain.ServerConfigMisc{
							KillRanking:     1,
							Tutorial:        2,
							TradeCancelTime: 3,
						},
						Commands: domain.ServerConfigCommands{
							MillisArrival:     1,
							CommandCancelTime: 2,
						},
						Newbie: domain.ServerConfigNewbie{
							Days:                 1,
							RatioDays:            2,
							Ratio:                3,
							RemoveNewbieVillages: 4,
						},
						Game: domain.ServerConfigGame{
							BuildtimeFormula:   1,
							Knight:             2,
							KnightNewItems:     3,
							Archer:             4,
							Tech:               5,
							FarmLimit:          6,
							Church:             7,
							Watchtower:         8,
							Stronghold:         9,
							FakeLimit:          10,
							BarbarianRise:      11,
							BarbarianShrink:    12,
							BarbarianMaxPoints: 13,
							Scavenging:         20,
							Hauls:              14,
							HaulsBase:          15,
							HaulsMax:           16,
							BaseProduction:     17,
							Event:              18,
							SuppressEvents:     19,
						},
						Buildings: domain.ServerConfigBuildings{
							CustomMain:       -1,
							CustomFarm:       10,
							CustomStorage:    11,
							CustomPlace:      12,
							CustomBarracks:   13,
							CustomChurch:     14,
							CustomSmith:      15,
							CustomWood:       16,
							CustomStone:      17,
							CustomIron:       18,
							CustomMarket:     19,
							CustomStable:     20,
							CustomWall:       21,
							CustomGarage:     22,
							CustomHide:       23,
							CustomSnob:       24,
							CustomStatue:     25,
							CustomWatchtower: 26,
						},
						Snob: domain.ServerConfigSnob{
							Gold:          100,
							CheapRebuild:  101,
							Rise:          102,
							MaxDist:       102,
							Factor:        103,
							CoinWood:      104,
							CoinStone:     105,
							CoinIron:      106,
							NoBarbConquer: 107,
						},
						Ally: domain.ServerConfigAlly{
							NoHarm:                60,
							NoOtherSupport:        61,
							NoOtherSupportType:    62,
							AllytimeSupport:       63,
							NoLeave:               64,
							NoJoin:                65,
							Limit:                 66,
							FixedAllies:           67,
							PointsMemberCount:     68,
							WarsMemberRequirement: 69,
							WarsPointsRequirement: 70,
							WarsAutoacceptDays:    71,
							Levels:                72,
							XpRequirements:        "v1",
						},
						Coord: domain.ServerConfigCoord{
							MapSize:         1000,
							Func:            1,
							EmptyVillages:   2,
							BonusVillages:   3,
							BonusNew:        4,
							Inner:           5,
							SelectStart:     6,
							VillageMoveWait: 7,
							NobleRestart:    8,
							StartVillages:   9,
						},
						Sitter: domain.ServerConfigSitter{
							Allow: 1,
						},
						Sleep: domain.ServerConfigSleep{
							Active:   50,
							Delay:    51,
							Min:      52,
							Max:      53,
							MinAwake: 54,
							MaxAwake: 55,
							WarnTime: 56,
						},
						Night: domain.ServerConfigNight{
							Active:    1,
							StartHour: 23,
							EndHour:   8,
							DefFactor: 1.5,
							Duration:  11,
						},
						Win: domain.ServerConfigWin{
							Check: 16,
						},
					},
				}, nil)
			},
			versionCode:    version.Code,
			key:            "pl151",
			expectedStatus: http.StatusOK,
			target:         &model.GetServerConfigResp{},
			expectedResponse: &model.GetServerConfigResp{
				Data: model.ServerConfig{
					Speed:     2,
					UnitSpeed: 0.5,
					Moral:     1,
					Build: model.ServerConfigBuild{
						Destroy: 0,
					},
					Misc: model.ServerConfigMisc{
						KillRanking:     1,
						Tutorial:        2,
						TradeCancelTime: 3,
					},
					Commands: model.ServerConfigCommands{
						MillisArrival:     1,
						CommandCancelTime: 2,
					},
					Newbie: model.ServerConfigNewbie{
						Days:                 1,
						RatioDays:            2,
						Ratio:                3,
						RemoveNewbieVillages: 4,
					},
					Game: model.ServerConfigGame{
						BuildtimeFormula:   1,
						Knight:             2,
						KnightNewItems:     3,
						Archer:             4,
						Tech:               5,
						FarmLimit:          6,
						Church:             7,
						Watchtower:         8,
						Stronghold:         9,
						FakeLimit:          10,
						BarbarianRise:      11,
						BarbarianShrink:    12,
						BarbarianMaxPoints: 13,
						Scavenging:         20,
						Hauls:              14,
						HaulsBase:          15,
						HaulsMax:           16,
						BaseProduction:     17,
						Event:              18,
						SuppressEvents:     19,
					},
					Buildings: model.ServerConfigBuildings{
						CustomMain:       -1,
						CustomFarm:       10,
						CustomStorage:    11,
						CustomPlace:      12,
						CustomBarracks:   13,
						CustomChurch:     14,
						CustomSmith:      15,
						CustomWood:       16,
						CustomStone:      17,
						CustomIron:       18,
						CustomMarket:     19,
						CustomStable:     20,
						CustomWall:       21,
						CustomGarage:     22,
						CustomHide:       23,
						CustomSnob:       24,
						CustomStatue:     25,
						CustomWatchtower: 26,
					},
					Snob: model.ServerConfigSnob{
						Gold:          100,
						CheapRebuild:  101,
						Rise:          102,
						MaxDist:       102,
						Factor:        103,
						CoinWood:      104,
						CoinStone:     105,
						CoinIron:      106,
						NoBarbConquer: 107,
					},
					Ally: model.ServerConfigAlly{
						NoHarm:                60,
						NoOtherSupport:        61,
						NoOtherSupportType:    62,
						AllytimeSupport:       63,
						NoLeave:               64,
						NoJoin:                65,
						Limit:                 66,
						FixedAllies:           67,
						PointsMemberCount:     68,
						WarsMemberRequirement: 69,
						WarsPointsRequirement: 70,
						WarsAutoacceptDays:    71,
						Levels:                72,
						XpRequirements:        "v1",
					},
					Coord: model.ServerConfigCoord{
						MapSize:         1000,
						Func:            1,
						EmptyVillages:   2,
						BonusVillages:   3,
						BonusNew:        4,
						Inner:           5,
						SelectStart:     6,
						VillageMoveWait: 7,
						NobleRestart:    8,
						StartVillages:   9,
					},
					Sitter: model.ServerConfigSitter{
						Allow: 1,
					},
					Sleep: model.ServerConfigSleep{
						Active:   50,
						Delay:    51,
						Min:      52,
						Max:      53,
						MinAwake: 54,
						MaxAwake: 55,
						WarnTime: 56,
					},
					Night: model.ServerConfigNight{
						Active:    1,
						StartHour: 23,
						EndHour:   8,
						DefFactor: 1.5,
						Duration:  11,
					},
					Win: model.ServerConfigWin{
						Check: 16,
					},
				},
			},
		},
		{
			name: "ERR: version not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			key:            "pl151",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
		},
		{
			name: "ERR: server not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{}, domain.ServerNotFoundError{Key: "pl1512"})
			},
			versionCode:    version.Code,
			key:            "pl1512",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: "server (key=pl1512) not found",
				},
			},
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			tt.setup(versionSvc, serverSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
			)

			target := fmt.Sprintf("/v1/versions/%s/servers/%s/config", tt.versionCode, tt.key)
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}

func TestServer_getUnitInfo(t *testing.T) {
	t.Parallel()

	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	tests := []struct {
		name             string
		setup            func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService)
		versionCode      string
		key              string
		expectedStatus   int
		target           any
		expectedResponse any
	}{
		{
			name: "OK",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{
					UnitInfo: domain.UnitInfo{
						Spear: domain.Unit{
							BuildTime: 5.8777,
						},
						Sword: domain.Unit{
							Pop: 1,
						},
						Axe: domain.Unit{
							Speed: 13.7777,
						},
						Archer: domain.Unit{
							Attack: 10,
						},
						Spy: domain.Unit{
							Defense: 1,
						},
						Light: domain.Unit{
							DefenseCavalry: 15,
						},
						Marcher: domain.Unit{
							DefenseArcher: 10,
						},
						Heavy: domain.Unit{
							BuildTime: 30.1234,
							Pop:       6,
						},
						Ram: domain.Unit{
							Speed:  30.849,
							Attack: 1,
						},
						Catapult: domain.Unit{
							Defense:        12,
							DefenseCavalry: 123,
						},
						Knight: domain.Unit{
							DefenseArcher: 1234,
							Carry:         10,
						},
						Snob: domain.Unit{
							BuildTime: 1234.4,
							Pop:       100,
							Speed:     35,
						},
						Militia: domain.Unit{
							BuildTime:      1,
							Pop:            2,
							Speed:          3,
							Attack:         4,
							Defense:        5,
							DefenseCavalry: 6,
							DefenseArcher:  7,
							Carry:          8,
						},
					},
				}, nil)
			},
			versionCode:    version.Code,
			key:            "pl151",
			expectedStatus: http.StatusOK,
			target:         &model.GetUnitInfoResp{},
			expectedResponse: &model.GetUnitInfoResp{
				Data: model.UnitInfo{
					Spear: model.Unit{
						BuildTime: 5.8777,
					},
					Sword: model.Unit{
						Pop: 1,
					},
					Axe: model.Unit{
						Speed: 13.7777,
					},
					Archer: model.Unit{
						Attack: 10,
					},
					Spy: model.Unit{
						Defense: 1,
					},
					Light: model.Unit{
						DefenseCavalry: 15,
					},
					Marcher: model.Unit{
						DefenseArcher: 10,
					},
					Heavy: model.Unit{
						BuildTime: 30.1234,
						Pop:       6,
					},
					Ram: model.Unit{
						Speed:  30.849,
						Attack: 1,
					},
					Catapult: model.Unit{
						Defense:        12,
						DefenseCavalry: 123,
					},
					Knight: model.Unit{
						DefenseArcher: 1234,
						Carry:         10,
					},
					Snob: model.Unit{
						BuildTime: 1234.4,
						Pop:       100,
						Speed:     35,
					},
					Militia: model.Unit{
						BuildTime:      1,
						Pop:            2,
						Speed:          3,
						Attack:         4,
						Defense:        5,
						DefenseCavalry: 6,
						DefenseArcher:  7,
						Carry:          8,
					},
				},
			},
		},
		{
			name: "ERR: version not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			key:            "pl151",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
		},
		{
			name: "ERR: server not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{}, domain.ServerNotFoundError{Key: "pl1512"})
			},
			versionCode:    version.Code,
			key:            "pl1512",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: "server (key=pl1512) not found",
				},
			},
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			tt.setup(versionSvc, serverSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
			)

			target := fmt.Sprintf("/v1/versions/%s/servers/%s/unit-info", tt.versionCode, tt.key)
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}

func TestServer_getBuildingInfo(t *testing.T) {
	t.Parallel()

	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	tests := []struct {
		name             string
		setup            func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService)
		versionCode      string
		key              string
		expectedStatus   int
		target           any
		expectedResponse any
	}{
		{
			name: "OK",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{
					BuildingInfo: domain.BuildingInfo{
						Main: domain.Building{
							MaxLevel: 124,
							Wood:     12224,
						},
						Barracks: domain.Building{
							MinLevel: 22,
							Stone:    12341,
						},
						Stable: domain.Building{
							Wood: 22241,
							Pop:  111,
						},
						Garage: domain.Building{
							Stone:      11121,
							WoodFactor: 0.99,
						},
						Watchtower: domain.Building{
							Iron:        123412,
							StoneFactor: 0.878,
						},
						Snob: domain.Building{
							Pop:       789,
							PopFactor: 0.11,
						},
						Smith: domain.Building{
							BuildTime:       23.49999,
							BuildTimeFactor: 0.999,
						},
						Place: domain.Building{
							MaxLevel: 112,
						},
						Statue: domain.Building{
							MinLevel: 123,
						},
						Market: domain.Building{
							Pop: 1234,
						},
						Wood: domain.Building{
							WoodFactor: 0.998,
						},
						Stone: domain.Building{
							StoneFactor: 0.777,
						},
						Iron: domain.Building{
							IronFactor: 0.5689,
						},
						Farm: domain.Building{
							PopFactor: 0.9,
						},
						Storage: domain.Building{
							MaxLevel: 31,
						},
						Hide: domain.Building{
							MinLevel: 1,
						},
						Wall: domain.Building{
							MaxLevel:        20,
							MinLevel:        0,
							Wood:            1234,
							Stone:           1231,
							Iron:            3214,
							Pop:             1,
							WoodFactor:      0.999,
							StoneFactor:     0.998,
							IronFactor:      0.977,
							PopFactor:       2,
							BuildTime:       30.5,
							BuildTimeFactor: 0.5,
						},
					},
				}, nil)
			},
			versionCode:    version.Code,
			key:            "pl151",
			expectedStatus: http.StatusOK,
			target:         &model.GetBuildingInfoResp{},
			expectedResponse: &model.GetBuildingInfoResp{
				Data: model.BuildingInfo{
					Main: model.Building{
						MaxLevel: 124,
						Wood:     12224,
					},
					Barracks: model.Building{
						MinLevel: 22,
						Stone:    12341,
					},
					Stable: model.Building{
						Wood: 22241,
						Pop:  111,
					},
					Garage: model.Building{
						Stone:      11121,
						WoodFactor: 0.99,
					},
					Watchtower: model.Building{
						Iron:        123412,
						StoneFactor: 0.878,
					},
					Snob: model.Building{
						Pop:       789,
						PopFactor: 0.11,
					},
					Smith: model.Building{
						BuildTime:       23.49999,
						BuildTimeFactor: 0.999,
					},
					Place: model.Building{
						MaxLevel: 112,
					},
					Statue: model.Building{
						MinLevel: 123,
					},
					Market: model.Building{
						Pop: 1234,
					},
					Wood: model.Building{
						WoodFactor: 0.998,
					},
					Stone: model.Building{
						StoneFactor: 0.777,
					},
					Iron: model.Building{
						IronFactor: 0.5689,
					},
					Farm: model.Building{
						PopFactor: 0.9,
					},
					Storage: model.Building{
						MaxLevel: 31,
					},
					Hide: model.Building{
						MinLevel: 1,
					},
					Wall: model.Building{
						MaxLevel:        20,
						MinLevel:        0,
						Wood:            1234,
						Stone:           1231,
						Iron:            3214,
						Pop:             1,
						WoodFactor:      0.999,
						StoneFactor:     0.998,
						IronFactor:      0.977,
						PopFactor:       2,
						BuildTime:       30.5,
						BuildTimeFactor: 0.5,
					},
				},
			},
		},
		{
			name: "ERR: version not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			key:            "pl151",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
		},
		{
			name: "ERR: server not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{}, domain.ServerNotFoundError{Key: "pl1512"})
			},
			versionCode:    version.Code,
			key:            "pl1512",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: "server (key=pl1512) not found",
				},
			},
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			tt.setup(versionSvc, serverSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
			)

			target := fmt.Sprintf("/v1/versions/%s/servers/%s/building-info", tt.versionCode, tt.key)
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}
