package rest

import (
	"context"
	"net/http"

	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
	"github.com/go-chi/chi/v5"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
)

type tribeCtxKey struct{}

//counterfeiter:generate -o internal/mock/tribe_service.gen.go . TribeService
type TribeService interface {
	ListCount(ctx context.Context, params domain.ListTribesParams) ([]domain.Tribe, int64, error)
	GetByServerKeyAndID(ctx context.Context, serverKey string, id int64) (domain.Tribe, error)
}

type tribe struct {
	svc TribeService
}

// @ID          listTribes
// @Summary     List tribes
// @Description List all tribes
// @Tags        versions,servers,tribes
// @Produce     json
// @Success     200         {object}  model.ListTribesResp
// @Success     400         {object}  model.ErrorResp
// @Success     404         {object}  model.ErrorResp
// @Failure     500         {object}  model.ErrorResp
// @Header      200         {integer} X-Total-Count "Total number of records"
// @Param       versionCode path      string        true  "Version code"
// @Param       serverKey   path      string        true  "Server key"
// @Param       deleted     query     boolean       false "true=only deleted tribes, false=only existing tribes, by default both existing and deleted tribes are returned"
// @Param       offset      query     int           false "specifies where to start a page" minimum(0) default(0)
// @Param       limit       query     int           false "page size"                       minimum(1) maximum(200) default(200)
// @Param       tag         query     []string      false "tribe tags, max 20"
// @Param       sort        query     []string      false "format: field:direction, default: [id:asc]" Enums(id:asc,id:desc,scoreAtt:asc,scoreAtt:desc,scoreDef:asc,scoreDef:desc,scoreTotal:asc,scoreTotal:desc,points:asc,points:desc,dominance:asc,dominance:desc,deletedAt:asc,deletedAt:desc)
// @Router      /versions/{versionCode}/servers/{serverKey}/tribes [get]
func (t *tribe) list(w http.ResponseWriter, r *http.Request) {
	var err error
	ctx := r.Context()
	srv, _ := serverFromContext(ctx)
	params := domain.ListTribesParams{
		ServerKeys: []string{srv.Key},
	}
	query := queryParams[domain.TribeSort]{
		values:  r.URL.Query(),
		newSort: domain.NewTribeSort,
	}

	params.Tags = query.strings("tag")

	params.Deleted, err = query.nullBool("deleted")
	if err != nil {
		renderErr(w, err)
		return
	}

	params.Pagination, err = query.pagination()
	if err != nil {
		renderErr(w, err)
		return
	}

	params.Sort, err = query.sort()
	if err != nil {
		renderErr(w, err)
		return
	}

	tribes, count, err := t.svc.ListCount(ctx, params)
	if err != nil {
		renderErr(w, err)
		return
	}

	setTotalCountHeader(w.Header(), count)
	renderJSON(w, http.StatusOK, model.NewListTribesResp(tribes))
}

// @ID          getTribe
// @Summary     Get a tribe
// @Description Get a tribe
// @Tags        versions,servers,tribes
// @Produce     json
// @Success     200         {object} model.GetTribeResp
// @Failure     400         {object} model.ErrorResp
// @Failure     404         {object} model.ErrorResp
// @Failure     500         {object} model.ErrorResp
// @Param       versionCode path     string  true "Version code"
// @Param       serverKey   path     string  true "Server key"
// @Param       tribeId     path     integer true "Tribe ID"
// @Router      /versions/{versionCode}/servers/{serverKey}/tribes/{tribeId} [get]
func (t *tribe) getByID(w http.ResponseWriter, r *http.Request) {
	trb, _ := tribeFromContext(r.Context())
	renderJSON(w, http.StatusOK, model.NewGetTribeResp(trb))
}

func tribeMiddleware(svc TribeService) func(http.Handler) http.Handler {
	return func(next http.Handler) http.Handler {
		return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
			ctx := r.Context()
			srv, _ := serverFromContext(ctx)
			routeCtx := chi.RouteContext(ctx)

			tribeID, err := urlParamInt64(routeCtx, "tribeId")
			if err != nil {
				renderErr(w, err)
				return
			}

			trb, err := svc.GetByServerKeyAndID(ctx, srv.Key, tribeID)
			if err != nil {
				renderErr(w, err)
				return
			}

			next.ServeHTTP(w, r.WithContext(tribeToContext(ctx, trb)))
		})
	}
}

func tribeToContext(ctx context.Context, trb domain.Tribe) context.Context {
	return context.WithValue(ctx, tribeCtxKey{}, trb)
}

func tribeFromContext(ctx context.Context) (domain.Tribe, bool) {
	trb, ok := ctx.Value(tribeCtxKey{}).(domain.Tribe)
	return trb, ok
}
