package rest

import (
	"context"
	"net/http"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
)

//counterfeiter:generate -o internal/mock/tribe_change_service.gen.go . TribeChangeService
type TribeChangeService interface {
	ListCountWithRelations(ctx context.Context, params domain.ListTribeChangesParams) ([]domain.TribeChangeWithRelations, int64, error)
}

type tribeChange struct {
	svc TribeChangeService
}

// @ID          listPlayerTribeChanges
// @Summary     List tribe changes of a particular player
// @Description List tribe changes of a particular player
// @Tags        versions,servers,players,tribe_changes
// @Produce     json
// @Success     200         {object}  model.ListTribeChangesResp
// @Success     400         {object}  model.ErrorResp
// @Success     404         {object}  model.ErrorResp
// @Failure     500         {object}  model.ErrorResp
// @Header      200         {integer} X-Total-Count "Total number of records"
// @Param       versionCode path      string        true  "Version code"
// @Param       serverKey   path      string        true  "Server key"
// @Param       playerId    path      integer       true  "Player ID"
// @Param       offset      query     int           false "specifies where to start a page"                          minimum(0) default(0)
// @Param       limit       query     int           false "page size"                                                minimum(1) maximum(100) default(100)
// @Param       sort        query     []string      false "format: field:direction, default: [createdAt:asc,id:asc]" Enums(id:asc,id:desc,createdAt:asc,createdAt:desc)
// @Router      /versions/{versionCode}/servers/{serverKey}/players/{playerId}/tribe-changes [get]
func (t *tribeChange) listPlayer(w http.ResponseWriter, r *http.Request) {
	t.renderList(w, r, func(r *http.Request, params domain.ListTribeChangesParams) domain.ListTribeChangesParams {
		pl, _ := playerFromContext(r.Context())
		params.PlayerIDs = []int64{pl.ID}
		return params
	})
}

// @ID          listTribeTribeChanges
// @Summary     List tribe changes of a particular tribe
// @Description List tribe changes of a particular tribe
// @Tags        versions,servers,tribes,tribe_changes
// @Produce     json
// @Success     200         {object}  model.ListTribeChangesResp
// @Success     400         {object}  model.ErrorResp
// @Success     404         {object}  model.ErrorResp
// @Failure     500         {object}  model.ErrorResp
// @Header      200         {integer} X-Total-Count "Total number of records"
// @Param       versionCode path      string        true  "Version code"
// @Param       serverKey   path      string        true  "Server key"
// @Param       tribeId     path      integer       true  "Tribe ID"
// @Param       offset      query     int           false "specifies where to start a page"                          minimum(0) default(0)
// @Param       limit       query     int           false "page size"                                                minimum(1) maximum(100) default(100)
// @Param       sort        query     []string      false "format: field:direction, default: [createdAt:asc,id:asc]" Enums(id:asc,id:desc,createdAt:asc,createdAt:desc)
// @Router      /versions/{versionCode}/servers/{serverKey}/tribes/{tribeId}/tribe-changes [get]
func (t *tribeChange) listTribe(w http.ResponseWriter, r *http.Request) {
	t.renderList(w, r, func(r *http.Request, params domain.ListTribeChangesParams) domain.ListTribeChangesParams {
		trb, _ := tribeFromContext(r.Context())
		params.TribeIDs = []int64{trb.ID}
		return params
	})
}

func (t *tribeChange) renderList(
	w http.ResponseWriter,
	r *http.Request,
	paramsModifierFunc func(*http.Request, domain.ListTribeChangesParams) domain.ListTribeChangesParams,
) {
	var err error
	ctx := r.Context()
	srv, _ := serverFromContext(ctx)
	query := queryParams[domain.TribeChangeSort]{
		values:  r.URL.Query(),
		newSort: domain.NewTribeChangeSort,
	}

	params := domain.ListTribeChangesParams{
		ServerKeys: []string{srv.Key},
	}

	params.Pagination, err = query.pagination()
	if err != nil {
		renderErr(w, err)
		return
	}

	params.Sort, err = query.sort()
	if err != nil {
		renderErr(w, err)
		return
	}

	tcs, count, err := t.svc.ListCountWithRelations(ctx, paramsModifierFunc(r, params))
	if err != nil {
		renderErr(w, err)
		return
	}

	setTotalCountHeader(w.Header(), count)
	renderJSON(w, http.StatusOK, model.NewListTribeChangesResp(tcs))
}
