package rest_test

import (
	"context"
	"fmt"
	"net/http"
	"net/url"
	"strconv"
	"testing"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/mock"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
	"github.com/google/go-cmp/cmp"
)

func TestTribeChange_listPlayer(t *testing.T) {
	t.Parallel()

	now := time.Now()
	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	server := domain.Server{
		Key:         "pl151",
		URL:         "https://pl151.plemiona.pl",
		Open:        true,
		Special:     false,
		VersionCode: "pl",
	}
	player := domain.PlayerWithRelations{
		Player: domain.Player{
			BasePlayer: domain.BasePlayer{
				ID:         124,
				Name:       "player 124",
				ProfileURL: "player-124",
			},
		},
	}
	playerMeta := domain.PlayerMeta{
		ID:         player.ID,
		Name:       player.Name,
		ProfileURL: player.ProfileURL,
	}
	playerID := strconv.FormatInt(player.ID, 10)

	tests := []struct {
		name  string
		setup func(
			versionSvc *mock.FakeVersionService,
			serverSvc *mock.FakeServerService,
			playerSvc *mock.FakePlayerService,
			tribeChangeSvc *mock.FakeTribeChangeService,
		)
		versionCode        string
		serverKey          string
		playerID           string
		queryParams        url.Values
		expectedStatus     int
		target             any
		expectedResponse   any
		expectedTotalCount string
	}{
		{
			name: "OK: without params",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
				tribeChangeSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListTribeChangesParams,
				) ([]domain.TribeChangeWithRelations, int64, error) {
					expectedParams := domain.ListTribeChangesParams{
						ServerKeys: []string{server.Key},
						PlayerIDs:  []int64{player.ID},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					tcs := []domain.TribeChangeWithRelations{
						{
							TribeChange: domain.TribeChange{
								ID:         111,
								PlayerID:   player.ID,
								NewTribeID: 1113,
								OldTribeID: 1114,
								ServerKey:  server.Key,
								CreatedAt:  now,
							},
							Player: playerMeta,
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         1113,
									Name:       "name-1113",
									Tag:        "tag-1113",
									ProfileURL: "profile-1113",
								},
							},
							OldTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         1114,
									Name:       "name-1114",
									Tag:        "tag-1114",
									ProfileURL: "profile-1114",
								},
							},
						},
						{
							TribeChange: domain.TribeChange{
								ID:         112,
								PlayerID:   player.ID,
								NewTribeID: 1120,
								OldTribeID: 0,
								ServerKey:  server.Key,
								CreatedAt:  now.Add(time.Hour),
							},
							Player: playerMeta,
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         1120,
									Name:       "name-1120",
									Tag:        "tag-1120",
									ProfileURL: "profile-1120",
								},
							},
							OldTribe: domain.NullTribeMeta{
								Valid: false,
							},
						},
						{
							TribeChange: domain.TribeChange{
								ID:         113,
								PlayerID:   player.ID,
								NewTribeID: 0,
								OldTribeID: 1114,
								ServerKey:  server.Key,
								CreatedAt:  now.Add(2 * time.Hour),
							},
							Player: playerMeta,
							NewTribe: domain.NullTribeMeta{
								Valid: false,
							},
							OldTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         1114,
									Name:       "name-1114",
									Tag:        "tag-1114",
									ProfileURL: "profile-1114",
								},
							},
						},
					}
					return tcs, int64(len(tcs)), nil
				})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			playerID:       playerID,
			queryParams:    nil,
			expectedStatus: http.StatusOK,
			target:         &model.ListTribeChangesResp{},
			expectedResponse: &model.ListTribeChangesResp{
				Data: []model.TribeChange{
					{
						ID:        111,
						CreatedAt: now,
						Player: model.PlayerMeta{
							ID:         player.ID,
							Name:       player.Name,
							ProfileURL: player.ProfileURL,
							Tribe: model.NullTribeMeta{
								Valid: true,
								Tribe: model.TribeMeta{
									ID:         1114,
									Name:       "name-1114",
									Tag:        "tag-1114",
									ProfileURL: "profile-1114",
								},
							},
						},
						NewTribe: model.NullTribeMeta{
							Valid: true,
							Tribe: model.TribeMeta{
								ID:         1113,
								Name:       "name-1113",
								Tag:        "tag-1113",
								ProfileURL: "profile-1113",
							},
						},
					},
					{
						ID:        112,
						CreatedAt: now.Add(time.Hour),
						Player: model.PlayerMeta{
							ID:         player.ID,
							Name:       player.Name,
							ProfileURL: player.ProfileURL,
						},
						NewTribe: model.NullTribeMeta{
							Valid: true,
							Tribe: model.TribeMeta{
								ID:         1120,
								Name:       "name-1120",
								Tag:        "tag-1120",
								ProfileURL: "profile-1120",
							},
						},
					},
					{
						ID:        113,
						CreatedAt: now.Add(2 * time.Hour),
						Player: model.PlayerMeta{
							ID:         player.ID,
							Name:       player.Name,
							ProfileURL: player.ProfileURL,
							Tribe: model.NullTribeMeta{
								Valid: true,
								Tribe: model.TribeMeta{
									ID:         1114,
									Name:       "name-1114",
									Tag:        "tag-1114",
									ProfileURL: "profile-1114",
								},
							},
						},
					},
				},
			},
			expectedTotalCount: "3",
		},
		{
			name: "OK: limit=1,offset=100,sort=[date:desc,id:desc]",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
				tribeChangeSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListTribeChangesParams,
				) ([]domain.TribeChangeWithRelations, int64, error) {
					expectedParams := domain.ListTribeChangesParams{
						ServerKeys: []string{server.Key},
						PlayerIDs:  []int64{player.ID},
						Pagination: domain.Pagination{
							Limit:  1,
							Offset: 100,
						},
						Sort: []domain.TribeChangeSort{
							{By: domain.TribeChangeSortByCreatedAt, Direction: domain.SortDirectionDESC},
							{By: domain.TribeChangeSortByID, Direction: domain.SortDirectionDESC},
						},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					tcs := []domain.TribeChangeWithRelations{
						{
							TribeChange: domain.TribeChange{
								ID:         111,
								PlayerID:   player.ID,
								NewTribeID: 1113,
								OldTribeID: 1114,
								ServerKey:  server.Key,
								CreatedAt:  now,
							},
							Player: playerMeta,
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         1113,
									Name:       "name-1113",
									Tag:        "tag-1113",
									ProfileURL: "profile-1113",
								},
							},
							OldTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         1114,
									Name:       "name-1114",
									Tag:        "tag-1114",
									ProfileURL: "profile-1114",
								},
							},
						},
					}
					return tcs, int64(len(tcs)) + int64(params.Pagination.Offset), nil
				})
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"limit":  []string{"1"},
				"offset": []string{"100"},
				"sort":   []string{"createdAt:desc", "id:desc"},
			},
			expectedStatus: http.StatusOK,
			target:         &model.ListTribeChangesResp{},
			expectedResponse: &model.ListTribeChangesResp{
				Data: []model.TribeChange{
					{
						ID:        111,
						CreatedAt: now,
						Player: model.PlayerMeta{
							ID:         player.ID,
							Name:       player.Name,
							ProfileURL: player.ProfileURL,
							Tribe: model.NullTribeMeta{
								Valid: true,
								Tribe: model.TribeMeta{
									ID:         1114,
									Name:       "name-1114",
									Tag:        "tag-1114",
									ProfileURL: "profile-1114",
								},
							},
						},
						NewTribe: model.NullTribeMeta{
							Valid: true,
							Tribe: model.TribeMeta{
								ID:         1113,
								Name:       "name-1113",
								Tag:        "tag-1113",
								ProfileURL: "profile-1113",
							},
						},
					},
				},
			},
			expectedTotalCount: "101",
		},
		{
			name: "ERR: limit is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"limit": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "limit: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: offset is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"offset": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "offset: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - invalid format",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc", "test"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: parsing \"test\": invalid syntax, expected field:direction",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported field",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc", "test:asc"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: unsupported sort by: \"test\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported direction",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc2"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[0]: unsupported sort direction: \"asc2\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: player id is not a valid int64",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			playerID:       "asdf",
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "playerId: strconv.ParseInt: parsing \"asdf\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: version not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			serverKey:      server.Key,
			playerID:       playerID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: server not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{}, domain.ServerNotFoundError{Key: server.Key + "2"})
			},
			versionCode:    version.Code,
			serverKey:      server.Key + "2",
			playerID:       playerID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("server (key=%s) not found", server.Key+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: player not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(domain.PlayerWithRelations{}, domain.PlayerNotFoundError{ID: 1241})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			playerID:       playerID + "1",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("player (id=%s) not found", playerID+"1"),
				},
			},
			expectedTotalCount: "",
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			tribeChangeSvc := &mock.FakeTribeChangeService{}
			playerSvc := &mock.FakePlayerService{}
			tt.setup(versionSvc, serverSvc, playerSvc, tribeChangeSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
				withTribeChangeService(tribeChangeSvc),
				withPlayerService(playerSvc),
			)

			target := fmt.Sprintf(
				"/v1/versions/%s/servers/%s/players/%s/tribe-changes?%s",
				tt.versionCode,
				tt.serverKey,
				tt.playerID,
				tt.queryParams.Encode(),
			)
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertTotalCount(t, resp.Header, tt.expectedTotalCount)
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}

func TestTribeChange_listTribe(t *testing.T) {
	t.Parallel()

	now := time.Now()
	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	server := domain.Server{
		Key:         "pl151",
		URL:         "https://pl151.plemiona.pl",
		Open:        true,
		Special:     false,
		VersionCode: "pl",
	}
	tribe := domain.Tribe{
		BaseTribe: domain.BaseTribe{
			ID:         124,
			Name:       "tribe 124",
			Tag:        "tag 124",
			ProfileURL: "tribe-124",
		},
	}
	tribeMeta := domain.TribeMeta{
		ID:         tribe.ID,
		Name:       tribe.Name,
		Tag:        tribe.Tag,
		ProfileURL: tribe.ProfileURL,
	}
	tribeID := strconv.FormatInt(tribe.ID, 10)

	tests := []struct {
		name  string
		setup func(
			versionSvc *mock.FakeVersionService,
			serverSvc *mock.FakeServerService,
			tribeSvc *mock.FakeTribeService,
			tribeChangeSvc *mock.FakeTribeChangeService,
		)
		versionCode        string
		serverKey          string
		tribeID            string
		queryParams        url.Values
		expectedStatus     int
		target             any
		expectedResponse   any
		expectedTotalCount string
	}{
		{
			name: "OK: without params",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
				tribeChangeSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListTribeChangesParams,
				) ([]domain.TribeChangeWithRelations, int64, error) {
					expectedParams := domain.ListTribeChangesParams{
						ServerKeys: []string{server.Key},
						TribeIDs:   []int64{tribe.ID},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					tcs := []domain.TribeChangeWithRelations{
						{
							TribeChange: domain.TribeChange{
								ID:         111,
								PlayerID:   1111,
								NewTribeID: 1113,
								OldTribeID: tribe.ID,
								ServerKey:  server.Key,
								CreatedAt:  now,
							},
							Player: domain.PlayerMeta{
								ID:         1111,
								Name:       "name-1111",
								ProfileURL: "profile-1111",
							},
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         1113,
									Name:       "name-1113",
									Tag:        "tag-1113",
									ProfileURL: "profile-1113",
								},
							},
							OldTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: tribeMeta,
							},
						},
						{
							TribeChange: domain.TribeChange{
								ID:         112,
								PlayerID:   1235,
								NewTribeID: tribe.ID,
								OldTribeID: 0,
								ServerKey:  server.Key,
								CreatedAt:  now.Add(time.Hour),
							},
							Player: domain.PlayerMeta{
								ID:         1235,
								Name:       "name-1235",
								ProfileURL: "profile-1235",
							},
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: tribeMeta,
							},
							OldTribe: domain.NullTribeMeta{
								Valid: false,
							},
						},
						{
							TribeChange: domain.TribeChange{
								ID:         113,
								PlayerID:   1235,
								NewTribeID: 0,
								OldTribeID: tribe.ID,
								ServerKey:  server.Key,
								CreatedAt:  now.Add(2 * time.Hour),
							},
							Player: domain.PlayerMeta{
								ID:         1235,
								Name:       "name-1235",
								ProfileURL: "profile-1235",
							},
							NewTribe: domain.NullTribeMeta{
								Valid: false,
							},
							OldTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: tribeMeta,
							},
						},
					}
					return tcs, int64(len(tcs)), nil
				})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			tribeID:        tribeID,
			queryParams:    nil,
			expectedStatus: http.StatusOK,
			target:         &model.ListTribeChangesResp{},
			expectedResponse: &model.ListTribeChangesResp{
				Data: []model.TribeChange{
					{
						ID:        111,
						CreatedAt: now,
						Player: model.PlayerMeta{
							ID:         1111,
							Name:       "name-1111",
							ProfileURL: "profile-1111",
							Tribe: model.NullTribeMeta{
								Valid: true,
								Tribe: model.TribeMeta{
									ID:         tribe.ID,
									Name:       tribe.Name,
									Tag:        tribe.Tag,
									ProfileURL: tribe.ProfileURL,
								},
							},
						},
						NewTribe: model.NullTribeMeta{
							Valid: true,
							Tribe: model.TribeMeta{
								ID:         1113,
								Name:       "name-1113",
								Tag:        "tag-1113",
								ProfileURL: "profile-1113",
							},
						},
					},
					{
						ID:        112,
						CreatedAt: now.Add(time.Hour),
						Player: model.PlayerMeta{
							ID:         1235,
							Name:       "name-1235",
							ProfileURL: "profile-1235",
						},
						NewTribe: model.NullTribeMeta{
							Valid: true,
							Tribe: model.TribeMeta{
								ID:         tribe.ID,
								Name:       tribe.Name,
								Tag:        tribe.Tag,
								ProfileURL: tribe.ProfileURL,
							},
						},
					},
					{
						ID:        113,
						CreatedAt: now.Add(2 * time.Hour),
						Player: model.PlayerMeta{
							ID:         1235,
							Name:       "name-1235",
							ProfileURL: "profile-1235",
							Tribe: model.NullTribeMeta{
								Valid: true,
								Tribe: model.TribeMeta{
									ID:         tribe.ID,
									Name:       tribe.Name,
									Tag:        tribe.Tag,
									ProfileURL: tribe.ProfileURL,
								},
							},
						},
					},
				},
			},
			expectedTotalCount: "3",
		},
		{
			name: "OK: limit=1,offset=100,sort=[date:desc,id:desc]",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
				tribeChangeSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListTribeChangesParams,
				) ([]domain.TribeChangeWithRelations, int64, error) {
					expectedParams := domain.ListTribeChangesParams{
						ServerKeys: []string{server.Key},
						TribeIDs:   []int64{tribe.ID},
						Pagination: domain.Pagination{
							Limit:  1,
							Offset: 100,
						},
						Sort: []domain.TribeChangeSort{
							{By: domain.TribeChangeSortByCreatedAt, Direction: domain.SortDirectionDESC},
							{By: domain.TribeChangeSortByID, Direction: domain.SortDirectionDESC},
						},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					tcs := []domain.TribeChangeWithRelations{
						{
							TribeChange: domain.TribeChange{
								ID:         111,
								PlayerID:   tribe.ID,
								NewTribeID: 1113,
								OldTribeID: 1114,
								ServerKey:  server.Key,
								CreatedAt:  now,
							},
							Player: domain.PlayerMeta{
								ID:         1111,
								Name:       "name-1111",
								ProfileURL: "profile-1111",
							},
							NewTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: tribeMeta,
							},
							OldTribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         1114,
									Name:       "name-1114",
									Tag:        "tag-1114",
									ProfileURL: "profile-1114",
								},
							},
						},
					}
					return tcs, int64(len(tcs)) + int64(params.Pagination.Offset), nil
				})
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"limit":  []string{"1"},
				"offset": []string{"100"},
				"sort":   []string{"createdAt:desc", "id:desc"},
			},
			expectedStatus: http.StatusOK,
			target:         &model.ListTribeChangesResp{},
			expectedResponse: &model.ListTribeChangesResp{
				Data: []model.TribeChange{
					{
						ID:        111,
						CreatedAt: now,
						Player: model.PlayerMeta{
							ID:         1111,
							Name:       "name-1111",
							ProfileURL: "profile-1111",
							Tribe: model.NullTribeMeta{
								Valid: true,
								Tribe: model.TribeMeta{
									ID:         1114,
									Name:       "name-1114",
									Tag:        "tag-1114",
									ProfileURL: "profile-1114",
								},
							},
						},
						NewTribe: model.NullTribeMeta{
							Valid: true,
							Tribe: model.TribeMeta{
								ID:         tribe.ID,
								Name:       tribe.Name,
								Tag:        tribe.Tag,
								ProfileURL: tribe.ProfileURL,
							},
						},
					},
				},
			},
			expectedTotalCount: "101",
		},
		{
			name: "ERR: limit is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"limit": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "limit: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: offset is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"offset": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "offset: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - invalid format",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc", "test"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: parsing \"test\": invalid syntax, expected field:direction",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported field",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc", "test:asc"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: unsupported sort by: \"test\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported direction",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"sort": []string{"createdAt:asc2"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[0]: unsupported sort direction: \"asc2\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: tribe id is not a valid int64",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			tribeID:        "asdf",
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "tribeId: strconv.ParseInt: parsing \"asdf\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: version not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			serverKey:      server.Key,
			tribeID:        tribeID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: server not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{}, domain.ServerNotFoundError{Key: server.Key + "2"})
			},
			versionCode:    version.Code,
			serverKey:      server.Key + "2",
			tribeID:        tribeID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("server (key=%s) not found", server.Key+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: tribe not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeChangeSvc *mock.FakeTribeChangeService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(domain.Tribe{}, domain.TribeNotFoundError{ID: 1241})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			tribeID:        tribeID + "1",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("tribe (id=%s) not found", tribeID+"1"),
				},
			},
			expectedTotalCount: "",
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			tribeChangeSvc := &mock.FakeTribeChangeService{}
			tribeSvc := &mock.FakeTribeService{}
			tt.setup(versionSvc, serverSvc, tribeSvc, tribeChangeSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
				withTribeChangeService(tribeChangeSvc),
				withTribeService(tribeSvc),
			)

			target := fmt.Sprintf(
				"/v1/versions/%s/servers/%s/tribes/%s/tribe-changes?%s",
				tt.versionCode,
				tt.serverKey,
				tt.tribeID,
				tt.queryParams.Encode(),
			)
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertTotalCount(t, resp.Header, tt.expectedTotalCount)
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}
