package rest

import (
	"context"
	"net/http"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
)

//counterfeiter:generate -o internal/mock/tribe_snapshot_service.gen.go . TribeSnapshotService
type TribeSnapshotService interface {
	ListCount(ctx context.Context, params domain.ListTribeSnapshotsParams) ([]domain.TribeSnapshot, int64, error)
}

type tribeSnapshot struct {
	svc TribeSnapshotService
}

// @ID          listTribeSnapshotsDeprecated
// @Summary     Deprecated, use /versions/{versionCode}/servers/{serverKey}/tribes/{tribeId}/snapshots instead.
// @Description Deprecated, use /versions/{versionCode}/servers/{serverKey}/tribes/{tribeId}/snapshots instead.
// @Tags        versions,servers,tribes,snapshots
// @Produce     json
// @Success     200         {object}  model.ListTribeSnapshotsResp
// @Success     400         {object}  model.ErrorResp
// @Success     404         {object}  model.ErrorResp
// @Failure     500         {object}  model.ErrorResp
// @Header      200         {integer} X-Total-Count "Total number of records"
// @Param       versionCode path      string        true  "Version code"
// @Param       serverKey   path      string        true  "Server key"
// @Param       tribeId     path      integer       true  "Tribe ID"
// @Param       offset      query     int           false "specifies where to start a page"                     minimum(0) default(0)
// @Param       limit       query     int           false "page size"                                           minimum(1) maximum(100) default(100)
// @Param       sort        query     []string      false "format: field:direction, default: [date:asc,id:asc]" Enums(id:asc,id:desc,date:asc,date:desc)
// @Router      /versions/{versionCode}/servers/{serverKey}/tribes/{tribeId}/history [get]
// @Deprecated
func (t *tribeSnapshot) history(w http.ResponseWriter, r *http.Request) {
	t.list(w, r)
}

// @ID          listTribeSnapshots
// @Summary     List snapshots of a particular tribe
// @Description List snapshots of a particular tribe
// @Tags        versions,servers,tribes,snapshots
// @Produce     json
// @Success     200         {object}  model.ListTribeSnapshotsResp
// @Success     400         {object}  model.ErrorResp
// @Success     404         {object}  model.ErrorResp
// @Failure     500         {object}  model.ErrorResp
// @Header      200         {integer} X-Total-Count "Total number of records"
// @Param       versionCode path      string        true  "Version code"
// @Param       serverKey   path      string        true  "Server key"
// @Param       tribeId     path      integer       true  "Tribe ID"
// @Param       offset      query     int           false "specifies where to start a page"                     minimum(0) default(0)
// @Param       limit       query     int           false "page size"                                           minimum(1) maximum(100) default(100)
// @Param       sort        query     []string      false "format: field:direction, default: [date:asc,id:asc]" Enums(id:asc,id:desc,date:asc,date:desc)
// @Router      /versions/{versionCode}/servers/{serverKey}/tribes/{tribeId}/snapshots [get]
func (t *tribeSnapshot) list(w http.ResponseWriter, r *http.Request) {
	var err error
	ctx := r.Context()
	srv, _ := serverFromContext(ctx)

	params := domain.ListTribeSnapshotsParams{
		ServerKeys: []string{srv.Key},
	}
	query := queryParams[domain.TribeSnapshotSort]{
		values:  r.URL.Query(),
		newSort: domain.NewTribeSnapshotSort,
	}

	trb, _ := tribeFromContext(r.Context())
	params.TribeIDs = []int64{trb.ID}

	params.Pagination, err = query.pagination()
	if err != nil {
		renderErr(w, err)
		return
	}

	params.Sort, err = query.sort()
	if err != nil {
		renderErr(w, err)
		return
	}

	snapshots, count, err := t.svc.ListCount(ctx, params)
	if err != nil {
		renderErr(w, err)
		return
	}

	setTotalCountHeader(w.Header(), count)
	renderJSON(w, http.StatusOK, model.NewListTribeSnapshotsResp(snapshots))
}
