package rest_test

import (
	"context"
	"fmt"
	"net/http"
	"net/url"
	"strconv"
	"testing"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/mock"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
	"github.com/google/go-cmp/cmp"
)

func TestTribeSnapshot_list(t *testing.T) {
	t.Parallel()

	now := time.Now()
	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	server := domain.Server{
		Key:         "pl151",
		URL:         "https://pl151.plemiona.pl",
		Open:        true,
		Special:     false,
		VersionCode: "pl",
	}
	tribe := domain.Tribe{
		BaseTribe: domain.BaseTribe{
			ID:         124,
			Name:       "tribe 124",
			Tag:        "tag 124",
			ProfileURL: "tribe-124",
		},
	}
	tribeID := strconv.FormatInt(tribe.ID, 10)

	tests := []struct {
		name  string
		setup func(
			versionSvc *mock.FakeVersionService,
			serverSvc *mock.FakeServerService,
			tribeSvc *mock.FakeTribeService,
			tribeSnapshotSvc *mock.FakeTribeSnapshotService,
		)
		versionCode        string
		serverKey          string
		tribeID            string
		queryParams        url.Values
		expectedStatus     int
		target             any
		expectedResponse   any
		expectedTotalCount string
	}{
		{
			name: "OK: without params",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeSnapshotSvc *mock.FakeTribeSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
				tribeSnapshotSvc.ListCountCalls(func(
					_ context.Context,
					params domain.ListTribeSnapshotsParams,
				) ([]domain.TribeSnapshot, int64, error) {
					expectedParams := domain.ListTribeSnapshotsParams{
						ServerKeys: []string{server.Key},
						TribeIDs:   []int64{tribe.ID},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					snapshots := []domain.TribeSnapshot{
						{
							OpponentsDefeated: domain.OpponentsDefeated{
								RankAtt:    99,
								ScoreAtt:   100,
								RankDef:    101,
								ScoreDef:   102,
								RankSup:    103,
								ScoreSup:   104,
								RankTotal:  105,
								ScoreTotal: 106,
							},
							ID:          999,
							TribeID:     tribe.ID,
							ServerKey:   server.Key,
							NumMembers:  1235,
							NumVillages: 100,
							Points:      101,
							AllPoints:   5321,
							Rank:        102,
							Dominance:   5.15,
							Date:        now,
							CreatedAt:   now.Add(-5 * time.Minute),
						},
						{
							OpponentsDefeated: domain.OpponentsDefeated{
								RankAtt:    999,
								ScoreAtt:   1000,
								RankDef:    1010,
								ScoreDef:   1020,
								RankSup:    1030,
								ScoreSup:   1040,
								RankTotal:  1050,
								ScoreTotal: 1060,
							},
							ID:          9990,
							TribeID:     tribe.ID,
							ServerKey:   server.Key,
							NumMembers:  123,
							NumVillages: 1000,
							Points:      1010,
							AllPoints:   321,
							Rank:        1020,
							Dominance:   1.15,
							Date:        now.Add(-24 * time.Hour),
							CreatedAt:   now.Add(-36 * time.Hour),
						},
					}
					return snapshots, int64(len(snapshots)), nil
				})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			tribeID:        tribeID,
			queryParams:    nil,
			expectedStatus: http.StatusOK,
			target:         &model.ListTribeSnapshotsResp{},
			expectedResponse: &model.ListTribeSnapshotsResp{
				Data: []model.TribeSnapshot{
					{
						RankAtt:     99,
						ScoreAtt:    100,
						RankDef:     101,
						ScoreDef:    102,
						RankTotal:   105,
						ScoreTotal:  106,
						ID:          999,
						NumMembers:  1235,
						NumVillages: 100,
						Points:      101,
						AllPoints:   5321,
						Rank:        102,
						Dominance:   5.15,
						Date:        model.Date{Time: now.Truncate(24 * time.Hour)},
					},
					{
						RankAtt:     999,
						ScoreAtt:    1000,
						RankDef:     1010,
						ScoreDef:    1020,
						RankTotal:   1050,
						ScoreTotal:  1060,
						ID:          9990,
						NumMembers:  123,
						NumVillages: 1000,
						Points:      1010,
						AllPoints:   321,
						Rank:        1020,
						Dominance:   1.15,
						Date:        model.Date{Time: now.Add(-24 * time.Hour).Truncate(24 * time.Hour)},
					},
				},
			},
			expectedTotalCount: "2",
		},
		{
			name: "OK: limit=1,offset=100,sort=[date:desc,id:desc]",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeSnapshotSvc *mock.FakeTribeSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
				tribeSnapshotSvc.ListCountCalls(func(
					_ context.Context,
					params domain.ListTribeSnapshotsParams,
				) ([]domain.TribeSnapshot, int64, error) {
					expectedParams := domain.ListTribeSnapshotsParams{
						ServerKeys: []string{server.Key},
						TribeIDs:   []int64{tribe.ID},
						Pagination: domain.Pagination{
							Limit:  1,
							Offset: 100,
						},
						Sort: []domain.TribeSnapshotSort{
							{By: domain.TribeSnapshotSortByDate, Direction: domain.SortDirectionDESC},
							{By: domain.TribeSnapshotSortByID, Direction: domain.SortDirectionDESC},
						},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					snapshots := []domain.TribeSnapshot{
						{
							OpponentsDefeated: domain.OpponentsDefeated{
								RankAtt:    999,
								ScoreAtt:   1000,
								RankDef:    1010,
								ScoreDef:   1020,
								RankSup:    1030,
								ScoreSup:   1040,
								RankTotal:  1050,
								ScoreTotal: 1060,
							},
							ID:          9990,
							TribeID:     tribe.ID,
							ServerKey:   server.Key,
							NumMembers:  123,
							NumVillages: 1000,
							Points:      1010,
							AllPoints:   321,
							Rank:        1020,
							Dominance:   1.15,
							Date:        now.Add(-24 * time.Hour),
							CreatedAt:   now.Add(-36 * time.Hour),
						},
					}
					return snapshots, int64(len(snapshots)) + int64(params.Pagination.Offset), nil
				})
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"limit":  []string{"1"},
				"offset": []string{"100"},
				"sort":   []string{"date:desc", "id:desc"},
			},
			expectedStatus: http.StatusOK,
			target:         &model.ListTribeSnapshotsResp{},
			expectedResponse: &model.ListTribeSnapshotsResp{
				Data: []model.TribeSnapshot{
					{
						RankAtt:     999,
						ScoreAtt:    1000,
						RankDef:     1010,
						ScoreDef:    1020,
						RankTotal:   1050,
						ScoreTotal:  1060,
						ID:          9990,
						NumMembers:  123,
						NumVillages: 1000,
						Points:      1010,
						AllPoints:   321,
						Rank:        1020,
						Dominance:   1.15,
						Date:        model.Date{Time: now.Add(-24 * time.Hour).Truncate(24 * time.Hour)},
					},
				},
			},
			expectedTotalCount: "101",
		},
		{
			name: "ERR: limit is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeSnapshotSvc *mock.FakeTribeSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"limit": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "limit: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: offset is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeSnapshotSvc *mock.FakeTribeSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"offset": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "offset: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - invalid format",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeSnapshotSvc *mock.FakeTribeSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"sort": []string{"date:asc", "test"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: parsing \"test\": invalid syntax, expected field:direction",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported field",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeSnapshotSvc *mock.FakeTribeSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"sort": []string{"date:asc", "test:asc"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[1]: unsupported sort by: \"test\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: sort - unsupported direction",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeSnapshotSvc *mock.FakeTribeSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(tribe, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			tribeID:     tribeID,
			queryParams: url.Values{
				"sort": []string{"date:asc2"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "sort[0]: unsupported sort direction: \"asc2\"",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: tribe id is not a valid int64",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeSnapshotSvc *mock.FakeTribeSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			tribeID:        "asdf",
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "tribeId: strconv.ParseInt: parsing \"asdf\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: version not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeSnapshotSvc *mock.FakeTribeSnapshotService,
			) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			serverKey:      server.Key,
			tribeID:        tribeID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: server not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeSnapshotSvc *mock.FakeTribeSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{}, domain.ServerNotFoundError{Key: server.Key + "2"})
			},
			versionCode:    version.Code,
			serverKey:      server.Key + "2",
			tribeID:        tribeID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("server (key=%s) not found", server.Key+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: tribe not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				tribeSvc *mock.FakeTribeService,
				tribeSnapshotSvc *mock.FakeTribeSnapshotService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				tribeSvc.GetByServerKeyAndIDReturns(domain.Tribe{}, domain.TribeNotFoundError{ID: 1241})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			tribeID:        tribeID + "1",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("tribe (id=%s) not found", tribeID+"1"),
				},
			},
			expectedTotalCount: "",
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			tribeSnapshotSvc := &mock.FakeTribeSnapshotService{}
			tribeSvc := &mock.FakeTribeService{}
			tt.setup(versionSvc, serverSvc, tribeSvc, tribeSnapshotSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
				withTribeSnapshotService(tribeSnapshotSvc),
				withTribeService(tribeSvc),
			)

			target := fmt.Sprintf(
				"/v1/versions/%s/servers/%s/tribes/%s/snapshots?%s",
				tt.versionCode,
				tt.serverKey,
				tt.tribeID,
				tt.queryParams.Encode(),
			)
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertTotalCount(t, resp.Header, tt.expectedTotalCount)
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}
