package rest

import (
	"context"
	"net/http"

	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"

	"github.com/go-chi/chi/v5"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
)

type versionCtxKey struct{}

//counterfeiter:generate -o internal/mock/version_service.gen.go . VersionService
type VersionService interface {
	ListCount(ctx context.Context) ([]domain.Version, int64, error)
	GetByCode(ctx context.Context, code string) (domain.Version, error)
}

type version struct {
	svc VersionService
}

// @ID          listVersions
// @Summary     List versions
// @Description List all versions
// @Tags        versions
// @Produce     json
// @Success     200 {object}  model.ListVersionsResp
// @Failure     500 {object}  model.ErrorResp
// @Header      200 {integer} X-Total-Count "Total number of records"
// @Router      /versions [get]
func (v *version) list(w http.ResponseWriter, r *http.Request) {
	versions, count, err := v.svc.ListCount(r.Context())
	if err != nil {
		renderErr(w, err)
		return
	}

	setTotalCountHeader(w.Header(), count)
	renderJSON(w, http.StatusOK, model.NewListVersionsResp(versions))
}

// @ID          getVersion
// @Summary     Get a version
// @Description Get a version
// @Tags        versions
// @Produce     json
// @Success     200         {object} model.GetVersionResp
// @Failure     404         {object} model.ErrorResp
// @Failure     500         {object} model.ErrorResp
// @Param       versionCode path     string true "Version code"
// @Router      /versions/{versionCode} [get]
func (v *version) getByCode(w http.ResponseWriter, r *http.Request) {
	ver, _ := versionFromContext(r.Context())
	renderJSON(w, http.StatusOK, model.NewGetVersionResp(ver))
}

func versionMiddleware(svc VersionService) func(http.Handler) http.Handler {
	return func(next http.Handler) http.Handler {
		return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
			ctx := r.Context()

			ver, err := svc.GetByCode(ctx, chi.URLParamFromCtx(ctx, "versionCode"))
			if err != nil {
				renderErr(w, err)
				return
			}

			next.ServeHTTP(w, r.WithContext(versionToContext(ctx, ver)))
		})
	}
}

func versionToContext(ctx context.Context, ver domain.Version) context.Context {
	return context.WithValue(ctx, versionCtxKey{}, ver)
}

func versionFromContext(ctx context.Context) (domain.Version, bool) {
	ver, ok := ctx.Value(versionCtxKey{}).(domain.Version)
	return ver, ok
}
