package rest

import (
	"context"
	"net/http"

	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
	"github.com/go-chi/chi/v5"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
)

type villageCtxKey struct{}

//counterfeiter:generate -o internal/mock/village_service.gen.go . VillageService
type VillageService interface {
	ListCountWithRelations(ctx context.Context, params domain.ListVillagesParams) ([]domain.VillageWithRelations, int64, error)
	GetByServerKeyAndIDWithRelations(ctx context.Context, serverKey string, id int64) (domain.VillageWithRelations, error)
}

type village struct {
	svc VillageService
}

// @ID          listVillages
// @Summary     List villages
// @Description List all villages
// @Tags        versions,servers,villages
// @Produce     json
// @Success     200         {object}  model.ListVillagesResp
// @Success     400         {object}  model.ErrorResp
// @Success     404         {object}  model.ErrorResp
// @Failure     500         {object}  model.ErrorResp
// @Header      200         {integer} X-Total-Count "Total number of records"
// @Param       versionCode path      string        true  "Version code"
// @Param       serverKey   path      string        true  "Server key"
// @Param       offset      query     int           false "specifies where to start a page" minimum(0) default(0)
// @Param       limit       query     int           false "page size"                       minimum(1) maximum(500) default(500)
// @Param       coords      query     []string      false "village coordinates, max 50"
// @Router      /versions/{versionCode}/servers/{serverKey}/villages [get]
func (v *village) list(w http.ResponseWriter, r *http.Request) {
	v.renderList(w, r, func(r *http.Request, params domain.ListVillagesParams) domain.ListVillagesParams {
		return params
	})
}

// @ID          listPlayerVillages
// @Summary     List villages of a particular player
// @Description List villages of a particular player
// @Tags        versions,servers,players,villages
// @Produce     json
// @Success     200         {object}  model.ListVillagesResp
// @Success     400         {object}  model.ErrorResp
// @Success     404         {object}  model.ErrorResp
// @Failure     500         {object}  model.ErrorResp
// @Header      200         {integer} X-Total-Count "Total number of records"
// @Param       versionCode path      string        true  "Version code"
// @Param       serverKey   path      string        true  "Server key"
// @Param       playerId    path      integer       true  "Player ID"
// @Param       offset      query     int           false "specifies where to start a page" minimum(0) default(0)
// @Param       limit       query     int           false "page size"                       minimum(1) maximum(500) default(500)
// @Param       coords      query     []string      false "village coordinates, max 50"
// @Router      /versions/{versionCode}/servers/{serverKey}/players/{playerId}/villages [get]
func (v *village) listPlayer(w http.ResponseWriter, r *http.Request) {
	v.renderList(w, r, func(r *http.Request, params domain.ListVillagesParams) domain.ListVillagesParams {
		p, _ := playerFromContext(r.Context())
		params.PlayerID = domain.NullInt64{
			Int64: p.ID,
			Valid: true,
		}
		return params
	})
}

func (v *village) renderList(
	w http.ResponseWriter,
	r *http.Request,
	paramsModifierFunc func(*http.Request, domain.ListVillagesParams) domain.ListVillagesParams,
) {
	var err error
	ctx := r.Context()
	srv, _ := serverFromContext(ctx)
	params := domain.ListVillagesParams{
		ServerKeys: []string{srv.Key},
	}
	query := queryParams[any]{
		values: r.URL.Query(),
	}

	params.Pagination, err = query.pagination()
	if err != nil {
		renderErr(w, err)
		return
	}

	params.Coords, err = query.coords("coords")
	if err != nil {
		renderErr(w, err)
		return
	}

	villages, count, err := v.svc.ListCountWithRelations(ctx, paramsModifierFunc(r, params))
	if err != nil {
		renderErr(w, err)
		return
	}

	setTotalCountHeader(w.Header(), count)
	renderJSON(w, http.StatusOK, model.NewListVillagesResp(villages))
}

// @ID          getVillage
// @Summary     Get a village
// @Description Get a village
// @Tags        versions,servers,villages
// @Produce     json
// @Success     200         {object} model.GetVillageResp
// @Failure     400         {object} model.ErrorResp
// @Failure     404         {object} model.ErrorResp
// @Failure     500         {object} model.ErrorResp
// @Param       versionCode path     string  true "Version code"
// @Param       serverKey   path     string  true "Server key"
// @Param       villageId   path     integer true "Village ID"
// @Router      /versions/{versionCode}/servers/{serverKey}/villages/{villageId} [get]
func (v *village) getByID(w http.ResponseWriter, r *http.Request) {
	vlg, _ := villageFromContext(r.Context())
	renderJSON(w, http.StatusOK, model.NewGetVillageResp(vlg))
}

func villageMiddleware(svc VillageService) func(http.Handler) http.Handler {
	return func(next http.Handler) http.Handler {
		return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
			ctx := r.Context()
			srv, _ := serverFromContext(ctx)
			routeCtx := chi.RouteContext(ctx)

			villageID, err := urlParamInt64(routeCtx, "villageId")
			if err != nil {
				renderErr(w, err)
				return
			}

			vlg, err := svc.GetByServerKeyAndIDWithRelations(ctx, srv.Key, villageID)
			if err != nil {
				renderErr(w, err)
				return
			}

			next.ServeHTTP(w, r.WithContext(villageToContext(ctx, vlg)))
		})
	}
}

func villageToContext(ctx context.Context, vlg domain.VillageWithRelations) context.Context {
	return context.WithValue(ctx, villageCtxKey{}, vlg)
}

func villageFromContext(ctx context.Context) (domain.VillageWithRelations, bool) {
	vlg, ok := ctx.Value(villageCtxKey{}).(domain.VillageWithRelations)
	return vlg, ok
}
