package rest_test

import (
	"context"
	"fmt"
	"net/http"
	"net/url"
	"strconv"
	"testing"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/mock"
	"gitea.dwysokinski.me/twhelp/core/internal/router/rest/internal/model"
	"github.com/google/go-cmp/cmp"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
)

func TestVillage_list(t *testing.T) {
	t.Parallel()

	now := time.Now()
	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	server := domain.Server{
		Key:         "pl151",
		URL:         "https://pl151.plemiona.pl",
		Open:        true,
		Special:     false,
		VersionCode: "pl",
	}

	tests := []struct {
		name               string
		setup              func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService)
		versionCode        string
		serverKey          string
		queryParams        url.Values
		expectedStatus     int
		target             any
		expectedResponse   any
		expectedTotalCount string
	}{
		{
			name: "OK: without params",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				villageSvc.ListCountWithRelationsCalls(func(_ context.Context, params domain.ListVillagesParams) ([]domain.VillageWithRelations, int64, error) {
					expectedParams := domain.ListVillagesParams{
						ServerKeys: []string{server.Key},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					villages := []domain.VillageWithRelations{
						{
							Village: domain.Village{
								BaseVillage: domain.BaseVillage{
									ID:         1234,
									Name:       "name",
									Points:     1,
									X:          2,
									Y:          3,
									Continent:  "K11",
									Bonus:      1,
									PlayerID:   997,
									ProfileURL: "profile-1234",
								},
								ServerKey: server.Key,
								CreatedAt: now,
							},
							Player: domain.NullPlayerMetaWithRelations{
								Valid: true,
								Player: domain.PlayerMetaWithRelations{
									PlayerMeta: domain.PlayerMeta{
										ID:         997,
										Name:       "name 997",
										ProfileURL: "profile-997",
									},
									Tribe: domain.NullTribeMeta{
										Valid: true,
										Tribe: domain.TribeMeta{
											ID:         1234,
											Name:       "name 997",
											Tag:        "tag 997",
											ProfileURL: "profile-1234",
										},
									},
								},
							},
						},
						{
							Village: domain.Village{
								BaseVillage: domain.BaseVillage{
									ID:         12345,
									Name:       "name 2",
									Points:     1,
									X:          22,
									Y:          22,
									Continent:  "K22",
									Bonus:      0,
									PlayerID:   0,
									ProfileURL: "profile-12345",
								},
								ServerKey: server.Key,
								CreatedAt: now,
							},
							Player: domain.NullPlayerMetaWithRelations{},
						},
					}
					return villages, int64(len(villages)), nil
				})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			expectedStatus: http.StatusOK,
			target:         &model.ListVillagesResp{},
			expectedResponse: &model.ListVillagesResp{
				Data: []model.Village{
					{
						ID:         1234,
						Name:       "name",
						FullName:   "name (2|3) K11",
						Points:     1,
						X:          2,
						Y:          3,
						Continent:  "K11",
						Bonus:      1,
						ProfileURL: "profile-1234",
						Player: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         997,
								Name:       "name 997",
								ProfileURL: "profile-997",
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         1234,
										Name:       "name 997",
										Tag:        "tag 997",
										ProfileURL: "profile-1234",
									},
								},
							},
						},
						CreatedAt: now,
					},
					{
						ID:         12345,
						Name:       "name 2",
						FullName:   "name 2 (22|22) K22",
						Points:     1,
						X:          22,
						Y:          22,
						Continent:  "K22",
						Bonus:      0,
						ProfileURL: "profile-12345",
						Player: model.NullPlayerMeta{
							Valid:  false,
							Player: model.PlayerMeta{},
						},
						CreatedAt: now,
					},
				},
			},
			expectedTotalCount: "2",
		},
		{
			name: "OK: limit=1, offset=15",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				villageSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListVillagesParams,
				) ([]domain.VillageWithRelations, int64, error) {
					expectedParams := domain.ListVillagesParams{
						ServerKeys: []string{server.Key},
						Pagination: domain.Pagination{
							Limit:  1,
							Offset: 15,
						},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					villages := []domain.VillageWithRelations{
						{
							Village: domain.Village{
								BaseVillage: domain.BaseVillage{
									ID:         12345,
									Name:       "name 2",
									Points:     1,
									X:          22,
									Y:          22,
									Continent:  "K22",
									Bonus:      0,
									PlayerID:   0,
									ProfileURL: "profile-12345",
								},
								ServerKey: server.Key,
								CreatedAt: now,
							},
							Player: domain.NullPlayerMetaWithRelations{},
						},
					}
					return villages, int64(len(villages)) + 12354, nil
				})
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			queryParams: url.Values{
				"limit":  []string{"1"},
				"offset": []string{"15"},
			},
			expectedStatus: http.StatusOK,
			target:         &model.ListVillagesResp{},
			expectedResponse: &model.ListVillagesResp{
				Data: []model.Village{
					{
						ID:        12345,
						Name:      "name 2",
						FullName:  "name 2 (22|22) K22",
						Points:    1,
						X:         22,
						Y:         22,
						Continent: "K22",
						Bonus:     0,
						Player: model.NullPlayerMeta{
							Valid:  false,
							Player: model.PlayerMeta{},
						},
						ProfileURL: "profile-12345",
						CreatedAt:  now,
					},
				},
			},
			expectedTotalCount: "12355",
		},
		{
			name: "OK: coords=[22|23]",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				villageSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListVillagesParams,
				) ([]domain.VillageWithRelations, int64, error) {
					expectedParams := domain.ListVillagesParams{
						ServerKeys: []string{server.Key},
						Coords:     []domain.Coords{{X: 22, Y: 23}},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						fmt.Println(diff)
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					return []domain.VillageWithRelations{
						{
							Village: domain.Village{
								BaseVillage: domain.BaseVillage{
									ID:         12345,
									Name:       "name 2",
									Points:     1,
									X:          22,
									Y:          23,
									Continent:  "K22",
									Bonus:      0,
									PlayerID:   0,
									ProfileURL: "profile-12345",
								},
								ServerKey: server.Key,
								CreatedAt: now,
							},
							Player: domain.NullPlayerMetaWithRelations{},
						},
					}, 1, nil
				})
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			queryParams: url.Values{
				"coords": []string{"22|23"},
			},
			expectedStatus: http.StatusOK,
			target:         &model.ListVillagesResp{},
			expectedResponse: &model.ListVillagesResp{
				Data: []model.Village{
					{
						ID:        12345,
						Name:      "name 2",
						FullName:  "name 2 (22|23) K22",
						Points:    1,
						X:         22,
						Y:         23,
						Continent: "K22",
						Bonus:     0,
						Player: model.NullPlayerMeta{
							Valid:  false,
							Player: model.PlayerMeta{},
						},
						ProfileURL: "profile-12345",
						CreatedAt:  now,
					},
				},
			},
			expectedTotalCount: "1",
		},
		{
			name: "ERR: limit is not a valid int32",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			queryParams: url.Values{
				"limit": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "limit: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: offset is not a valid int32",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			queryParams: url.Values{
				"offset": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "offset: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: coords - incorrect format 1",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			queryParams: url.Values{
				"coords": []string{"22|-22"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "coords[0]: " + domain.ErrIncorrectCoordsFormat.Error(),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: coords - incorrect format 2",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			queryParams: url.Values{
				"coords": []string{"22|22", "235"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "coords[1]: " + domain.ErrIncorrectCoordsFormat.Error(),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: version not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			serverKey:      server.Key,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: server not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{}, domain.ServerNotFoundError{Key: server.Key + "2"})
			},
			versionCode:    version.Code,
			serverKey:      server.Key + "2",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("server (key=%s) not found", server.Key+"2"),
				},
			},
			expectedTotalCount: "",
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			villageSvc := &mock.FakeVillageService{}
			tt.setup(versionSvc, serverSvc, villageSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
				withVillageService(villageSvc),
			)

			target := fmt.Sprintf(
				"/v1/versions/%s/servers/%s/villages?%s",
				tt.versionCode,
				tt.serverKey,
				tt.queryParams.Encode(),
			)
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertTotalCount(t, resp.Header, tt.expectedTotalCount)
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}

func TestVillage_getByID(t *testing.T) {
	t.Parallel()

	now := time.Now()
	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	server := domain.Server{
		Key:         "pl151",
		URL:         "https://pl151.plemiona.pl",
		Open:        true,
		Special:     false,
		VersionCode: "pl",
	}
	tests := []struct {
		name             string
		setup            func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService)
		versionCode      string
		serverKey        string
		id               string
		expectedStatus   int
		target           any
		expectedResponse any
	}{
		{
			name: "OK",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				villageSvc.GetByServerKeyAndIDWithRelationsReturns(domain.VillageWithRelations{
					Village: domain.Village{
						BaseVillage: domain.BaseVillage{
							ID:         1234,
							Name:       "name",
							Points:     1,
							X:          2,
							Y:          3,
							Continent:  "K11",
							Bonus:      1,
							PlayerID:   997,
							ProfileURL: "profile-1234",
						},
						ServerKey: server.Key,
						CreatedAt: now,
					},
					Player: domain.NullPlayerMetaWithRelations{
						Valid: true,
						Player: domain.PlayerMetaWithRelations{
							PlayerMeta: domain.PlayerMeta{
								ID:         997,
								Name:       "name 997",
								ProfileURL: "profile-997",
							},
							Tribe: domain.NullTribeMeta{
								Valid: true,
								Tribe: domain.TribeMeta{
									ID:         1234,
									Name:       "name 997",
									Tag:        "tag 997",
									ProfileURL: "profile-1234",
								},
							},
						},
					},
				}, nil)
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			id:             "1234",
			expectedStatus: http.StatusOK,
			target:         &model.GetVillageResp{},
			expectedResponse: &model.GetVillageResp{
				Data: model.Village{
					ID:         1234,
					Name:       "name",
					FullName:   "name (2|3) K11",
					Points:     1,
					X:          2,
					Y:          3,
					Continent:  "K11",
					Bonus:      1,
					ProfileURL: "profile-1234",
					Player: model.NullPlayerMeta{
						Valid: true,
						Player: model.PlayerMeta{
							ID:         997,
							Name:       "name 997",
							ProfileURL: "profile-997",
							Tribe: model.NullTribeMeta{
								Valid: true,
								Tribe: model.TribeMeta{
									ID:         1234,
									Name:       "name 997",
									Tag:        "tag 997",
									ProfileURL: "profile-1234",
								},
							},
						},
					},
					CreatedAt: now,
				},
			},
		},
		{
			name: "ERR: village id is not a valid int64",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			id:             "true",
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "villageId: strconv.ParseInt: parsing \"true\": invalid syntax",
				},
			},
		},
		{
			name: "ERR: version not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			serverKey:      server.Key,
			id:             "1234",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
		},
		{
			name: "ERR: server not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{}, domain.ServerNotFoundError{Key: server.Key + "2"})
			},
			versionCode:    version.Code,
			serverKey:      server.Key + "2",
			id:             "1234",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("server (key=%s) not found", server.Key+"2"),
				},
			},
		},
		{
			name: "ERR: village not found",
			setup: func(versionSvc *mock.FakeVersionService, serverSvc *mock.FakeServerService, villageSvc *mock.FakeVillageService) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				villageSvc.GetByServerKeyAndIDWithRelationsReturns(domain.VillageWithRelations{}, domain.VillageNotFoundError{ID: 12345551})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			id:             "12345551",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: "village (id=12345551) not found",
				},
			},
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			villageSvc := &mock.FakeVillageService{}
			tt.setup(versionSvc, serverSvc, villageSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
				withVillageService(villageSvc),
			)

			target := fmt.Sprintf(
				"/v1/versions/%s/servers/%s/villages/%s",
				tt.versionCode,
				tt.serverKey,
				tt.id,
			)
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}

func TestVillage_listPlayer(t *testing.T) {
	t.Parallel()

	now := time.Now()
	version := domain.Version{
		Code:     "pl",
		Name:     "Poland",
		Host:     "plemiona.pl",
		Timezone: "Europe/Warsaw",
	}
	server := domain.Server{
		Key:         "pl151",
		URL:         "https://pl151.plemiona.pl",
		Open:        true,
		Special:     false,
		VersionCode: "pl",
	}
	player := domain.PlayerWithRelations{
		Player: domain.Player{
			BasePlayer: domain.BasePlayer{
				ID:         124,
				Name:       "player 124",
				ProfileURL: "player-124",
				TribeID:    1233,
			},
		},
		Tribe: domain.NullTribeMeta{
			Valid: true,
			Tribe: domain.TribeMeta{
				ID:         1233,
				Name:       "tribe 1233",
				Tag:        "tag1233",
				ProfileURL: "profile 1233",
			},
		},
	}
	playerID := strconv.FormatInt(player.ID, 10)

	tests := []struct {
		name  string
		setup func(
			versionSvc *mock.FakeVersionService,
			serverSvc *mock.FakeServerService,
			playerSvc *mock.FakePlayerService,
			villageSvc *mock.FakeVillageService,
		)
		versionCode        string
		serverKey          string
		playerID           string
		queryParams        url.Values
		expectedStatus     int
		target             any
		expectedResponse   any
		expectedTotalCount string
	}{
		{
			name: "OK: without params",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				villageSvc *mock.FakeVillageService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
				villageSvc.ListCountWithRelationsCalls(func(_ context.Context, params domain.ListVillagesParams) ([]domain.VillageWithRelations, int64, error) {
					expectedParams := domain.ListVillagesParams{
						ServerKeys: []string{server.Key},
						PlayerID: domain.NullInt64{
							Int64: player.ID,
							Valid: true,
						},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					villages := []domain.VillageWithRelations{
						{
							Village: domain.Village{
								BaseVillage: domain.BaseVillage{
									ID:         1234,
									Name:       "name",
									Points:     1,
									X:          2,
									Y:          3,
									Continent:  "K11",
									Bonus:      1,
									PlayerID:   player.ID,
									ProfileURL: "profile-1234",
								},
								ServerKey: server.Key,
								CreatedAt: now,
							},
							Player: domain.NullPlayerMetaWithRelations{
								Valid: true,
								Player: domain.PlayerMetaWithRelations{
									PlayerMeta: domain.PlayerMeta{
										ID:         player.ID,
										Name:       player.Name,
										ProfileURL: player.ProfileURL,
									},
									Tribe: domain.NullTribeMeta{
										Valid: true,
										Tribe: domain.TribeMeta{
											ID:         player.Tribe.Tribe.ID,
											Name:       player.Tribe.Tribe.Name,
											Tag:        player.Tribe.Tribe.Tag,
											ProfileURL: player.Tribe.Tribe.ProfileURL,
										},
									},
								},
							},
						},
						{
							Village: domain.Village{
								BaseVillage: domain.BaseVillage{
									ID:         12345,
									Name:       "name 2",
									Points:     1,
									X:          22,
									Y:          22,
									Continent:  "K22",
									Bonus:      0,
									PlayerID:   player.ID,
									ProfileURL: "profile-12345",
								},
								ServerKey: server.Key,
								CreatedAt: now,
							},
							Player: domain.NullPlayerMetaWithRelations{
								Valid: true,
								Player: domain.PlayerMetaWithRelations{
									PlayerMeta: domain.PlayerMeta{
										ID:         player.ID,
										Name:       player.Name,
										ProfileURL: player.ProfileURL,
									},
									Tribe: domain.NullTribeMeta{
										Valid: true,
										Tribe: domain.TribeMeta{
											ID:         player.Tribe.Tribe.ID,
											Name:       player.Tribe.Tribe.Name,
											Tag:        player.Tribe.Tribe.Tag,
											ProfileURL: player.Tribe.Tribe.ProfileURL,
										},
									},
								},
							},
						},
					}
					return villages, int64(len(villages)), nil
				})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			playerID:       playerID,
			expectedStatus: http.StatusOK,
			target:         &model.ListVillagesResp{},
			expectedResponse: &model.ListVillagesResp{
				Data: []model.Village{
					{
						ID:         1234,
						Name:       "name",
						FullName:   "name (2|3) K11",
						Points:     1,
						X:          2,
						Y:          3,
						Continent:  "K11",
						Bonus:      1,
						ProfileURL: "profile-1234",
						Player: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         player.ID,
								Name:       player.Name,
								ProfileURL: player.ProfileURL,
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         player.Tribe.Tribe.ID,
										Name:       player.Tribe.Tribe.Name,
										Tag:        player.Tribe.Tribe.Tag,
										ProfileURL: player.Tribe.Tribe.ProfileURL,
									},
								},
							},
						},
						CreatedAt: now,
					},
					{
						ID:         12345,
						Name:       "name 2",
						FullName:   "name 2 (22|22) K22",
						Points:     1,
						X:          22,
						Y:          22,
						Continent:  "K22",
						Bonus:      0,
						ProfileURL: "profile-12345",
						Player: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         player.ID,
								Name:       player.Name,
								ProfileURL: player.ProfileURL,
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         player.Tribe.Tribe.ID,
										Name:       player.Tribe.Tribe.Name,
										Tag:        player.Tribe.Tribe.Tag,
										ProfileURL: player.Tribe.Tribe.ProfileURL,
									},
								},
							},
						},
						CreatedAt: now,
					},
				},
			},
			expectedTotalCount: "2",
		},
		{
			name: "OK: limit=1, offset=15",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				villageSvc *mock.FakeVillageService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
				villageSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListVillagesParams,
				) ([]domain.VillageWithRelations, int64, error) {
					expectedParams := domain.ListVillagesParams{
						ServerKeys: []string{server.Key},
						Pagination: domain.Pagination{
							Limit:  1,
							Offset: 15,
						},
						PlayerID: domain.NullInt64{
							Int64: player.ID,
							Valid: true,
						},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					villages := []domain.VillageWithRelations{
						{
							Village: domain.Village{
								BaseVillage: domain.BaseVillage{
									ID:         12345,
									Name:       "name 2",
									Points:     1,
									X:          22,
									Y:          22,
									Continent:  "K22",
									Bonus:      0,
									PlayerID:   player.ID,
									ProfileURL: "profile-12345",
								},
								ServerKey: server.Key,
								CreatedAt: now,
							},
							Player: domain.NullPlayerMetaWithRelations{
								Valid: true,
								Player: domain.PlayerMetaWithRelations{
									PlayerMeta: domain.PlayerMeta{
										ID:         player.ID,
										Name:       player.Name,
										ProfileURL: player.ProfileURL,
									},
									Tribe: domain.NullTribeMeta{
										Valid: true,
										Tribe: domain.TribeMeta{
											ID:         player.Tribe.Tribe.ID,
											Name:       player.Tribe.Tribe.Name,
											Tag:        player.Tribe.Tribe.Tag,
											ProfileURL: player.Tribe.Tribe.ProfileURL,
										},
									},
								},
							},
						},
					}
					return villages, int64(len(villages)) + 350, nil
				})
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"limit":  []string{"1"},
				"offset": []string{"15"},
			},
			expectedStatus: http.StatusOK,
			target:         &model.ListVillagesResp{},
			expectedResponse: &model.ListVillagesResp{
				Data: []model.Village{
					{
						ID:        12345,
						Name:      "name 2",
						FullName:  "name 2 (22|22) K22",
						Points:    1,
						X:         22,
						Y:         22,
						Continent: "K22",
						Bonus:     0,
						Player: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         player.ID,
								Name:       player.Name,
								ProfileURL: player.ProfileURL,
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         player.Tribe.Tribe.ID,
										Name:       player.Tribe.Tribe.Name,
										Tag:        player.Tribe.Tribe.Tag,
										ProfileURL: player.Tribe.Tribe.ProfileURL,
									},
								},
							},
						},
						ProfileURL: "profile-12345",
						CreatedAt:  now,
					},
				},
			},
			expectedTotalCount: "351",
		},
		{
			name: "OK: coords=[22|23]",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				villageSvc *mock.FakeVillageService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
				villageSvc.ListCountWithRelationsCalls(func(
					_ context.Context,
					params domain.ListVillagesParams,
				) ([]domain.VillageWithRelations, int64, error) {
					expectedParams := domain.ListVillagesParams{
						ServerKeys: []string{server.Key},
						Coords:     []domain.Coords{{X: 22, Y: 23}},
						PlayerID: domain.NullInt64{
							Int64: player.ID,
							Valid: true,
						},
					}
					if diff := cmp.Diff(params, expectedParams); diff != "" {
						fmt.Println(diff)
						return nil, 0, fmt.Errorf("validation failed: %s", diff)
					}

					return []domain.VillageWithRelations{
						{
							Village: domain.Village{
								BaseVillage: domain.BaseVillage{
									ID:         12345,
									Name:       "name 2",
									Points:     1,
									X:          22,
									Y:          23,
									Continent:  "K22",
									Bonus:      0,
									PlayerID:   player.ID,
									ProfileURL: "profile-12345",
								},
								ServerKey: server.Key,
								CreatedAt: now,
							},
							Player: domain.NullPlayerMetaWithRelations{
								Valid: true,
								Player: domain.PlayerMetaWithRelations{
									PlayerMeta: domain.PlayerMeta{
										ID:         player.ID,
										Name:       player.Name,
										ProfileURL: player.ProfileURL,
									},
									Tribe: domain.NullTribeMeta{
										Valid: true,
										Tribe: domain.TribeMeta{
											ID:         player.Tribe.Tribe.ID,
											Name:       player.Tribe.Tribe.Name,
											Tag:        player.Tribe.Tribe.Tag,
											ProfileURL: player.Tribe.Tribe.ProfileURL,
										},
									},
								},
							},
						},
					}, 1, nil
				})
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"coords": []string{"22|23"},
			},
			expectedStatus: http.StatusOK,
			target:         &model.ListVillagesResp{},
			expectedResponse: &model.ListVillagesResp{
				Data: []model.Village{
					{
						ID:        12345,
						Name:      "name 2",
						FullName:  "name 2 (22|23) K22",
						Points:    1,
						X:         22,
						Y:         23,
						Continent: "K22",
						Bonus:     0,
						Player: model.NullPlayerMeta{
							Valid: true,
							Player: model.PlayerMeta{
								ID:         player.ID,
								Name:       player.Name,
								ProfileURL: player.ProfileURL,
								Tribe: model.NullTribeMeta{
									Valid: true,
									Tribe: model.TribeMeta{
										ID:         player.Tribe.Tribe.ID,
										Name:       player.Tribe.Tribe.Name,
										Tag:        player.Tribe.Tribe.Tag,
										ProfileURL: player.Tribe.Tribe.ProfileURL,
									},
								},
							},
						},
						ProfileURL: "profile-12345",
						CreatedAt:  now,
					},
				},
			},
			expectedTotalCount: "1",
		},
		{
			name: "ERR: limit is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				villageSvc *mock.FakeVillageService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"limit": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "limit: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: offset is not a valid int32",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				villageSvc *mock.FakeVillageService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"offset": []string{"asd"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "offset: strconv.ParseInt: parsing \"asd\": invalid syntax",
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: coords - incorrect format 1",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				villageSvc *mock.FakeVillageService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"coords": []string{"22|-22"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "coords[0]: " + domain.ErrIncorrectCoordsFormat.Error(),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: coords - incorrect format 2",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				villageSvc *mock.FakeVillageService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(player, nil)
			},
			versionCode: version.Code,
			serverKey:   server.Key,
			playerID:    playerID,
			queryParams: url.Values{
				"coords": []string{"22|22", "235"},
			},
			expectedStatus: http.StatusBadRequest,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeValidationError.String(),
					Message: "coords[1]: " + domain.ErrIncorrectCoordsFormat.Error(),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: version not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				villageSvc *mock.FakeVillageService,
			) {
				versionSvc.GetByCodeReturns(domain.Version{}, domain.VersionNotFoundError{VerCode: version.Code + "2"})
			},
			versionCode:    version.Code + "2",
			serverKey:      server.Key,
			playerID:       playerID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("version (code=%s) not found", version.Code+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: server not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				villageSvc *mock.FakeVillageService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(domain.Server{}, domain.ServerNotFoundError{Key: server.Key + "2"})
			},
			versionCode:    version.Code,
			serverKey:      server.Key + "2",
			playerID:       playerID,
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("server (key=%s) not found", server.Key+"2"),
				},
			},
			expectedTotalCount: "",
		},
		{
			name: "ERR: player not found",
			setup: func(
				versionSvc *mock.FakeVersionService,
				serverSvc *mock.FakeServerService,
				playerSvc *mock.FakePlayerService,
				villageSvc *mock.FakeVillageService,
			) {
				versionSvc.GetByCodeReturns(version, nil)
				serverSvc.GetNormalByVersionCodeAndKeyReturns(server, nil)
				playerSvc.GetByServerKeyAndIDWithRelationsReturns(domain.PlayerWithRelations{}, domain.PlayerNotFoundError{ID: 1241})
			},
			versionCode:    version.Code,
			serverKey:      server.Key,
			playerID:       playerID + "1",
			expectedStatus: http.StatusNotFound,
			target:         &model.ErrorResp{},
			expectedResponse: &model.ErrorResp{
				Error: model.APIError{
					Code:    domain.ErrorCodeEntityNotFound.String(),
					Message: fmt.Sprintf("player (id=%s) not found", playerID+"1"),
				},
			},
			expectedTotalCount: "",
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			versionSvc := &mock.FakeVersionService{}
			serverSvc := &mock.FakeServerService{}
			playerSvc := &mock.FakePlayerService{}
			villageSvc := &mock.FakeVillageService{}
			tt.setup(versionSvc, serverSvc, playerSvc, villageSvc)

			router := newRouter(
				withVersionService(versionSvc),
				withServerService(serverSvc),
				withVillageService(villageSvc),
				withPlayerService(playerSvc),
			)

			target := fmt.Sprintf(
				"/v1/versions/%s/servers/%s/players/%s/villages?%s",
				tt.versionCode,
				tt.serverKey,
				tt.playerID,
				tt.queryParams.Encode(),
			)
			resp := doRequest(router, http.MethodGet, target, nil)
			defer resp.Body.Close()
			assertTotalCount(t, resp.Header, tt.expectedTotalCount)
			assertJSONResponse(t, resp, tt.expectedStatus, tt.expectedResponse, tt.target)
		})
	}
}
