package service

import (
	"context"
	"fmt"
	"time"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
)

const (
	playerSnapshotMaxLimit   = 100
	playerSnapshotSortMaxLen = 2
)

//counterfeiter:generate -o internal/mock/player_lister.gen.go . PlayerLister
type PlayerLister interface {
	List(ctx context.Context, params domain.ListPlayersParams) ([]domain.Player, error)
}

//counterfeiter:generate -o internal/mock/player_snapshot_repository.gen.go . PlayerSnapshotRepository
type PlayerSnapshotRepository interface {
	Create(ctx context.Context, params ...domain.CreatePlayerSnapshotParams) error
	ListCountWithRelations(ctx context.Context, params domain.ListPlayerSnapshotsParams) ([]domain.PlayerSnapshotWithRelations, int64, error)
	Delete(ctx context.Context, serverKey string, createdAtLTE time.Time) error
}

type PlayerSnapshot struct {
	repo      PlayerSnapshotRepository
	playerSvc PlayerLister
}

func NewPlayerSnapshot(repo PlayerSnapshotRepository, playerSvc PlayerLister) *PlayerSnapshot {
	return &PlayerSnapshot{repo: repo, playerSvc: playerSvc}
}

func (p *PlayerSnapshot) Create(ctx context.Context, key string, date time.Time) error {
	var lastID int64
	for {
		players, err := p.playerSvc.List(ctx, domain.ListPlayersParams{
			ServerKeys: []string{key},
			IDGT: domain.NullInt64{
				Int64: lastID,
				Valid: true,
			},
			Deleted: domain.NullBool{
				Valid: true,
				Bool:  false,
			},
			Pagination: domain.Pagination{
				Limit: playerMaxLimit,
			},
			Sort: []domain.PlayerSort{
				{By: domain.PlayerSortByID, Direction: domain.SortDirectionASC},
			},
		})
		if err != nil {
			return fmt.Errorf("PlayerRepository.List: %w", err)
		}

		params := make([]domain.CreatePlayerSnapshotParams, 0, len(players))
		for _, player := range players {
			params = append(params, domain.CreatePlayerSnapshotParams{
				OpponentsDefeated: player.OpponentsDefeated,
				PlayerID:          player.ID,
				NumVillages:       player.NumVillages,
				Points:            player.Points,
				Rank:              player.Rank,
				TribeID:           player.TribeID,
				ServerKey:         player.ServerKey,
				Date:              date,
			})
		}

		if err = p.repo.Create(ctx, params...); err != nil {
			return fmt.Errorf("PlayerSnapshotRepository.Create: %w", err)
		}

		if len(players) < playerMaxLimit {
			break
		}

		lastID = players[len(players)-1].ID
	}

	return nil
}

func (p *PlayerSnapshot) ListCountWithRelations(
	ctx context.Context,
	params domain.ListPlayerSnapshotsParams,
) ([]domain.PlayerSnapshotWithRelations, int64, error) {
	if len(params.Sort) == 0 {
		params.Sort = []domain.PlayerSnapshotSort{
			{
				By:        domain.PlayerSnapshotSortByDate,
				Direction: domain.SortDirectionASC,
			},
			{
				By:        domain.PlayerSnapshotSortByID,
				Direction: domain.SortDirectionASC,
			},
		}
	}

	if params.Pagination.Limit == 0 {
		params.Pagination.Limit = playerSnapshotMaxLimit
	}

	if len(params.Sort) > playerSnapshotSortMaxLen {
		return nil, 0, domain.ValidationError{
			Field: "sort",
			Err: domain.MaxLenError{
				Max: playerSnapshotSortMaxLen,
			},
		}
	}

	if err := validatePagination(params.Pagination, playerSnapshotMaxLimit); err != nil {
		return nil, 0, err
	}

	snapshots, count, err := p.repo.ListCountWithRelations(ctx, params)
	if err != nil {
		return nil, 0, fmt.Errorf("PlayerSnapshotRepository.ListCountWithRelations: %w", err)
	}

	return snapshots, count, nil
}

func (p *PlayerSnapshot) CleanUp(ctx context.Context, srv domain.Server) error {
	if srv.Special || srv.Open || srv.PlayerSnapshotsCreatedAt.After(time.Now().Add(-30*24*time.Hour) /* 30 days */) {
		return nil
	}

	// delete snapshots older than 6 months
	if err := p.repo.Delete(ctx, srv.Key, time.Now().Add(-180*24*time.Hour)); err != nil {
		return fmt.Errorf("couldn't delete old player snapshots: %w", err)
	}

	return nil
}
