package service_test

import (
	"context"
	"errors"
	"fmt"
	"testing"

	"github.com/google/go-cmp/cmp"

	"github.com/stretchr/testify/assert"

	"gitea.dwysokinski.me/twhelp/core/internal/service"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
	"gitea.dwysokinski.me/twhelp/core/internal/service/internal/mock"
)

func TestServer_Refresh(t *testing.T) {
	t.Parallel()

	url := "https://www.plemiona.pl"
	versionCode := "pl"

	openServers := []domain.OpenServer{
		{
			Key: "pl151",
			URL: "https://pl151.plemiona.pl",
		},
		{
			Key: "pl152",
			URL: "https://pl152.plemiona.pl",
		},
		{
			Key: "pls1",
			URL: "https://pls1.plemiona.pl",
		},
	}
	client := &mock.FakeOpenServerConfigInfoGetter{}
	client.GetOpenServersCalls(func(ctx context.Context, baseURL string) ([]domain.OpenServer, error) {
		if baseURL != url {
			return nil, errors.New("incorrect baseURL")
		}
		return openServers, nil
	})

	specialServers := []domain.Server{
		{
			Key:         "pls1",
			URL:         "https://pls1.plemiona.pl",
			Open:        false,
			Special:     true,
			VersionCode: versionCode,
		},
	}
	currentlyStoredOpenServers := []domain.Server{
		{
			Key:         "pl150",
			URL:         "https://pl150.plemiona.pl",
			Open:        true,
			Special:     false,
			VersionCode: versionCode,
		},
		{
			Key:         "pl151",
			URL:         "https://pl151.plemiona.pl",
			Open:        true,
			Special:     false,
			VersionCode: versionCode,
		},
	}
	repo := &mock.FakeServerRepository{}
	repo.ListReturnsOnCall(0, specialServers, nil)
	repo.ListReturnsOnCall(1, currentlyStoredOpenServers, nil)
	repo.CreateOrUpdateCalls(func(ctx context.Context, params ...domain.CreateServerParams) ([]domain.Server, error) {
		servers := make([]domain.Server, 0, len(params))
		for _, p := range params {
			servers = append(servers, domain.Server{
				Key:         p.Key,
				URL:         p.URL,
				Open:        p.Open,
				VersionCode: p.VersionCode,
			})
		}
		return servers, nil
	})

	res, err := service.NewServer(repo, client).Refresh(context.Background(), url, versionCode)
	assert.NoError(t, err)
	assert.Equal(t, []domain.Server{
		{
			Key:         "pl151",
			URL:         "https://pl151.plemiona.pl",
			Open:        true,
			VersionCode: versionCode,
		},
		{
			Key:         "pl152",
			URL:         "https://pl152.plemiona.pl",
			Open:        true,
			VersionCode: versionCode,
		},
		{
			Key:         "pl150",
			URL:         "https://pl150.plemiona.pl",
			Open:        false,
			VersionCode: versionCode,
		},
	}, res)
}

func TestServer_List_ListCount(t *testing.T) {
	t.Parallel()

	var defaultLimit int32 = 500
	tests := []struct {
		name           string
		params         domain.ListServersParams
		expectedParams domain.ListServersParams
		expectedErr    error
	}{
		{
			name:   "OK: default limit",
			params: domain.ListServersParams{},
			expectedParams: domain.ListServersParams{
				Pagination: domain.Pagination{
					Limit: defaultLimit,
				},
			},
		},
		{
			name: "OK: custom limit",
			params: domain.ListServersParams{
				Pagination: domain.Pagination{
					Limit:  499,
					Offset: 1,
				},
			},
			expectedParams: domain.ListServersParams{
				Pagination: domain.Pagination{
					Limit:  499,
					Offset: 1,
				},
			},
		},
		{
			name: "ERR: params.Pagination.Limit < 0",
			params: domain.ListServersParams{
				Pagination: domain.Pagination{
					Limit: -1,
				},
			},
			expectedErr: domain.ValidationError{
				Field: "limit",
				Err: domain.MinError{
					Min: 1,
				},
			},
		},
		{
			name: "ERR: params.Pagination.Limit > 500",
			params: domain.ListServersParams{
				Pagination: domain.Pagination{
					Limit: 501,
				},
			},
			expectedErr: domain.ValidationError{
				Field: "limit",
				Err: domain.MaxError{
					Max: 500,
				},
			},
		},
		{
			name: "ERR: params.Pagination.Offset < 0",
			params: domain.ListServersParams{
				Pagination: domain.Pagination{
					Offset: -1,
				},
			},
			expectedErr: domain.ValidationError{
				Field: "offset",
				Err: domain.MinError{
					Min: 0,
				},
			},
		},
	}

	for _, tt := range tests {
		tt := tt

		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			repo := &mock.FakeServerRepository{}
			repo.ListCountCalls(func(_ context.Context, params domain.ListServersParams) ([]domain.Server, int64, error) {
				if diff := cmp.Diff(params, tt.expectedParams); diff != "" {
					return nil, 0, fmt.Errorf("validation failed: %s", diff)
				}
				return make([]domain.Server, params.Pagination.Limit), int64(params.Pagination.Limit), nil
			})
			repo.ListCalls(func(ctx context.Context, params domain.ListServersParams) ([]domain.Server, error) {
				servers, _, err := repo.ListCount(ctx, params)
				return servers, err
			})

			svc := service.NewServer(repo, &mock.FakeOpenServerConfigInfoGetter{})

			serversListCount, count, err := svc.ListCount(context.Background(), tt.params)
			assert.ErrorIs(t, err, tt.expectedErr)
			assert.EqualValues(t, tt.expectedParams.Pagination.Limit, count)
			assert.Len(t, serversListCount, int(tt.expectedParams.Pagination.Limit))

			serversList, err := svc.List(context.Background(), tt.params)
			assert.ErrorIs(t, err, tt.expectedErr)
			assert.Equal(t, serversListCount, serversList)
		})
	}
}

func TestServer_GetNormalByVersionCodeAndKey(t *testing.T) {
	t.Parallel()

	t.Run("OK", func(t *testing.T) {
		t.Parallel()

		server := domain.Server{
			Key:         "pl151",
			VersionCode: "pl",
		}
		repo := &mock.FakeServerRepository{}
		repo.ListReturns([]domain.Server{server}, nil)
		repo.ListCalls(func(ctx context.Context, params domain.ListServersParams) ([]domain.Server, error) {
			if diff := cmp.Diff(domain.ListServersParams{
				Special: domain.NullBool{
					Bool:  false,
					Valid: true,
				},
				Keys:         []string{server.Key},
				VersionCodes: []string{server.VersionCode},
				Pagination: domain.Pagination{
					Limit: 1,
				},
			}, params); diff != "" {
				return nil, fmt.Errorf("validation failed: %s", diff)
			}

			return []domain.Server{server}, nil
		})

		res, err := service.NewServer(repo, &mock.FakeOpenServerConfigInfoGetter{}).
			GetNormalByVersionCodeAndKey(context.Background(), server.VersionCode, server.Key)
		assert.NoError(t, err)
		assert.Equal(t, server, res)
	})

	t.Run("ERR: server not found", func(t *testing.T) {
		t.Parallel()

		repo := &mock.FakeServerRepository{}
		repo.ListReturns(nil, nil)

		key := "pl151"
		res, err := service.NewServer(repo, &mock.FakeOpenServerConfigInfoGetter{}).
			GetNormalByVersionCodeAndKey(context.Background(), "en", key)
		assert.ErrorIs(t, err, domain.ServerNotFoundError{Key: key})
		assert.Zero(t, res)
	})
}
