package service

import (
	"context"
	"fmt"

	"gitea.dwysokinski.me/twhelp/core/internal/domain"
)

const (
	tribeChangeChunkSize  = 500
	tribeChangeMaxLimit   = 100
	tribeChangeSortMaxLen = 2
)

//counterfeiter:generate -o internal/mock/tribe_change_repository.gen.go . TribeChangeRepository
type TribeChangeRepository interface {
	Create(ctx context.Context, params ...domain.CreateTribeChangeParams) error
	ListCountWithRelations(ctx context.Context, params domain.ListTribeChangesParams) ([]domain.TribeChangeWithRelations, int64, error)
}

type TribeChange struct {
	repo TribeChangeRepository
}

func NewTribeChange(repo TribeChangeRepository) *TribeChange {
	return &TribeChange{repo: repo}
}

func (t *TribeChange) Create(ctx context.Context, params ...domain.CreateTribeChangeParams) error {
	for i := 0; i < len(params); i += tribeChangeChunkSize {
		end := i + tribeChangeChunkSize
		if end > len(params) {
			end = len(params)
		}

		chunk := params[i:end]

		if err := t.repo.Create(ctx, chunk...); err != nil {
			return fmt.Errorf("TribeChangeRepository.Create: %w", err)
		}
	}

	return nil
}

func (t *TribeChange) ListCountWithRelations(
	ctx context.Context,
	params domain.ListTribeChangesParams,
) ([]domain.TribeChangeWithRelations, int64, error) {
	if len(params.Sort) == 0 {
		params.Sort = []domain.TribeChangeSort{
			{
				By:        domain.TribeChangeSortByCreatedAt,
				Direction: domain.SortDirectionASC,
			},
			{
				By:        domain.TribeChangeSortByID,
				Direction: domain.SortDirectionASC,
			},
		}
	}

	if params.Pagination.Limit == 0 {
		params.Pagination.Limit = tribeChangeMaxLimit
	}

	if len(params.Sort) > tribeChangeSortMaxLen {
		return nil, 0, domain.ValidationError{
			Field: "sort",
			Err: domain.MaxLenError{
				Max: tribeChangeSortMaxLen,
			},
		}
	}

	if err := validatePagination(params.Pagination, tribeChangeMaxLimit); err != nil {
		return nil, 0, err
	}

	tcs, count, err := t.repo.ListCountWithRelations(ctx, params)
	if err != nil {
		return nil, 0, fmt.Errorf("TribeChangeRepository.ListCountWithRelations: %w", err)
	}

	return tcs, count, nil
}
